"""API handlers for reviews of click packages."""
from django.db.models import Count
from django.http import HttpResponseBadRequest
from django.shortcuts import get_object_or_404
from piston.handler import AnonymousBaseHandler, BaseHandler
from piston.utils import translate_mime

from clickreviews.models import (
    ClickPackage,
    ClickPackageModeration,
    ClickPackageReview,
    ClickPackageReviewModeration,
)
from clickreviews.forms import (
    ClickPackageFlagForm,
    ClickPackageModerationStatusForm,
    ClickPackageReviewFilterForm,
    ClickPackageReviewFlagForm,
    ClickPackageReviewForm,
    ClickPackageReviewModerationStatusForm,
    ClickPackageReviewUpdateForm,
)
from core.api.pagination import make_paginated_response
from core.models import BaseModeration


REVIEW_FIELDS = (
    'id',
    'package_name',
    'language',
    'date_created',
    'rating',
    'reviewer_username',
    'reviewer_displayname',
    'summary',
    'review_text',
    'hide',
    'date_deleted',
    'version',
    'usefulness_total',
    'usefulness_favorable',
)


class ClickReviewAnonymousHandler(AnonymousBaseHandler):
    allowed_methods = ('GET',)
    fields = REVIEW_FIELDS

    def read(self, request):
        """Return the reviews for a given package name.

        Currently the api supports filtering by package name only.
        """
        form = ClickPackageReviewFilterForm(request.GET)
        if not form.is_valid():
            return HttpResponseBadRequest(form.errors_json())

        package_name = form.cleaned_data['package_name']

        return (ClickPackageReview.objects.published()
                .prefetch_related('reviewer__useropenid_set')
                .select_related('click_package', 'reviewer')
                .filter(click_package__package_name=package_name)
                # XXX: Order by date until usefulness is implemented.
                .order_by('-date_created'))


class ClickReviewHandler(BaseHandler):
    allowed_methods = ('DELETE', 'POST', 'PUT')
    anonymous = ClickReviewAnonymousHandler
    fields = REVIEW_FIELDS

    def create(self, request):
        review = ClickPackageReview(reviewer=request.user)
        form = ClickPackageReviewForm(request.data, instance=review)
        if not form.is_valid():
            return HttpResponseBadRequest(form.errors_json())
        review = form.save()

        # Update stats on the software item
        review.click_package.update_stats()

        return review

    def update(self, request, review_id):
        """Use PUT verb to modify an existing review."""
        review = get_object_or_404(
            ClickPackageReview, id=review_id, reviewer=request.user)

        form = ClickPackageReviewUpdateForm(request.data, instance=review)
        if form.is_valid():
            review = form.save()
            # XXX: Refactor this to be triggered on review save/delete.
            review.click_package.update_stats()
            response = review
        else:
            response = HttpResponseBadRequest(form.errors_json())
        return response

    def delete(self, request, review_id):
        review = get_object_or_404(
            ClickPackageReview.objects.published(),
            id=review_id, reviewer=request.user)
        review.delete()
        return review


class ClickReviewsStatsHandler(BaseHandler):
    allowed_methods = ('GET',)
    fields = ('id', 'histogram', 'package_name',
              'ratings_average', 'ratings_total')

    def read(self, request, package_name=None):
        # We do not allow requests with GET parameters which avoid
        # django's caching system (risk of DoS attack).
        if request.GET:
            return HttpResponseBadRequest(
                "This api call does not take GET parameters.")

        if package_name is not None:
            return get_object_or_404(
                ClickPackage, package_name=package_name)

        return ClickPackage.objects.all()


class ClickPackageFlagHandler(BaseHandler):
    allowed_methods = ('POST',)
    fields = ('id', 'summary', 'description', 'date_created')

    def create(self, request, package_name):
        user = request.user
        data = request.POST.copy()
        data.update({'user': user.pk, 'package_name': package_name})
        form = ClickPackageFlagForm(data)

        if form.is_valid():
            response = form.save()
        else:
            response = HttpResponseBadRequest(form.errors_json())

        return response


class ClickPackageReviewFlagHandler(BaseHandler):
    allowed_methods = ('POST',)
    fields = ('id', 'summary', 'description', 'date_created')

    def create(self, request, review_id):
        review = get_object_or_404(ClickPackageReview, pk=review_id)
        user = request.user
        data = request.POST.copy()
        data.update({'user': user.pk, 'review': review.pk})
        form = ClickPackageReviewFlagForm(data)

        if form.is_valid():
            response = form.save()
        else:
            response = HttpResponseBadRequest(form.errors_json())

        return response


class BaseClickPackageModerationHandler(BaseHandler):

    STATUS_MAPPINGS = {
        'pending': BaseModeration.PENDING_STATUS,
        'approved': BaseModeration.APPROVED_STATUS,
        'rejected': BaseModeration.REJECTED_STATUS
    }

    # Using '_model' to not clash with piston's API.
    _model = None
    _form = None

    def _prepare_queryset(self, queryset):
        return queryset

    def read(self, request, status=None, pk=None):
        if status is None:
            queryset = self._model.objects.all()
        elif status == 'pending':
            queryset = self._model.objects.pending()
        elif status == 'approved':
            queryset = self._model.objects.approved()
        elif status == 'rejected':
            queryset = self._model.objects.rejected()

        # Sort by number of flags; in case of tie, give priority to
        # moderations created earlier.
        queryset = (queryset.annotate(num_flags=Count('flag_set'))
                    .order_by('-num_flags', 'date_created')
                    .select_related()
                    .prefetch_related('flag_set__user'))
        queryset = self._prepare_queryset(queryset)

        if pk is not None:
            return get_object_or_404(queryset, pk=pk)
        else:
            return make_paginated_response(request, queryset)

    def _normalize_request_data(self, request):
        # Workarounds for piston:
        #  1) Piston won't translate request data when method is PATCH
        #  2) request.data could be absent or be None in absence of data
        translate_mime(request)
        if not hasattr(request, 'data') or request.data is None:
            request.data = {}

    def patch(self, request, status, pk):
        """Set status for a given moderation.

        Required fields in request data:
            moderator: An openid identifier string.
            moderation_text: A string with related comments.

        """
        self._normalize_request_data(request)
        moderation = get_object_or_404(self._model, pk=pk)
        data = {'moderator': request.data.get('moderator'),
                'status': self.STATUS_MAPPINGS.get(status),
                'moderation_text': request.data.get('moderation_text')}
        form = self._form(data, instance=moderation)
        if form.is_valid():
            response = form.save()
        else:
            response = HttpResponseBadRequest(form.errors_json())
        return response


class ClickPackageModerationHandler(BaseClickPackageModerationHandler):
    _model = ClickPackageModeration
    _form = ClickPackageModerationStatusForm
    allowed_methods = ('GET', 'PATCH')
    fields = (
        'id',
        'date_created',
        'date_moderated',
        'get_status_display',
        'moderation_text',
        ('moderator', ('id',
                       'email',
                       'first_name',
                       'get_full_name',
                       'last_name')),
        ('package', ('id',
                     'date_ratings_updated',
                     'histogram',
                     'package_name',
                     'ratings_average',
                     'ratings_total')),
        ('flags', (
            'date_created',
            'description',
            'summary',
            ('user', ('id',
                      'email',
                      'first_name',
                      'get_full_name',
                      'last_name')))),
        'status')


class ClickPackageReviewModerationHandler(BaseClickPackageModerationHandler):
    _model = ClickPackageReviewModeration
    _form = ClickPackageReviewModerationStatusForm
    allowed_methods = ('GET', 'PATCH')
    fields = (
        'id',
        'date_created',
        'date_moderated',
        'get_status_display',
        'moderation_text',
        ('moderator', ('id',
                       'email',
                       'first_name',
                       'get_full_name',
                       'last_name')),
        # Expose the full reviewer user data
        ('review', REVIEW_FIELDS + (
            ('reviewer', ('id',
                          'email',
                          'first_name',
                          'get_full_name',
                          'last_name')),)),
        ('flags', (
            'date_created',
            'description',
            'summary',
            ('user', ('id',
                      'email',
                      'first_name',
                      'get_full_name',
                      'last_name')))),
        'status')

    def _prepare_queryset(self, queryset):
        return queryset.prefetch_related('review__reviewer__useropenid_set')
