"""URI resources for click reviews API."""
from django.conf import settings
from django.conf.urls import url, patterns
from django.views.decorators.cache import (
    cache_control,
    never_cache,
)

from piston.resource import Resource
from core.api.auth import (
    DelegatedSSOAuthentication,
    SSOOAuthAuthentication,
)
from core.api.decorators import gzip_content, vary_only_on_headers
from core.api.resource import CSRFExemptResource
from clickreviews.api.handlers import (
    ClickReviewHandler,
    ClickReviewsStatsHandler,
    ClickPackageFlagHandler,
    ClickPackageReviewFlagHandler,
    ClickPackageModerationHandler,
    ClickPackageReviewModerationHandler,
)


auth = SSOOAuthAuthentication(realm="Ratings and Reviews")

review_resource = CSRFExemptResource(handler=ClickReviewHandler,
                                     authentication=auth)
review_resource.callmap['PATCH'] = 'patch'
review_resource = never_cache(review_resource)

# Although the review stats resource will be cached by the cache
# middleware, we update the cache-control so that upstream caches will
# keep it around for much longer (CACHE_REVIEW_STATS_SECONDS).
reviews_stats_resource = Resource(handler=ClickReviewsStatsHandler)
cached_resource = cache_control(max_age=settings.CACHE_REVIEW_STATS_SECONDS)
reviews_stats_resource.__name__ = "Resource"
reviews_stats_resource = vary_only_on_headers('Accept', 'Accept-Encoding')(
    gzip_content(cached_resource(reviews_stats_resource)))


click_package_flag_resource = CSRFExemptResource(
    handler=ClickPackageFlagHandler,
    authentication=auth)
click_package_flag_resource.__name__ = "Resource"
click_package_flag_resource = never_cache(click_package_flag_resource)

click_package_review_flag_resource = CSRFExemptResource(
    handler=ClickPackageReviewFlagHandler,
    authentication=auth)
click_package_review_flag_resource.__name__ = "Resource"
click_package_review_flag_resource = never_cache(
    click_package_review_flag_resource)

sca_delegated_sso_auth = DelegatedSSOAuthentication(
    secret=settings.SCA_DELEGATED_SSO_SECRET, realm="Ratings and Reviews")

click_package_moderations_resource = CSRFExemptResource(
    handler=ClickPackageModerationHandler,
    authentication=sca_delegated_sso_auth)
click_package_moderations_resource.callmap['PATCH'] = 'patch'
click_package_moderations_resource.__name__ = "Resource"
click_package_moderations_resource = never_cache(
    click_package_moderations_resource)

click_package_review_moderations_resource = CSRFExemptResource(
    handler=ClickPackageReviewModerationHandler,
    authentication=sca_delegated_sso_auth)
click_package_review_moderations_resource.callmap['PATCH'] = 'patch'
click_package_review_moderations_resource.__name__ = "Resource"
click_package_review_moderations_resource = never_cache(
    click_package_review_moderations_resource)

moderation_status_re = r'(?P<status>(pending|approved|rejected))'

urlpatterns = patterns(
    '',
    url(r'^1.0/reviews/$', review_resource, name='click-api-reviews'),
    url(r'^1.0/reviews/(?P<review_id>\d+)/$',
        review_resource, name='click-api-reviews'),
    url(r'^1.0/review-stats/$', reviews_stats_resource,
        name='click-api-reviews-stats'),
    url(r'^1.0/review-stats/(?P<package_name>[^/]+)/$',
        reviews_stats_resource, name='click-api-reviews-stats'),
    # flag inappropriate click packages
    url(r'^1.0/packages/(?P<package_name>[^/]+)/flags/$',
        click_package_flag_resource, name='click-api-package-flags'),
    # flag inappropriate reviews
    url(r'^1.0/reviews/(?P<review_id>\d+)/flags/$',
        click_package_review_flag_resource, name='click-api-review-flags'),
    url(r'^1.0/package-moderations/$', click_package_moderations_resource,
        name='click-api-package-moderations'),
    url(r'^1.0/package-moderations/(?P<pk>\d+)/$',
        click_package_moderations_resource,
        name='click-api-package-moderations'),
    url(r'^1.0/package-moderations/%s/$' % moderation_status_re,
        click_package_moderations_resource,
        name='click-api-package-moderations'),
    url(r'^1.0/package-moderations/%s/(?P<pk>\d+)/$' % moderation_status_re,
        click_package_moderations_resource,
        name='click-api-package-moderations'),
    url(r'^1.0/review-moderations/$',
        click_package_review_moderations_resource,
        name='click-api-review-moderations'),
    url(r'^1.0/review-moderations/(?P<pk>\d+)/$',
        click_package_review_moderations_resource,
        name='click-api-review-moderations'),
    url(r'^1.0/review-moderations/%s/(?P<pk>\d+)/$' % moderation_status_re,
        click_package_review_moderations_resource,
        name='click-api-review-moderations'),
    url(r'^1.0/review-moderations/%s/$' % moderation_status_re,
        click_package_review_moderations_resource,
        name='click-api-review-moderations'),
)
