from django import forms
from django.conf import settings
from django.core import validators
from django.utils.translation import ugettext as _

from clickreviews.models import (
    ClickPackage,
    ClickPackageFlag,
    ClickPackageModeration,
    ClickPackageReview,
    ClickPackageReviewFlag,
    ClickPackageReviewModeration,
)
from clickreviews.utilities import verify_click_package_published
from core.forms import JSONErrorMixin
from core.models import BaseModeration
from core.utilities import update_or_create_user_from_oauth


class ClickPackageReviewForm(forms.ModelForm, JSONErrorMixin):

    # The following extra fields are used to calculate their model
    # equivalents.
    package_name = forms.CharField()

    class Meta:
        model = ClickPackageReview
        exclude = ('click_package', 'reviewer', 'usefulness_total',
                   'usefulness_favorable', 'usefulness_percentage',
                   'usefulness_wilson_rating')

    def save(self, commit=True):
        """Save the extra model data obtained from fields above."""
        review = super(ClickPackageReviewForm, self).save(commit=False)

        review.click_package = self.cleaned_data['click_package']
        if commit:
            review.save()
        return review

    def clean(self):
        """Check fields that are dependent on each other."""
        # If the supplied data is not valid, no point in continuing.
        if not self.is_valid():
            return self.cleaned_data

        if not settings.ALLOW_MULTIPLE_REVIEWS_FOR_TESTING:
            # Ensure a user cannot submit a second review for the same app
            # unless their first one has been deleted by them
            previous_reviews = ClickPackageReview.objects.filter(
                reviewer=self.instance.reviewer,
                click_package__package_name=self.cleaned_data['package_name'],
                date_deleted=None)
            if previous_reviews:
                raise forms.ValidationError(
                    "A user cannot create multiple reviews for an app.")

        return self.cleaned_data

    def clean_package_name(self):
        """Validate that the click package exists.

        If it does not exist in the db, we check the click-package-index
        and create it if it is valid.
        """
        package_name = self.cleaned_data['package_name']

        try:
            click_package = ClickPackage.objects.get(
                package_name=package_name)
            self.cleaned_data['click_package'] = click_package
            return package_name
        except ClickPackage.DoesNotExist:
            pass

        exists_in_index = verify_click_package_published(package_name)

        if exists_in_index:
            click_package = ClickPackage.objects.create(
                package_name=package_name)
            self.cleaned_data['click_package'] = click_package
            return package_name

        raise forms.ValidationError(
            "Click package '{0}' is not published.".format(
                package_name))


class ClickPackageReviewUpdateForm(forms.ModelForm, JSONErrorMixin):

    can_modify = forms.BooleanField(required=False, widget=forms.HiddenInput)

    def clean_can_modify(self):
        review = self.instance
        if review.is_awaiting_moderation():
            raise forms.ValidationError(
                "This review cannot be edited (pending moderation).")
        return True

    class Meta:
        model = ClickPackageReview
        fields = ('summary', 'review_text', 'rating')


class ClickPackageReviewFilterForm(forms.Form, JSONErrorMixin):
    package_name = forms.CharField(max_length=255)


class BaseClickPackageFlagForm(forms.ModelForm, JSONErrorMixin):

    DESCRIPTION_MAX_LENGTH = 4098

    def __init__(self, *args, **kwargs):
        super(BaseClickPackageFlagForm, self).__init__(*args, **kwargs)
        # Impose a virtually unlimited max_length for description:
        # someone writting more than 4MB of review probably should be
        # thinking about writting a book instead.  Why do this here?
        # because we don't want to enforce this limit at the database
        # level but at validation until we settle in a fixed value.
        self.fields['description'].validators.append(
            validators.MaxLengthValidator(self.DESCRIPTION_MAX_LENGTH))

    # XXX: Django 1.7 has this one, delete it *after* upgrading.
    def add_error(self, field, msg):
        self._errors[field] = self.error_class([msg])


class ClickPackageFlagForm(BaseClickPackageFlagForm):

    package_name = forms.CharField()

    class Meta:
        model = ClickPackageFlag
        fields = ('package_name', 'summary', 'description', 'user')

    def clean_package_name(self):
        """Validate package_name and create ClickPackage if needed.

        When a package doesn't exist, its full name is checked against
        CPI for validity and if it's valid, it's created.  When
        validation checks pass, the ClickPackage instance is stored in
        cleaned_data's 'package' key.
        """
        package_name = self.cleaned_data['package_name']
        click_package = None

        try:
            click_package = ClickPackage.objects.get(
                package_name=package_name)
        except ClickPackage.DoesNotExist:
            exists_in_index = verify_click_package_published(package_name)
            if exists_in_index:
                # As `verify_click_package_published` takes a bit of
                # time, at this point we may have several requests
                # competing to reach for creating an instance,
                # get_or_create adds a bit of safety for that case.
                click_package, _ = ClickPackage.objects.get_or_create(
                    package_name=package_name)

        if click_package is not None:
            self.cleaned_data['package'] = click_package
            return package_name
        else:
            msg = "Click package '{0}' is not published.".format(package_name)
            self.add_error('package_name', msg)

    def clean(self):
        cleaned_data = super(ClickPackageFlagForm, self).clean()
        package = cleaned_data.get('package')
        if package is not None:
            user_already_flagged = ClickPackageFlag.objects.filter(
                user=cleaned_data['user'],
                moderation__package=package.pk).exists()
            if user_already_flagged:
                msg = _('User cannot flag same package twice.')
                self.add_error('package_name', msg)
        return cleaned_data

    def save(self, *args, **kwargs):
        package = self.cleaned_data['package']
        moderation, _ = ClickPackageModeration.objects.get_or_create(
            package=package, status=ClickPackageModeration.PENDING_STATUS)
        self.instance.moderation = moderation
        return super(ClickPackageFlagForm, self).save(*args, **kwargs)


class ClickPackageReviewFlagForm(BaseClickPackageFlagForm):

    review = forms.ModelChoiceField(queryset=ClickPackageReview.objects.all())

    class Meta:
        model = ClickPackageReviewFlag
        fields = ('review', 'summary', 'description', 'user')

    def clean(self):
        cleaned_data = super(ClickPackageReviewFlagForm, self).clean()
        review = cleaned_data.get('review')
        if review is not None:
            user_already_flagged = ClickPackageReviewFlag.objects.filter(
                user=cleaned_data['user'],
                moderation__review=review.pk).exists()
            if user_already_flagged:
                msg = _('User cannot flag same review twice.')
                self.add_error('review', msg)
        return cleaned_data

    def save(self, *args, **kwargs):
        review = self.cleaned_data['review']
        moderation, _ = ClickPackageReviewModeration.objects.get_or_create(
            review=review, status=ClickPackageModeration.PENDING_STATUS)
        self.instance.moderation = moderation
        return super(ClickPackageReviewFlagForm, self).save(*args, **kwargs)


class BaseClickPackageModerationStatusForm(forms.ModelForm, JSONErrorMixin):

    moderator = forms.CharField(max_length=64)

    def __init__(self, *args, **kwargs):
        super(BaseClickPackageModerationStatusForm, self).__init__(
            *args, **kwargs)
        self.fields['moderation_text'].required = True

    # XXX: Django 1.7 has this one, delete it *after* upgrading.
    def add_error(self, field, msg):
        self._errors[field] = self.error_class([msg])

    def clean_status(self):
        status = self.cleaned_data.get('status')

        if status == BaseModeration.PENDING_STATUS:
            self.add_error(
                'status', _('Moderation status cannot be set to pending.'))
        elif self.instance.status != BaseModeration.PENDING_STATUS:
            self.add_error(
                'status', _('Moderation must be pending to change status.'))

        return status

    def clean_moderator(self):
        openid = self.cleaned_data['moderator']
        user, __ = update_or_create_user_from_oauth(openid)
        if user is None:
            self.add_error(
                'moderator', _('User given as moderator not found.'))
        elif not self.instance.can_be_moderated_by(user):
            self.add_error(
                'moderator', _('User cannot modify this moderation.'))
        self.cleaned_data['moderator'] = user
        return user

    def save(self, commit=True, *args, **kwargs):
        moderator = self.cleaned_data['moderator']
        moderation_text = self.cleaned_data['moderation_text']
        status = self.cleaned_data['status']
        instance = super(BaseClickPackageModerationStatusForm, self).save(
            commit=False, *args, **kwargs)
        instance.moderate(moderator, status, moderation_text)
        return instance


class ClickPackageModerationStatusForm(BaseClickPackageModerationStatusForm):

    class Meta:
        model = ClickPackageModeration
        fields = ('status', 'moderator', 'moderation_text')


class ClickPackageReviewModerationStatusForm(
        BaseClickPackageModerationStatusForm):

    class Meta:
        model = ClickPackageReviewModeration
        fields = ('status', 'moderator', 'moderation_text')
