"""Models for storing and retrieving reviews of Click packages."""
import pytz

from datetime import datetime

from django.contrib.auth.models import User
from django.db import models

from core.fields import (
    DebVersionField,
    IntegerRangeField,
)
from core.models import (
    BaseModeration,
    BaseModerationFlag,
    ModerationManager,
    ObjectWithReviewStats,
)


class ClickPackage(ObjectWithReviewStats):
    """The click application for which reviews are stored."""

    package_name = models.CharField(max_length=255, unique=True)

    def __unicode__(self):
        return u'Click Package {0}'.format(self.package_name)

    def reviews_for_object(self):
        """Returns the reviews for this package for ObjectWithReviewStats."""
        return self.clickpackagereview_set.published().filter(
            rating__gt=0, rating__lte=5)

    class Meta:
        app_label = 'clickreviews'


class ClickPackageReviewManager(models.Manager):

    def published(self):
        return self.get_query_set().filter(
            hide=False, date_deleted__isnull=True)


class ClickPackageReview(models.Model):
    """A review of a click package."""

    click_package = models.ForeignKey(ClickPackage)

    version = DebVersionField(max_length=100, db_index=True)
    arch_tag = models.CharField(max_length=16)
    reviewer = models.ForeignKey(User)
    date_created = models.DateTimeField(
        default=lambda: datetime.now(pytz.utc),
        editable=False)
    rating = IntegerRangeField(min_value=1, max_value=5)
    summary = models.CharField(max_length=80)
    review_text = models.CharField(max_length=5000)
    language = models.SlugField(max_length=10)
    hide = models.BooleanField(
        'Hide review from the UI', db_index=True, default=False)
    date_deleted = models.DateTimeField(blank=True, null=True)

    # The following get updated when a usefulness vote is added
    usefulness_total = models.IntegerField(default=0)
    usefulness_favorable = models.IntegerField(default=0)
    usefulness_percentage = models.IntegerField(default=0)
    usefulness_wilson_rating = models.FloatField(default=0.0)

    objects = ClickPackageReviewManager()

    # Helpers for piston resource.
    def package_name(self):
        return self.click_package.package_name

    def reviewer_username(self):
        # Use .all() so that `prefetch_related` have effect.
        openid_objs = list(self.reviewer.useropenid_set.all())
        if openid_objs:
            openid_obj = openid_objs[-1]
            return openid_obj.claimed_id.split('/')[-1]
        # Production queries show this won't happen, use as safeguard.
        return self.reviewer.username

    def reviewer_displayname(self):
        return self.reviewer.get_full_name()

    def is_awaiting_moderation(self):
        """Return true if the review has any pending moderations."""
        return self.moderation_set.pending().exists()

    def delete(self):
        """Delete a review submitted by the user."""
        self.hide = True
        self.date_deleted = datetime.now(pytz.utc)
        self.save()
        self.click_package.update_stats()


# Q) So, what's this tonguetwister all about?  A) In click packages,
# not only reviews can be flagged for moderation but also the click
# packages themselves, meaning we have two very similar named models
# that track their flagging and moderation.

class ClickPackageModeration(BaseModeration):
    """Track moderation of Click Packages."""

    package = models.ForeignKey(
        ClickPackage, related_name='moderation_set')

    objects = ModerationManager()

    def get_moderated_object(self):
        return self.package

    def on_moderate(self, review):
        # Do nothing.  SCA will take care of unpublishing the
        # moderated packages and that's all we need.
        pass

    def can_be_moderated_by(self, user):
        # XXX: Not checking for any specific perms as moderation
        # endpoints use DelegatedSSO.
        flagged_by_user = self.flag_set.filter(
            user=user).exists()
        return not flagged_by_user

    # XXX: piston breaks `prefetch_related` on related managers; this
    # hack is to make piston's `Emitter.construct` use our method so
    # that `prefetch_related` is honored.  Never use this in your
    # code, use `flag_set` instead, leave this to piston only.
    def flags(self):
        return self.flag_set.all()


class ClickPackageFlag(BaseModerationFlag):
    """Track moderation flags of Click Packages."""

    moderation = models.ForeignKey(
        ClickPackageModeration, related_name='flag_set')

    class Meta:
        ordering = ['-date_created']

    def get_moderation_object(self):
        return self.moderation


class ClickPackageReviewModeration(BaseModeration):
    """Track moderation of Click Package reviews."""

    review = models.ForeignKey(
        ClickPackageReview, related_name='moderation_set')

    objects = ModerationManager()

    def get_moderated_object(self):
        return self.review

    def on_moderate(self, review):
        # Hide or unhide the related review as appropriate.
        should_hide = (self.status == self.REJECTED_STATUS)
        if review.hide != should_hide:
            review.hide = should_hide
            review.save()
            review.click_package.update_stats()

    def can_be_moderated_by(self, user):
        is_review_owner = self.review.reviewer.pk == user.pk
        flagged_by_user = self.flag_set.filter(
            user=user).exists()
        return not is_review_owner and not flagged_by_user

    # XXX: piston workaround. See `ClickPackageModeration.flags` for
    # details.
    def flags(self):
        return self.flag_set.all()


class ClickPackageReviewFlag(BaseModerationFlag):
    """Track moderation flags for Click Package reviews."""

    moderation = models.ForeignKey(
        ClickPackageReviewModeration, related_name='flag_set')

    class Meta:
        ordering = ['-date_created']

    def get_moderation_object(self):
        return self.moderation
