"""TestCases with ClickPackage reviews related object factory."""
import pytz

from datetime import datetime

from django.test import TestCase

from clickreviews.models import (
    ClickPackage,
    ClickPackageReview,
    ClickPackageFlag,
    ClickPackageModeration,
    ClickPackageReviewFlag,
    ClickPackageReviewModeration,
)
from core.tests.factory import BaseObjectFactory


__all__ = [
    'ClickReviewsObjectFactory',
    'TestCaseWithFactory',
]


class ClickReviewsObjectFactory(BaseObjectFactory):
    """A factory for creating model objects for testing."""

    def makeVersion(self):
        return '{0}.0'.format(self.getUniqueInteger())

    def makeClickPackage(self, package_name=None,
                         ratings_total=2, ratings_average=3.5,
                         histogram="[0, 0, 1, 1, 0]",
                         date_ratings_updated=None):
        if package_name is None:
            package_name = self.getUniqueString(prefix='package')
        if date_ratings_updated is None:
            date_ratings_updated = datetime.now(pytz.utc)

        return ClickPackage.objects.create(
            package_name=package_name,
            ratings_total=ratings_total,
            ratings_average=unicode(ratings_average),
            histogram=histogram,
            date_ratings_updated=date_ratings_updated)

    def makeClickPackageReview(
            self, reviewer=None, click_package=None, version=None, rating=5,
            summary=None, review_text=None, date_created=None, language=None,
            hide=False, num_flags=0, arch_tag='i386',
            usefulness_wilson_rating=None, deleted=False):
        if reviewer is None:
            reviewer = self.makeUser()
        if click_package is None:
            click_package = self.makeClickPackage()
        if version is None:
            version = self.makeVersion()
        if summary is None:
            summary = self.getUniqueString(prefix='summary')
        if review_text is None:
            review_text = self.getUniqueString(prefix='review_text')
        if date_created is None:
            date_created = datetime.now(pytz.utc)
        if language is None:
            language = self.getUniqueString(prefix='lang')
        if deleted:
            hide = True
            date_deleted = datetime.now(pytz.utc)
        else:
            date_deleted = None
        if usefulness_wilson_rating is None:
            usefulness_wilson_rating = 0.0

        review = ClickPackageReview.objects.create(
            reviewer=reviewer, click_package=click_package,
            version=version, rating=rating, summary=summary,
            review_text=review_text, language=language,
            hide=hide, date_created=date_created, arch_tag=arch_tag,
            usefulness_wilson_rating=usefulness_wilson_rating,
            date_deleted=date_deleted)

        review.click_package.update_stats()

        return review

    def makeClickPackageModeration(self, package=None, status=None):
        if package is None:
            package = self.makeClickPackage()
        if status is None:
            status = ClickPackageModeration.PENDING_STATUS

        return ClickPackageModeration.objects.create(
            package=package, status=status)

    def makeClickPackageFlag(
            self, summary='Summary', description='Description',
            moderation=None, package=None, user=None):
        if moderation is None:
            moderation = self.makeClickPackageModeration(package=package)
        if user is None:
            user = self.makeUser()

        return ClickPackageFlag.objects.create(
            summary='Summary', description='Description',
            moderation=moderation, user=user)

    def makeClickPackageReviewModeration(self, review=None, status=None):
        if review is None:
            review = self.makeClickPackageReview()
        if status is None:
            status = ClickPackageReviewModeration.PENDING_STATUS

        return ClickPackageReviewModeration.objects.create(
            review=review, status=status)

    def makeClickPackageReviewFlag(
            self, summary='Summary', description='Description',
            moderation=None, review=None, user=None):
        if moderation is None:
            moderation = self.makeClickPackageReviewModeration(review=review)
        if user is None:
            user = self.makeUser()

        return ClickPackageReviewFlag.objects.create(
            summary='Summary', description='Description',
            moderation=moderation, user=user)


class TestCaseWithFactory(TestCase):

    def setUp(self):
        super(TestCaseWithFactory, self).setUp()
        self.factory = ClickReviewsObjectFactory()
