import mock

from clickreviews.forms import (
    ClickPackageFlagForm,
    ClickPackageModerationStatusForm,
    ClickPackageReviewFlagForm,
    ClickPackageReviewForm,
    ClickPackageReviewModerationStatusForm,
    ClickPackageReviewUpdateForm,
)
from clickreviews.models import (
    ClickPackage,
    ClickPackageReview,
    ClickPackageReviewModeration,
)
from clickreviews.tests.factory import TestCaseWithFactory
from core.models import BaseModeration
from core.tests.helpers import SSOTestCaseMixin


class ClickPackageReviewFormTestCase(TestCaseWithFactory):

    def setUp(self):
        super(ClickPackageReviewFormTestCase, self).setUp()

        # Ensure we don't hit the network for our tests.
        click_verify_fn = ('clickreviews.forms.'
                           'verify_click_package_published')
        patcher = mock.patch(click_verify_fn)
        self.addCleanup(patcher.stop)
        self.mock_click_verify = patcher.start()
        self.mock_click_verify.return_value = True

    def make_form_data(self, package_name='com.example.myname.myapp',
                       arch_tag='i386'):
        return {
            'package_name': package_name,
            'summary': 'Excellent',
            'version': '0.1.5',
            'review_text': 'Best app eva',
            'rating': '5',
            'language': 'en',
            'arch_tag': arch_tag,
        }, ClickPackageReview(reviewer=self.factory.makeUser())

    def test_valid(self):
        valid_data, instance = self.make_form_data()

        form = ClickPackageReviewForm(data=valid_data, instance=instance)

        self.assertEqual(True, form.is_valid())

    def test_package_not_published(self):
        self.mock_click_verify.return_value = False
        data, instance = self.make_form_data()

        form = ClickPackageReviewForm(data=data, instance=instance)

        self.assertEqual(False, form.is_valid())
        self.assertEqual(['package_name'], form.errors.keys())

    def test_saves_click_package(self):
        valid_data, instance = self.make_form_data(
            package_name='com.example.me.app')

        form = ClickPackageReviewForm(data=valid_data, instance=instance)
        click_package_review = form.save()

        self.assertEqual('com.example.me.app',
                         click_package_review.click_package.package_name)

    def test_save_second_review_fails(self):
        valid_data, instance = self.make_form_data()
        click_package = self.factory.makeClickPackage(
            package_name=valid_data['package_name'])
        self.factory.makeClickPackageReview(
            reviewer=instance.reviewer,
            click_package=click_package)

        form = ClickPackageReviewForm(data=valid_data, instance=instance)

        self.assertEqual(False, form.is_valid())

    def test_save_second_review_succeeds_if_first_deleted(self):
        valid_data, instance = self.make_form_data()
        click_package = self.factory.makeClickPackage(
            package_name=valid_data['package_name'])
        self.factory.makeClickPackageReview(
            reviewer=instance.reviewer,
            click_package=click_package,
            deleted=True)

        form = ClickPackageReviewForm(data=valid_data, instance=instance)

        self.assertEqual(True, form.is_valid())


class ClickPackageReviewUpdateFormTestCase(TestCaseWithFactory):

    def setUp(self):
        super(ClickPackageReviewUpdateFormTestCase, self).setUp()
        self.valid_data = {
            'summary': 'New summary',
            'review_text': 'New review body',
            'rating': 5}

    def test_valid(self):
        review = self.factory.makeClickPackageReview(rating=3)
        form = ClickPackageReviewUpdateForm(self.valid_data, instance=review)
        self.assertTrue(form.is_valid())
        review = form.save()
        for key, value in self.valid_data.items():
            self.assertEqual(getattr(review, key), self.valid_data[key])

    def assert_field_required(self, field_name):
        del self.valid_data[field_name]
        review = self.factory.makeClickPackageReview()
        form = ClickPackageReviewUpdateForm(self.valid_data, instance=review)
        self.assertFalse(form.is_valid())
        self.assertIn(field_name, form.errors)
        self.assertIn('This field is required.', form.errors[field_name])

    def test_summary_required(self):
        self.assert_field_required('summary')

    def test_review_text_required(self):
        self.assert_field_required('review_text')

    def test_rating_required(self):
        self.assert_field_required('rating')

    def test_is_awaiting_moderation(self):
        review = self.factory.makeClickPackageReview()
        self.factory.makeClickPackageReviewModeration(review=review)
        self.assertTrue(review.is_awaiting_moderation())
        form = ClickPackageReviewUpdateForm(self.valid_data, instance=review)
        self.assertFalse(form.is_valid())
        self.assertIn('can_modify', form.errors)
        self.assertIn("This review cannot be edited (pending moderation).",
                      form.errors['can_modify'])


class BaseClickPackageFlagFormTestCase(TestCaseWithFactory):

    form = None

    def __init__(self, *args, **kwargs):
        super(BaseClickPackageFlagFormTestCase, self).__init__(
            *args, **kwargs)
        # skip this base class to be run as a TestCase
        if self.__class__ == BaseClickPackageFlagFormTestCase:
            self.run = lambda self, *args, **kwargs: None

    def setUp(self):
        super(BaseClickPackageFlagFormTestCase, self).setUp()

        # Ensure we don't hit the network for our tests.
        click_verify_fn = ('clickreviews.forms.'
                           'verify_click_package_published')
        patcher = mock.patch(click_verify_fn)
        self.addCleanup(patcher.stop)
        self.mock_click_verify = patcher.start()
        self.mock_click_verify.return_value = True

    def make_form_data(self, *args, **kwargs):
        raise NotImplementedError

    def make_moderated_object(self):
        raise NotImplementedError

    def get_moderated_object_from_data(self, data):
        raise NotImplementedError

    def make_moderation_object(self):
        raise NotImplementedError

    def check_cleaned_data(self, cleaned_data, form_data):
        raise NotImplementedError

    def test_with_non_existing_moderated_object_and_no_moderation(self):
        data = self.make_form_data()
        form = self.form(data)
        self.assertTrue(form.is_valid())

        self.check_cleaned_data(form.cleaned_data, data)

        # Must create a moderation
        flag = form.save()
        moderation = flag.moderation
        moderated_object = self.get_moderated_object_from_data(data)
        self.assertEqual(moderation.status, moderation.PENDING_STATUS)
        # Moderated object must be properly set
        self.assertEqual(moderation.get_moderated_object(), moderated_object)

    def test_valid_with_moderated_object_and_no_moderation(self):
        moderated_object = self.make_moderated_object()
        data = self.make_form_data(moderated_object=moderated_object)
        form = self.form(data)
        self.assertTrue(form.is_valid())

        self.check_cleaned_data(form.cleaned_data, data)

        # Must create a moderation
        flag = form.save()
        moderation = flag.moderation
        self.assertEqual(moderation.status, moderation.PENDING_STATUS)
        self.assertEqual(moderation.get_moderated_object(), moderated_object)

    def test_no_moderated_object_and_moderation(self):
        # This is a case that will never happen, a moderation always
        # points to the moderated object.
        pass

    def test_valid_with_moderated_object_and_moderation(self):
        moderation = self.make_moderation_object()
        moderated_object = moderation.get_moderated_object()
        data = self.make_form_data(moderated_object=moderated_object)
        form = self.form(data)
        self.assertTrue(form.is_valid())

        self.check_cleaned_data(form.cleaned_data, data)

        # Must re-use the moderation
        flag = form.save()
        self.assertEqual(flag.moderation, moderation)

    def test_valid_with_moderated_object_and_moderation_no_pending(self):
        moderation = self.make_moderation_object()
        moderation.status = moderation.REJECTED_STATUS
        moderation.save()

        moderated_object = moderation.get_moderated_object()
        data = self.make_form_data(moderated_object=moderated_object)
        form = self.form(data)
        self.assertTrue(form.is_valid())

        self.check_cleaned_data(form.cleaned_data, data)

        # Must create a moderation object with pending status
        flag = form.save()
        self.assertNotEqual(flag.moderation, moderation)
        self.assertEqual(
            flag.moderation.status, flag.moderation.PENDING_STATUS)

    def test_description_max_length(self):
        length = self.form.DESCRIPTION_MAX_LENGTH + 1
        description = 'a' * length
        data = self.make_form_data(description=description)
        form = self.form(data)
        self.assertFalse(form.is_valid())
        error = 'Ensure this value has at most 4098 characters (it has 4099).'
        self.assertIn(error, form.errors['description'])


class ClickPackageFlagFormTestCase(BaseClickPackageFlagFormTestCase):

    form = ClickPackageFlagForm

    def make_form_data(self, moderated_object=None, summary='Summary',
                       description='Description', user=None):
        if moderated_object is None:
            package_name = 'org.user.somepackage'
        else:
            package_name = moderated_object.package_name
        if user is None:
            user = self.factory.makeUser()
        return {
            'package_name': package_name,
            'summary': summary,
            'description': description,
            'user': user.pk,
        }

    def make_moderated_object(self):
        return self.factory.makeClickPackage()

    def get_moderated_object_from_data(self, data):
        return ClickPackage.objects.get(package_name=data['package_name'])

    def make_moderation_object(self):
        return self.factory.makeClickPackageModeration()

    def check_cleaned_data(self, cleaned_data, form_data):
        package = self.get_moderated_object_from_data(form_data)
        self.assertEqual(cleaned_data['package'], package)
        self.assertEqual(
            cleaned_data['package_name'], form_data['package_name'])
        self.assertEqual(cleaned_data['summary'], form_data['summary'])
        self.assertEqual(cleaned_data['description'], form_data['description'])
        self.assertEqual(cleaned_data['user'].pk, form_data['user'])

    def test_valid_clean_package_name_with_no_existing_package(self):
        data = self.make_form_data()
        package_exists = ClickPackage.objects.filter(
            package_name=data['package_name']).exists()
        # Ensure package was not created automatically
        self.assertFalse(package_exists)
        form = self.form(data)
        form.cleaned_data = {'package_name': data['package_name']}
        package_name = form.clean_package_name()
        package = ClickPackage.objects.get(package_name=package_name)
        self.assertEqual(package_name, data['package_name'])
        self.assertEqual(form.cleaned_data['package'], package)

    def test_valid_clean_package_name_with_existing_package(self):
        package = self.factory.makeClickPackage()
        data = self.make_form_data(moderated_object=package)
        form = self.form(data)
        form.cleaned_data = {'package_name': data['package_name']}
        package_name = form.clean_package_name()
        self.assertEqual(
            form.cleaned_data['package_name'], package_name)
        self.assertEqual(form.cleaned_data['package'], package)

    def test_invalid_clean_package_name(self):
        self.mock_click_verify.return_value = False
        data = self.make_form_data()
        form = self.form(data)
        self.assertFalse(form.is_valid())
        error = "Click package 'org.user.somepackage' is not published."
        self.assertIn(error, form.errors['package_name'])

    def test_invalid_flagging_more_than_once(self):
        package = self.factory.makeClickPackage()
        user = self.factory.makeUser()
        data = self.make_form_data(moderated_object=package, user=user)
        form = self.form(data)
        self.assertTrue(form.is_valid())
        form.save()
        form = self.form(data)
        self.assertFalse(form.is_valid())
        error = 'User cannot flag same package twice.'
        self.assertIn(error, form.errors['package_name'])


class ClickPackageReviewFlagFormTestCase(BaseClickPackageFlagFormTestCase):

    form = ClickPackageReviewFlagForm

    def make_form_data(self, moderated_object=None, summary='Summary',
                       description='Description', user=None):
        if moderated_object is None:
            review = self.factory.makeClickPackageReview()
            review_pk = review.pk
        else:
            review_pk = moderated_object.pk
        if user is None:
            user = self.factory.makeUser()
        return {
            'review': review_pk,
            'summary': summary,
            'description': description,
            'user': user.pk,
        }

    def make_moderated_object(self):
        return self.factory.makeClickPackageReview()

    def get_moderated_object_from_data(self, data):
        return ClickPackageReview.objects.get(pk=data['review'])

    def make_moderation_object(self):
        return self.factory.makeClickPackageReviewModeration()

    def check_cleaned_data(self, cleaned_data, form_data):
        review = self.get_moderated_object_from_data(form_data)
        self.assertEqual(cleaned_data['review'], review)
        self.assertEqual(cleaned_data['summary'], form_data['summary'])
        self.assertEqual(cleaned_data['description'], form_data['description'])
        self.assertEqual(cleaned_data['user'].pk, form_data['user'])

    def test_invalid_review_pk(self):
        data = self.make_form_data()
        data['review'] = -1
        form = ClickPackageReviewFlagForm(data)
        self.assertFalse(form.is_valid())
        error = ('Select a valid choice. '
                 'That choice is not one of the available choices.')
        self.assertIn(error, form.errors['review'])

    def test_invalid_flagging_more_than_once(self):
        review = self.factory.makeClickPackageReview()
        user = self.factory.makeUser()
        data = self.make_form_data(moderated_object=review, user=user)
        form = self.form(data)
        self.assertTrue(form.is_valid())
        form.save()
        form = self.form(data)
        self.assertFalse(form.is_valid())
        error = 'User cannot flag same review twice.'
        self.assertIn(error, form.errors['review'])


class BaseClickPackageModerationStatusForm(
        SSOTestCaseMixin, TestCaseWithFactory):

    form = None

    def __init__(self, *args, **kwargs):
        super(BaseClickPackageModerationStatusForm, self).__init__(
            *args, **kwargs)
        if self.__class__ == BaseClickPackageModerationStatusForm:
            self.run = lambda self, *args, **kwargs: None

    def make_moderation_object(self, status=None):
        raise NotImplementedError

    def setUp(self):
        super(BaseClickPackageModerationStatusForm, self).setUp()
        self.user = self.factory.makeUser()
        self.openid = self.user.useropenid_set.get().claimed_id.split('/')[-1]
        self.patch_sso()

    def test_set_not_pending_to_approved(self):
        moderation = self.make_moderation_object(
            status=BaseModeration.REJECTED_STATUS)
        data = {'status': BaseModeration.APPROVED_STATUS,
                'moderator': self.openid}
        form = self.form(data, instance=moderation)
        self.assertFalse(form.is_valid())
        self.assertIn('Moderation must be pending to change status.',
                      form.errors['status'])

    def test_set_not_pending_to_rejected(self):
        moderation = self.make_moderation_object(
            status=BaseModeration.APPROVED_STATUS)
        data = {'status': BaseModeration.REJECTED_STATUS,
                'moderator': self.openid}
        form = self.form(data, instance=moderation)
        self.assertFalse(form.is_valid())
        self.assertIn('Moderation must be pending to change status.',
                      form.errors['status'])

    def test_approved_set_to_approved(self):
        moderation = self.make_moderation_object(
            status=BaseModeration.APPROVED_STATUS)
        data = {'status': BaseModeration.APPROVED_STATUS,
                'moderator': self.openid}
        form = self.form(data, instance=moderation)
        self.assertFalse(form.is_valid())
        self.assertIn('Moderation must be pending to change status.',
                      form.errors['status'])

    def test_pending_set_to_pending(self):
        moderation = self.make_moderation_object()
        data = {'status': BaseModeration.PENDING_STATUS,
                'moderator': self.openid}
        form = self.form(data, instance=moderation)
        self.assertFalse(form.is_valid())
        self.assertIn('Moderation status cannot be set to pending.',
                      form.errors['status'])

    def test_rejected_set_to_rejected(self):
        moderation = self.make_moderation_object(
            status=BaseModeration.REJECTED_STATUS)
        data = {'status': BaseModeration.REJECTED_STATUS,
                'moderator': self.openid}
        form = self.form(data, instance=moderation)
        self.assertFalse(form.is_valid())
        self.assertIn('Moderation must be pending to change status.',
                      form.errors['status'])

    def test_moderation_text_required(self):
        moderation = self.make_moderation_object()
        data = {'status': BaseModeration.REJECTED_STATUS,
                'moderator': self.openid}
        form = self.form(data, instance=moderation)
        self.assertFalse(form.is_valid())
        self.assertIn(
            'This field is required.', form.errors['moderation_text'])

    def test_moderator_required(self):
        moderation = self.make_moderation_object()
        data = {'status': BaseModeration.REJECTED_STATUS,
                'moderation_text': 'text'}
        form = self.form(data, instance=moderation)
        self.assertFalse(form.is_valid())
        self.assertIn('This field is required.', form.errors['moderator'])

    def test_invalid_moderator_given(self):
        moderation = self.make_moderation_object()
        self.mock_get_data.return_value = None
        data = {'status': BaseModeration.REJECTED_STATUS,
                'moderator': 'inexistent',
                'moderation_text': 'text'}
        form = self.form(data, instance=moderation)
        self.assertFalse(form.is_valid())
        self.assertIn(
            'User given as moderator not found.', form.errors['moderator'])

    def test_valid_set_approved(self):
        moderation = self.make_moderation_object()
        data = {'status': BaseModeration.APPROVED_STATUS,
                'moderator': self.openid,
                'moderation_text': 'text'}
        form = self.form(data, instance=moderation)
        self.assertTrue(form.is_valid())

    def test_valid_set_rejected(self):
        moderation = self.make_moderation_object()
        data = {'status': BaseModeration.REJECTED_STATUS,
                'moderator': self.openid,
                'moderation_text': 'text'}
        form = self.form(data, instance=moderation)
        self.assertTrue(form.is_valid())


class ClickPackageModerationStatusForm(BaseClickPackageModerationStatusForm):

    form = ClickPackageModerationStatusForm

    def make_moderation_object(self, status=None):
        return self.factory.makeClickPackageModeration(status=status)

    def test_user_that_flagged_the_package_cannot_moderate_it(self):
        moderation = self.make_moderation_object()
        self.factory.makeClickPackageFlag(
            user=self.user, moderation=moderation)
        data = {'status': BaseModeration.REJECTED_STATUS,
                'moderator': self.openid,
                'moderation_text': 'text'}
        form = self.form(data, instance=moderation)
        self.assertFalse(form.is_valid())
        self.assertIn(
            'User cannot modify this moderation.', form.errors['moderator'])


class ClickPackageReviewModerationStatusForm(
        BaseClickPackageModerationStatusForm):

    form = ClickPackageReviewModerationStatusForm

    def make_moderation_object(self, status=None):
        return self.factory.makeClickPackageReviewModeration(status=status)

    def test_save_rejected_hides_review(self):
        moderation = self.make_moderation_object()
        data = {'status': BaseModeration.REJECTED_STATUS,
                'moderator': self.openid,
                'moderation_text': 'text'}
        form = self.form(data, instance=moderation)
        self.assertTrue(form.is_valid())
        form.save()
        obj = ClickPackageReviewModeration.objects.get(pk=moderation.pk)
        self.assertTrue(obj.review.hide)

    def test_save_approved_doesnt_hide_review(self):
        moderation = self.make_moderation_object()
        data = {'status': BaseModeration.APPROVED_STATUS,
                'moderator': self.openid,
                'moderation_text': 'text'}
        form = self.form(data, instance=moderation)
        self.assertTrue(form.is_valid())
        form.save()
        obj = ClickPackageReviewModeration.objects.get(pk=moderation.pk)
        self.assertFalse(obj.review.hide)

    def test_user_that_flagged_the_review_cannot_moderate_it(self):
        moderation = self.make_moderation_object()
        self.factory.makeClickPackageReviewFlag(
            user=self.user, moderation=moderation)
        data = {'status': BaseModeration.REJECTED_STATUS,
                'moderator': self.openid,
                'moderation_text': 'text'}
        form = self.form(data, instance=moderation)
        self.assertFalse(form.is_valid())
        self.assertIn(
            'User cannot modify this moderation.', form.errors['moderator'])

    def test_review_owner_cannot_moderate_its_flags(self):
        moderation = self.make_moderation_object()
        review = moderation.review
        review.reviewer = self.user
        review.save()
        self.factory.makeClickPackageReviewFlag(
            user=self.user, moderation=moderation)
        data = {'status': BaseModeration.REJECTED_STATUS,
                'moderator': self.openid,
                'moderation_text': 'text'}
        form = self.form(data, instance=moderation)
        self.assertFalse(form.is_valid())
        self.assertIn(
            'User cannot modify this moderation.', form.errors['moderator'])
