import httplib
import hmac
import gzip
import json
import mock
import pytz
from datetime import datetime
from hashlib import sha256
from io import BytesIO

from django.conf import settings
from django.core.cache import cache
from django.core.serializers.json import DateTimeAwareJSONEncoder
from django.core.urlresolvers import reverse

from clickreviews.api.urls import sca_delegated_sso_auth
from clickreviews.models import (
    ClickPackage,
    ClickPackageFlag,
    ClickPackageReview,
    ClickPackageReviewFlag,
    ClickPackageReviewModeration,
)
from clickreviews.tests.factory import TestCaseWithFactory
from core.models import BaseModeration
from core.tests.helpers import SSOTestCaseMixin, assert_no_extra_queries_after


class ClickReviewsHandlerTestCase(TestCaseWithFactory):

    def setUp(self):
        super(ClickReviewsHandlerTestCase, self).setUp()

        # Ensure we don't hit the network for our tests.
        is_authed_fn = ('core.api.auth.SSOOAuthAuthentication.'
                        'is_authenticated')
        patcher = mock.patch(
            is_authed_fn, new_callable=mock.Mock, return_value=True)
        self.addCleanup(patcher.stop)
        self.mock_is_authenticated = patcher.start()

        click_verify_fn = ('clickreviews.forms.'
                           'verify_click_package_published')
        patcher = mock.patch(click_verify_fn)
        self.addCleanup(patcher.stop)
        self.mock_click_verify = patcher.start()
        self.mock_click_verify.return_value = True

    def make_data(self, package_name='com.ubuntu.dev.uname.pkgid',
                  rating=5):
        return {
            'package_name': package_name,
            'summary': 'Excellent',
            'version': '0.1.2',
            'review_text': 'Plain brilliant piece of software',
            'rating': rating,
            'language': 'en',
            'arch_tag': 'i386',
        }

    def _post_new_review(self, data, user=None, verify_result=True):
        # Post a review as an authenticated user.
        url = reverse('click-api-reviews')
        if user is None:
            user = self.factory.makeUser()
        self.client.login(username=user.username, password='test')

        response = self.client.post(
            url, data=data, content_type='application/json')
        return response

    def test_create(self):
        data = self.make_data(package_name='com.example.me.myapp',
                              rating=5)

        response = self._post_new_review(json.dumps(data))

        # Piston uses 200, rather than 201.
        self.assertEqual(httplib.OK, response.status_code)
        self.assertEqual('application/json; charset=utf-8',
                         response['content-type'])
        reviews = ClickPackageReview.objects.filter(
            click_package__package_name='com.example.me.myapp')
        self.assertEqual(1, reviews.count())
        review = reviews[0]
        response_dict = json.loads(response.content)
        self.assertEqual({
            'id': review.id,
            'package_name': 'com.example.me.myapp',
            'language': data['language'],
            'date_created': DateTimeAwareJSONEncoder().encode(
                review.date_created).strip('"'),
            'rating': 5,
            'reviewer_username': review.reviewer_username(),
            'reviewer_displayname': review.reviewer.get_full_name(),
            'summary': 'Excellent',
            'review_text': 'Plain brilliant piece of software',
            'hide': False,
            'date_deleted': None,
            'version': '0.1.2',
            'usefulness_total': 0,
            'usefulness_favorable': 0,
        }, response_dict)

    def test_invalid_rating(self):
        data = self.make_data(rating=6)

        response = self._post_new_review(json.dumps(data))

        self.assertEqual(httplib.BAD_REQUEST, response.status_code)

    def test_missing_args(self):
        # If required args are not included it is a bad request.
        valid_data = self.make_data()
        for key in valid_data:
            data = valid_data.copy()
            del data[key]

            response = self._post_new_review(json.dumps(data))

            self.assertEqual(httplib.BAD_REQUEST, response.status_code)

    def test_stats_updated(self):
        # Posting a review updates the associated stats.
        data = self.make_data(package_name='com.example.me.myapp',
                              rating=5)
        click_package = ClickPackage.objects.create(
            package_name='com.example.me.myapp')

        self._post_new_review(json.dumps(data))

        click_package = ClickPackage.objects.get(id=click_package.id)
        self.assertEqual(1, click_package.ratings_total)
        self.assertEqual(5, click_package.ratings_average)

    def _get_reviews(self, data=None):
        url = reverse('click-api-reviews')
        kwargs = {}
        if data is not None:
            kwargs['data'] = data
        self.mock_is_authenticated.return_value = False
        return self.client.get(url, **kwargs)

    def test_get_multiple_reviews_for_package(self):
        package = self.factory.makeClickPackage()
        for _ in range(5):
            self.factory.makeClickPackageReview(click_package=package)

        json_result = self._get_reviews(data={
            'package_name': package.package_name,
        })

        self.assertEqual(200, json_result.status_code)
        result = json.loads(json_result.content)
        self.assertEqual(5, len(result))

    def test_get_reviews_contains_correct_data(self):
        package = self.factory.makeClickPackage(
            package_name="com.example.me.myapp")
        review = self.factory.makeClickPackageReview(
            click_package=package, language='en',
            rating=5, summary='Best app', review_text='Unlimited fun!',
            version='0.1.1', date_created=datetime(
                2014, 1, 15, 11, 13, 55, 123456, tzinfo=pytz.UTC))

        json_result = self._get_reviews(data={
            'package_name': package.package_name,
        })
        self.assertEqual(200, json_result.status_code)
        result = json.loads(json_result.content)
        self.assertEqual(1, len(result))
        result_review = result[0]
        DateTimeAwareJSONEncoder().encode(
            review.date_created).strip('"')
        self.assertEqual({
            'id': review.id,
            'package_name': 'com.example.me.myapp',
            'language': 'en',
            'date_created': '2014-01-15T11:13:55.123Z',
            'rating': 5,
            'reviewer_username': review.reviewer_username(),
            'reviewer_displayname': review.reviewer_displayname(),
            'summary': 'Best app',
            'review_text': 'Unlimited fun!',
            'hide': False,
            'date_deleted': None,
            'version': '0.1.1',
            'usefulness_total': 0,
            'usefulness_favorable': 0,
        }, result_review)

    def test_must_filter_by_packagename(self):
        result = self._get_reviews()

        self.assertEqual(400, result.status_code)
        self.assertEqual(['package_name'],
                         json.loads(result.content)['errors'].keys())

    def test_doesnt_include_deleted(self):
        package = self.factory.makeClickPackage()
        self.factory.makeClickPackageReview(
            click_package=package,
            deleted=True)

        json_result = self._get_reviews(data={
            'package_name': package.package_name,
        })
        self.assertEqual(200, json_result.status_code)
        result = json.loads(json_result.content)
        self.assertEqual(0, len(result))

    def test_doesnt_include_hidden(self):
        package = self.factory.makeClickPackage()
        self.factory.makeClickPackageReview(click_package=package, hide=True)

        json_result = self._get_reviews(data={
            'package_name': package.package_name,
        })
        self.assertEqual(200, json_result.status_code)
        result = json.loads(json_result.content)
        self.assertEqual(0, len(result))

    def test_ordered_by_date_created(self):
        package = self.factory.makeClickPackage()
        dates = [
            datetime(2013, 3, 15, 0, 0, 0, 0, tzinfo=pytz.UTC),  # 0: 3rd
            datetime(2013, 6, 15, 0, 0, 0, 0, tzinfo=pytz.UTC),  # 1: 2nd
            datetime(2013, 1, 15, 0, 0, 0, 0, tzinfo=pytz.UTC),  # 2: 4th
            datetime(2014, 1, 15, 0, 0, 0, 0, tzinfo=pytz.UTC)]  # 3: 1st
        reviews = [
            self.factory.makeClickPackageReview(
                click_package=package,
                date_created=date)
            for date in dates
        ]
        expected_review_ids = [
            reviews[3].id,
            reviews[1].id,
            reviews[0].id,
            reviews[2].id,
        ]

        json_result = self._get_reviews(data={
            'package_name': package.package_name,
        })
        self.assertEqual(200, json_result.status_code)
        results = json.loads(json_result.content)
        self.assertEqual(4, len(results))
        actual_review_ids = [result['id'] for result in results]
        self.assertEqual(actual_review_ids, expected_review_ids)

    def test_query_count(self):
        package = self.factory.makeClickPackage()
        for _ in range(5):
            self.factory.makeClickPackageReview(click_package=package)
        assert_no_extra_queries_after(
            lambda: self.factory.makeClickPackageReview(click_package=package),
            self._get_reviews, data={'package_name': package.package_name}
        )

    def _update_review(self, review, data, user=None):
        # Post a review as an authenticated user.
        url = reverse('click-api-reviews', args=(review.pk,))
        if user is None:
            user = review.reviewer
        self.client.login(username=user.username, password='test')
        response = self.client.put(
            url, data=json.dumps(data), content_type='application/json')
        return response

    def test_update_review(self):
        review = self.factory.makeClickPackageReview()
        data = {'summary': 'New summary',
                'review_text': 'Review text',
                'rating': 5}
        response = self._update_review(review, data)
        self.assertEqual(response.status_code, httplib.OK)
        review = json.loads(response.content)
        for key, value in data.items():
            self.assertEqual(review[key], data[key])

    def test_update_review_missing_data(self):
        review = self.factory.makeClickPackageReview()
        response = self._update_review(review, {})
        self.assertEqual(response.status_code, httplib.BAD_REQUEST)
        errors = json.loads(response.content)
        expected = {'errors':
                    {'rating': ['This field is required.'],
                     'review_text': ['This field is required.'],
                     'summary': ['This field is required.']}}
        self.assertEqual(errors, expected)

    def test_update_review_wrong_user(self):
        review = self.factory.makeClickPackageReview()
        user = self.factory.makeUser()
        response = self._update_review(review, {}, user)
        self.assertEqual(response.status_code, httplib.NOT_FOUND)

    def _delete_review(self, review, user=None):
        # Delete a review as an authenticated user.
        url = reverse('click-api-reviews', args=(review.pk,))
        if user is None:
            user = review.reviewer
        self.client.login(username=user.username, password='test')
        response = self.client.delete(url)
        return response

    def test_delete_review(self):
        review = self.factory.makeClickPackageReview()
        response = self._delete_review(review)
        self.assertEqual(response.status_code, httplib.OK)
        review = json.loads(response.content)

    def test_delete_review_wrong_id(self):
        review = self.factory.makeClickPackageReview()
        ClickPackageReview.objects.all().delete()
        response = self._delete_review(review)
        self.assertEqual(response.status_code, httplib.NOT_FOUND)

    def test_delete_already_deleted(self):
        review = self.factory.makeClickPackageReview()
        review.delete()
        response = self._delete_review(review)
        self.assertEqual(response.status_code, httplib.NOT_FOUND)

    def test_delete_hidden_review(self):
        review = self.factory.makeClickPackageReview(hide=True)
        response = self._delete_review(review)
        self.assertEqual(response.status_code, httplib.NOT_FOUND)

    def test_delete_review_wrong_user(self):
        review = self.factory.makeClickPackageReview()
        user = self.factory.makeUser()
        response = self._delete_review(review, user=user)
        self.assertEqual(response.status_code, httplib.NOT_FOUND)


class ClickReviewsStatsHandlerTestCase(TestCaseWithFactory):

    def setUp(self):
        super(ClickReviewsStatsHandlerTestCase, self).setUp()
        self.clear_cache()

    def make_stats_response(
            self, pk=1, package_name='org.some.package',
            histogram='[0, 0, 0, 0, 1]', ratings_total=1,
            ratings_average='5.00'):

        # XXX: To be cleaned up once support for sqlite is dropped.
        # Using sqlite does not honor the settings for the
        # ratings_average DecimalField spec, ignoring max_digits and
        # decimal_places when returning the formatted value.
        db_engine = settings.DATABASES['default']['ENGINE']
        if db_engine == 'django.db.backends.sqlite3':
            ratings_average = ratings_average.rstrip('0')
            if ratings_average.endswith('.'):
                ratings_average = ratings_average[0:-1]

        return {
            'id': pk,
            'package_name': package_name,
            'histogram': histogram,
            'ratings_total': ratings_total,
            'ratings_average': ratings_average,
        }

    def clear_cache(self):
        cache._cache.clear()
        cache._expire_info.clear()

    def _request_stats(self, data=None, headers=None, **kwargs):
        url = reverse(
            'click-api-reviews-stats', kwargs=kwargs)
        if data is None:
            data = {}
        if headers is None:
            headers = {}
        return self.client.get(url, data, **headers)

    def test_read(self):
        foo_review = self.factory.makeClickPackageReview(
            click_package=self.factory.makeClickPackage('package_foo'),
            language='de')
        bar_review = self.factory.makeClickPackageReview(
            click_package=self.factory.makeClickPackage('package_bar'),
            language='de')

        response = self._request_stats()
        reviews = json.loads(response.content)

        foo = foo_review.click_package
        bar = bar_review.click_package
        self.assertEqual(2, len(reviews))
        self.assertEqual(reviews, [
            self.make_stats_response(
                pk=foo.pk, package_name=foo.package_name),
            self.make_stats_response(
                pk=bar.pk, package_name=bar.package_name)])

    def test_read_for_package(self):
        foo_review = self.factory.makeClickPackageReview(
            click_package=self.factory.makeClickPackage('package_foo'),
            language='de')
        self.factory.makeClickPackageReview(
            click_package=self.factory.makeClickPackage('package_bar'),
            language='de')

        response = self._request_stats(package_name='package_foo')
        reviews = json.loads(response.content)

        foo = foo_review.click_package
        self.assertEqual(reviews, self.make_stats_response(
            pk=foo.pk, package_name=foo.package_name))

    def test_read_for_missing_package(self):
        response = self._request_stats(package_name='not_found')
        self.assertEqual(response.status_code, 404)

    def test_request_cached_with_extra_headers(self):
        # Requests for server status are never cached.
        response = self._request_stats()

        self.assertEqual(
            'max-age={0}'.format(settings.CACHE_REVIEW_STATS_SECONDS),
            response['cache-control'])

        # A subsequent call will not hit the view.
        handler = 'clickreviews.api.handlers.ClickReviewsStatsHandler.read'
        with mock.patch(handler) as mock_read:
            response = self._request_stats()
            self.assertFalse(mock_read.called)

    def test_cannot_request_with_get_params(self):
        # Requesting the un-cached response (by adding get params) is
        # not allowed.
        response = self._request_stats(data=dict(foo='bar'))

        self.assertEqual(400, response.status_code)

    @mock.patch.multiple(settings, ENABLE_GZIP_CONTENT=True)
    def test_is_gzipped(self):
        """Test response is compressed when conditions are met."""
        foo_review = self.factory.makeClickPackageReview(
            click_package=self.factory.makeClickPackage('package_foo'),
            language='de')
        bar_review = self.factory.makeClickPackageReview(
            click_package=self.factory.makeClickPackage('package_bar'),
            language='de')

        response = self._request_stats(
            headers={'HTTP_ACCEPT_ENCODING': 'gzip, deflate'})

        fileobj = BytesIO()
        fileobj.write(response.content)
        fileobj.seek(0)

        with gzip.GzipFile(fileobj=fileobj) as fd:
            data = fd.read()

        reviews = json.loads(data)

        foo = foo_review.click_package
        bar = bar_review.click_package
        self.assertEqual(reviews, [
            self.make_stats_response(
                pk=foo.pk, package_name=foo.package_name),
            self.make_stats_response(
                pk=bar.pk, package_name=bar.package_name)])

    def test_vary_headers(self):
        response = self._request_stats()
        self.assertEqual(200, response.status_code)
        self.assertTrue(response.has_header('vary'))
        vary = response['vary']
        self.assertEqual('Accept, Accept-Encoding', vary)


class BaseClickPackageFlagHandlerTestCase(TestCaseWithFactory):

    url_name = None

    def setUp(self):
        super(BaseClickPackageFlagHandlerTestCase, self).setUp()
        is_authenticated_method = (
            'core.api.auth.SSOOAuthAuthentication.is_authenticated')
        p = mock.patch(is_authenticated_method)
        self.mock_is_authenticated = p.start()
        self.mock_is_authenticated.return_value = True
        self.addCleanup(p.stop)
        self.user = self.factory.makeUser()
        self.client.login(username=self.user.username, password='test')
        click_verify_fn = ('clickreviews.forms.'
                           'verify_click_package_published')
        patcher = mock.patch(click_verify_fn)
        self.addCleanup(patcher.stop)
        self.mock_click_verify = patcher.start()
        self.mock_click_verify.return_value = True

    def make_data(self, summary='Summary', description='Description'):
        return {
            'summary': summary,
            'description': description}

    def do_request(
            self, data=None, headers=None, user=None, **kwargs):
        url = reverse(self.url_name, kwargs=kwargs)
        if data is None:
            data = {}
        if headers is None:
            headers = {}
        if user is None:
            user = self.factory.makeUser()
        return self.client.post(url, data, **headers)


class ClickPackageFlagHandlerTestCase(BaseClickPackageFlagHandlerTestCase):

    url_name = 'click-api-package-flags'

    def test_requires_authentication(self):
        self.mock_is_authenticated.return_value = False
        package = self.factory.makeClickPackage()
        response = self.do_request(package_name=package.package_name)
        self.assertEqual(response.status_code, httplib.UNAUTHORIZED)

    def test_create_valid_non_existing_package(self):
        assert not ClickPackageFlag.objects.exists()
        data = self.make_data()
        response = self.do_request(package_name='some_package', data=data)
        self.assertEqual(response.status_code, httplib.OK)
        self.assertEqual(ClickPackageFlag.objects.count(), 1)

    def test_create_valid_existing_package(self):
        package = self.factory.makeClickPackage(package_name='some_package')
        assert not ClickPackageFlag.objects.exists()
        data = self.make_data()
        response = self.do_request(
            package_name=package.package_name, data=data)
        self.assertEqual(response.status_code, httplib.OK)
        self.assertEqual(ClickPackageFlag.objects.count(), 1)

    def test_create_invalid_non_existing_package(self):
        self.mock_click_verify.return_value = False
        assert not ClickPackageFlag.objects.exists()
        data = self.make_data()
        response = self.do_request(package_name='some_package', data=data)
        self.assertEqual(response.status_code, httplib.BAD_REQUEST)
        response_data = json.loads(response.content)
        error = "Click package 'some_package' is not published."
        self.assertIn(error,  response_data['errors']['package_name'])
        self.assertEqual(ClickPackageFlag.objects.count(), 0)

    def test_create_summary_description_are_required(self):
        self.mock_click_verify.return_value = False
        assert not ClickPackageFlag.objects.exists()
        response = self.do_request(package_name='some_package', data={})
        self.assertEqual(response.status_code, httplib.BAD_REQUEST)
        response_data = json.loads(response.content)
        error = "This field is required."
        self.assertIn(error,  response_data['errors']['summary'])
        self.assertIn(error,  response_data['errors']['description'])
        self.assertEqual(ClickPackageFlag.objects.count(), 0)

    def test_create_description_max_length(self):
        assert not ClickPackageFlag.objects.exists()
        data = self.make_data(description='a' * 4099)
        response = self.do_request(package_name='some_package', data=data)
        self.assertEqual(response.status_code, httplib.BAD_REQUEST)
        response_data = json.loads(response.content)
        error = 'Ensure this value has at most 4098 characters (it has 4099).'
        self.assertIn(error,  response_data['errors']['description'])
        self.assertEqual(ClickPackageFlag.objects.count(), 0)

    def test_cannot_flag_twice(self):
        package = self.factory.makeClickPackage()
        assert not ClickPackageFlag.objects.exists()
        data = self.make_data()
        response = self.do_request(
            package_name=package.package_name, data=data)
        self.assertEqual(response.status_code, httplib.OK)
        self.assertEqual(ClickPackageFlag.objects.count(), 1)
        response = self.do_request(
            package_name=package.package_name, data=data)
        self.assertEqual(response.status_code, httplib.BAD_REQUEST)
        response_data = json.loads(response.content)
        error = 'User cannot flag same package twice.'
        self.assertIn(error,  response_data['errors']['package_name'])
        self.assertEqual(ClickPackageFlag.objects.count(), 1)


class ClickPackageReviewFlagHandlerTestCase(
        BaseClickPackageFlagHandlerTestCase):

    url_name = 'click-api-review-flags'

    def test_requires_authentication(self):
        self.mock_is_authenticated.return_value = False
        review = self.factory.makeClickPackageReview()
        response = self.do_request(review_id=review.pk)
        self.assertEqual(response.status_code, httplib.UNAUTHORIZED)

    def test_create_invalid_non_existing_review(self):
        assert not ClickPackageReviewFlag.objects.exists()
        data = self.make_data()
        response = self.do_request(review_id=10000, data=data)
        self.assertEqual(response.status_code, httplib.NOT_FOUND)
        self.assertEqual(ClickPackageReviewFlag.objects.count(), 0)

    def test_create_valid(self):
        review = self.factory.makeClickPackageReview()
        assert not ClickPackageReviewFlag.objects.exists()
        data = self.make_data()
        response = self.do_request(review_id=review.pk, data=data)
        self.assertEqual(response.status_code, httplib.OK)
        self.assertEqual(ClickPackageReviewFlag.objects.count(), 1)

    def test_create_summary_description_are_required(self):
        self.mock_click_verify.return_value = False
        assert not ClickPackageReviewFlag.objects.exists()
        review = self.factory.makeClickPackageReview()
        response = self.do_request(review_id=review.pk, data={})
        self.assertEqual(response.status_code, httplib.BAD_REQUEST)
        response_data = json.loads(response.content)
        error = "This field is required."
        self.assertIn(error,  response_data['errors']['summary'])
        self.assertIn(error,  response_data['errors']['description'])
        self.assertEqual(ClickPackageReviewFlag.objects.count(), 0)

    def test_create_description_max_length(self):
        assert not ClickPackageReviewFlag.objects.exists()
        data = self.make_data(description='a' * 4099)
        review = self.factory.makeClickPackageReview()
        response = self.do_request(review_id=review.pk, data=data)
        self.assertEqual(response.status_code, httplib.BAD_REQUEST)
        response_data = json.loads(response.content)
        error = 'Ensure this value has at most 4098 characters (it has 4099).'
        self.assertIn(error,  response_data['errors']['description'])
        self.assertEqual(ClickPackageReviewFlag.objects.count(), 0)

    def test_cannot_flag_twice(self):
        review = self.factory.makeClickPackageReview()
        assert not ClickPackageReviewFlag.objects.exists()
        data = self.make_data()
        response = self.do_request(review_id=review.pk, data=data)
        self.assertEqual(response.status_code, httplib.OK)
        self.assertEqual(ClickPackageReviewFlag.objects.count(), 1)
        response = self.do_request(review_id=review.pk, data=data)
        self.assertEqual(response.status_code, httplib.BAD_REQUEST)
        response_data = json.loads(response.content)
        error = 'User cannot flag same review twice.'
        self.assertIn(error,  response_data['errors']['review'])
        self.assertEqual(ClickPackageReviewFlag.objects.count(), 1)


class BaseClickPackageModerationHandlerTestCase(
        SSOTestCaseMixin, TestCaseWithFactory):

    SCA_DELEGATED_SSO_SECRET = 'secret'
    url_name = None

    def __init__(self, *args, **kwargs):
        super(BaseClickPackageModerationHandlerTestCase, self).__init__(
            *args, **kwargs)
        # skip this base class to be run as a TestCase
        if self.__class__ == BaseClickPackageModerationHandlerTestCase:
            self.run = lambda self, *args, **kwargs: None

    def make_moderation_object(self, status=None):
        raise NotImplementedError

    def make_flag(self, *args, **kwargs):
        raise NotImplementedError

    def setUp(self):
        super(BaseClickPackageModerationHandlerTestCase, self).setUp()
        self.time = 1411667041.559495
        self.user = self.factory.makeUser()
        self.openid = self.user.useropenid_set.get().claimed_id.split('/')[-1]
        self.data = {"moderation_text": "text", "moderator": self.openid}

        old_secret = sca_delegated_sso_auth.secret

        def clean_secret():
            sca_delegated_sso_auth.secret = old_secret

        sca_delegated_sso_auth.secret = self.SCA_DELEGATED_SSO_SECRET
        self.addCleanup(clean_secret)

        p = mock.patch(
            'core.api.auth.delegated_sso.time.time', return_value=self.time)
        p.start()
        self.addCleanup(p.stop)
        self.patch_sso()

    def _make_delegated_sso_signed_headers(self, url, data=None):
        if data is None:
            data = ''
        url = 'http://testserver' + url
        msg = "%s%f%s" % (url, self.time, data)
        signature = hmac.new(
            self.SCA_DELEGATED_SSO_SECRET, msg, sha256).hexdigest()
        auth = 'DelegatedSSO signature="%s", timestamp="%f"' % (
            signature, self.time)
        return {'HTTP_AUTHORIZATION': auth}

    def request_get(self, url, authenticated=True, **headers):
        if authenticated:
            headers.update(self._make_delegated_sso_signed_headers(url))
        return self.client.get(url, **headers)

    def request_patch(
            self, url, data=None, content_type='application/json',
            authenticated=True, **headers):
        if data is None:
            data = ''
        if authenticated:
            headers.update(self._make_delegated_sso_signed_headers(url, data))
        return self.client.patch(
            url, data=data, content_type=content_type, **headers)

    def test_cannot_read_list(self):
        url = reverse(self.url_name)
        response = self.request_get(url, authenticated=False)
        self.assertEqual(response.status_code, httplib.UNAUTHORIZED)

    def test_cannot_read_with_pk(self):
        moderation = self.make_moderation_object()
        url = reverse(self.url_name, kwargs={'pk': moderation.pk})
        response = self.request_get(url, authenticated=False)
        self.assertEqual(response.status_code, httplib.UNAUTHORIZED)

    def test_cannot_read_list_with_status(self):
        url = reverse(self.url_name, kwargs={'status': 'pending'})
        response = self.request_get(url, authenticated=False)
        self.assertEqual(response.status_code, httplib.UNAUTHORIZED)

    def test_cannot_read_list_with_status_and_pk(self):
        url = reverse(self.url_name, kwargs={'status': 'pending', 'pk': 1})
        response = self.request_get(url, authenticated=False)
        self.assertEqual(response.status_code, httplib.UNAUTHORIZED)

    def test_cannot_patch(self):
        url = reverse(self.url_name, kwargs={'status': 'approved', 'pk': 1})
        response = self.request_patch(
            url, json.dumps(self.data), authenticated=False)
        self.assertEqual(response.status_code, httplib.UNAUTHORIZED)

    def test_can_read_list(self):
        url = reverse(self.url_name)
        response = self.request_get(url)
        self.assertEqual(response.status_code, httplib.OK)

    def test_can_read_with_pk(self):
        moderation = self.make_moderation_object()
        url = reverse(self.url_name, kwargs={'pk': moderation.pk})
        response = self.request_get(url)
        self.assertEqual(response.status_code, httplib.OK)

    def test_can_read_list_with_status(self):
        url = reverse(self.url_name, kwargs={'status': 'pending'})
        response = self.request_get(url)
        self.assertEqual(response.status_code, httplib.OK)

    def test_can_read_list_with_status_and_pk(self):
        moderation = self.make_moderation_object()
        url = reverse(self.url_name, kwargs={
            'status': 'pending', 'pk': moderation.pk})
        response = self.request_get(url)
        self.assertEqual(response.status_code, httplib.OK)

    def test_can_patch(self):
        moderation = self.make_moderation_object()
        url = reverse(self.url_name, kwargs={
            'status': 'approved', 'pk': moderation.pk})
        response = self.request_patch(url, json.dumps(self.data))
        self.assertEqual(response.status_code, httplib.OK)

    def make_moderation_objects_with_statuses(self):
        for _ in range(2):
            self.make_moderation_object(status=BaseModeration.APPROVED_STATUS)
        for _ in range(3):
            self.make_moderation_object(status=BaseModeration.PENDING_STATUS)
        for _ in range(4):
            self.make_moderation_object(status=BaseModeration.REJECTED_STATUS)

    def test_read_for_all_objects(self):
        self.make_moderation_objects_with_statuses()
        url = reverse(self.url_name)
        response = self.request_get(url)
        self.assertEqual(response.status_code, httplib.OK)
        data = json.loads(response.content)
        self.assertEqual(len(data['results']), 9)

    def test_read_with_approved_status(self):
        self.make_moderation_objects_with_statuses()
        url = reverse(self.url_name, kwargs={'status': 'approved'})
        response = self.request_get(url)
        self.assertEqual(response.status_code, httplib.OK)
        data = json.loads(response.content)
        self.assertEqual(len(data['results']), 2)
        for obj in data['results']:
            self.assertEqual(obj['status'], BaseModeration.APPROVED_STATUS)

    def test_read_with_pending_status(self):
        self.make_moderation_objects_with_statuses()
        url = reverse(self.url_name, kwargs={'status': 'pending'})
        response = self.request_get(url)
        self.assertEqual(response.status_code, httplib.OK)
        data = json.loads(response.content)
        self.assertEqual(len(data['results']), 3)
        for obj in data['results']:
            self.assertEqual(obj['status'], BaseModeration.PENDING_STATUS)

    def test_read_with_rejected_status(self):
        self.make_moderation_objects_with_statuses()
        url = reverse(self.url_name, kwargs={'status': 'rejected'})
        response = self.request_get(url)
        self.assertEqual(response.status_code, httplib.OK)
        data = json.loads(response.content)
        self.assertEqual(len(data['results']), 4)
        for obj in data['results']:
            self.assertEqual(obj['status'], BaseModeration.REJECTED_STATUS)

    def test_read_with_invalid_status_is_404(self):
        url = reverse(self.url_name)
        response = self.request_get(url + 'invalid/')
        self.assertEqual(response.status_code, httplib.NOT_FOUND)

    def test_read_with_pk_return_single_object(self):
        moderation = self.make_moderation_object()
        url = reverse(self.url_name, kwargs={'pk': moderation.pk})
        response = self.request_get(url)
        self.assertEqual(response.status_code, httplib.OK)
        data = json.loads(response.content)
        self.assertEqual(data['id'], moderation.pk)

    def test_read_with_invalid_pk_is_404(self):
        url = reverse(self.url_name, kwargs={
            'pk': 100000})
        response = self.request_get(url)
        self.assertEqual(response.status_code, httplib.NOT_FOUND)

    def test_read_with_status_and_pk_return_single_object(self):
        moderation = self.make_moderation_object()
        url = reverse(self.url_name, kwargs={
            'status': 'pending', 'pk': moderation.pk})
        response = self.request_get(url)
        self.assertEqual(response.status_code, httplib.OK)
        data = json.loads(response.content)
        self.assertEqual(data['id'], moderation.pk)

    def test_read_with_wrong_status_for_pk_is_404(self):
        moderation = self.make_moderation_object()
        url = reverse(self.url_name, kwargs={
            'status': 'approved', 'pk': moderation.pk})
        response = self.request_get(url)
        self.assertEqual(response.status_code, httplib.NOT_FOUND)

    def test_patch_approved(self):
        moderation = self.make_moderation_object()
        url = reverse(self.url_name, kwargs={
            'status': 'approved', 'pk': moderation.pk})
        response = self.request_patch(url, json.dumps(self.data))
        self.assertEqual(response.status_code, httplib.OK)
        data = json.loads(response.content)
        self.assertEqual(data['id'], moderation.pk)
        self.assertEqual(data['status'], BaseModeration.APPROVED_STATUS)

    def test_patch_rejected(self):
        moderation = self.make_moderation_object()
        url = reverse(self.url_name, kwargs={
            'status': 'rejected', 'pk': moderation.pk})
        response = self.request_patch(url, json.dumps(self.data))
        self.assertEqual(response.status_code, httplib.OK)
        data = json.loads(response.content)
        self.assertEqual(data['id'], moderation.pk)
        self.assertEqual(data['status'], BaseModeration.REJECTED_STATUS)

    def test_patch_pending(self):
        moderation = self.make_moderation_object()
        url = reverse(self.url_name, kwargs={
            'status': 'pending', 'pk': moderation.pk})
        response = self.request_patch(url, json.dumps(self.data))
        self.assertEqual(response.status_code, httplib.BAD_REQUEST)
        data = json.loads(response.content)
        self.assertIn('Moderation status cannot be set to pending.',
                      data['errors']['status'])

    def test_moderation_text_required(self):
        moderation = self.make_moderation_object()
        url = reverse(self.url_name, kwargs={
            'status': 'rejected', 'pk': moderation.pk})
        response = self.request_patch(url)
        self.assertEqual(response.status_code, httplib.BAD_REQUEST)
        data = json.loads(response.content)
        self.assertIn(
            'This field is required.', data['errors']['moderation_text'])

    def test_moderator_required(self):
        moderation = self.make_moderation_object()
        url = reverse(self.url_name, kwargs={
            'status': 'rejected', 'pk': moderation.pk})
        response = self.request_patch(url)
        self.assertEqual(response.status_code, httplib.BAD_REQUEST)
        data = json.loads(response.content)
        self.assertIn(
            'This field is required.', data['errors']['moderator'])

    def test_cannot_change_status_when_approved(self):
        moderation = self.make_moderation_object(
            status=BaseModeration.APPROVED_STATUS)
        url = reverse(self.url_name, kwargs={
            'status': 'rejected', 'pk': moderation.pk})
        response = self.request_patch(url, json.dumps(self.data))
        self.assertEqual(response.status_code, httplib.BAD_REQUEST)
        data = json.loads(response.content)
        self.assertIn('Moderation must be pending to change status.',
                      data['errors']['status'])

    def test_cannot_change_status_when_rejected(self):
        moderation = self.make_moderation_object(
            status=BaseModeration.REJECTED_STATUS)
        url = reverse(self.url_name, kwargs={
            'status': 'approved', 'pk': moderation.pk})
        response = self.request_patch(
            url, json.dumps(self.data), 'application/json')
        self.assertEqual(response.status_code, httplib.BAD_REQUEST)
        data = json.loads(response.content)
        self.assertIn('Moderation must be pending to change status.',
                      data['errors']['status'])

    def test_patch_same_status_approved(self):
        moderation = self.make_moderation_object(
            status=BaseModeration.APPROVED_STATUS)
        url = reverse(self.url_name, kwargs={
            'status': 'approved', 'pk': moderation.pk})
        response = self.request_patch(
            url, json.dumps(self.data))
        self.assertEqual(response.status_code, httplib.BAD_REQUEST)
        data = json.loads(response.content)
        self.assertIn('Moderation must be pending to change status.',
                      data['errors']['status'])

    def test_patch_same_status_rejected(self):
        moderation = self.make_moderation_object(
            status=BaseModeration.REJECTED_STATUS)
        url = reverse(self.url_name, kwargs={
            'status': 'rejected', 'pk': moderation.pk})
        response = self.request_patch(url, json.dumps(self.data))
        self.assertEqual(response.status_code, httplib.BAD_REQUEST)
        data = json.loads(response.content)
        self.assertIn('Moderation must be pending to change status.',
                      data['errors']['status'])

    def test_read_query_count(self):
        for _ in range(5):
            self.make_moderation_object()

        assert_no_extra_queries_after(
            lambda: self.make_moderation_object(),
            self.request_get, reverse(self.url_name))

    def test_ordering(self):
        # the following are sorted in reverse expected order.
        num_flags_and_dates = (
            (1, datetime(2014, 01, 01, tzinfo=pytz.utc)),
            (2, datetime(2014, 06, 01, tzinfo=pytz.utc)),
            (3, datetime(2014, 12, 01, tzinfo=pytz.utc)),
            # In case of flag number tie, earlier wins
            (3, datetime(2014, 03, 01, tzinfo=pytz.utc)),
            (4, datetime(2014, 02, 01, tzinfo=pytz.utc)))
        expected = []
        with mock.patch('core.models.datetime') as datetime_mock:
            for num_flags, date in num_flags_and_dates:
                datetime_mock.datetime.now.return_value = date
                moderation = self.make_moderation_object()
                for _ in range(num_flags):
                    self.make_flag(moderation=moderation)
                expected.append(moderation.pk)
                datetime_mock.reset_mock()
        expected = list(reversed(expected))
        url = reverse(self.url_name)
        response = self.request_get(url, json.dumps(self.data))
        data = json.loads(response.content)
        self.assertEqual([m['id'] for m in data['results']], expected)


class ClickPackageModerationHandlerTestCase(
        BaseClickPackageModerationHandlerTestCase):

    url_name = 'click-api-package-moderations'

    def make_moderation_object(self, status=None):
        return self.factory.makeClickPackageModeration(status=status)

    def make_flag(self, *args, **kwargs):
        return self.factory.makeClickPackageFlag(*args, **kwargs)

    def test_user_that_flagged_the_package_cannot_moderate_it(self):
        moderation = self.make_moderation_object()
        self.factory.makeClickPackageFlag(
            user=self.user, moderation=moderation)
        url = reverse(self.url_name, kwargs={
            'status': 'rejected', 'pk': moderation.pk})
        response = self.request_patch(url, json.dumps(self.data))
        self.assertEqual(response.status_code, httplib.BAD_REQUEST)
        data = json.loads(response.content)
        self.assertIn('User cannot modify this moderation.',
                      data['errors']['moderator'])

    def test_returned_fields(self):
        encoder = DateTimeAwareJSONEncoder()

        def encode_date(date):
            return encoder.encode(date).strip('"')

        moderation = self.make_moderation_object()
        flag = self.factory.makeClickPackageFlag(
            user=self.user, moderation=moderation)
        package = moderation.package
        date_created = encode_date(moderation.date_created)
        date_ratings_updated = encode_date(package.date_ratings_updated)
        expected = {
            'date_created': date_created,
            'date_moderated': None,
            'flags': [{
                'date_created': encode_date(flag.date_created),
                'description': flag.description,
                'summary': flag.summary,
                'user': {
                    'id': flag.user.pk,
                    'email': flag.user.email,
                    'first_name': flag.user.first_name,
                    'last_name': flag.user.last_name,
                    'get_full_name': flag.user.get_full_name()
                }
            }],
            'get_status_display': moderation.get_status_display(),
            'id': moderation.pk,
            'moderation_text': '',
            'package': {
                'id': package.pk,
                'date_ratings_updated': date_ratings_updated,
                'histogram': package.histogram,
                'package_name': package.package_name,
                'ratings_average': package.ratings_average,
                'ratings_total': package.ratings_total},
            'status': moderation.status}
        url = reverse(self.url_name, kwargs={
            'status': 'pending', 'pk': moderation.pk})
        response = self.request_get(url)
        data = json.loads(response.content)
        self.assertEqual(data, expected)


class ClickPackageReviewModerationHandlerTestCase(
        BaseClickPackageModerationHandlerTestCase):

    url_name = 'click-api-review-moderations'

    def make_moderation_object(self, status=None):
        return self.factory.makeClickPackageReviewModeration(status=status)

    def make_flag(self, *args, **kwargs):
        return self.factory.makeClickPackageReviewFlag(*args, **kwargs)

    def test_set_review_moderation_to_rejected_hides_review(self):
        moderation = self.make_moderation_object()
        url = reverse(self.url_name, kwargs={
            'status': 'rejected', 'pk': moderation.pk})
        response = self.request_patch(url, json.dumps(self.data))
        self.assertEqual(response.status_code, httplib.OK)
        obj = ClickPackageReviewModeration.objects.get(pk=moderation.pk)
        self.assertTrue(obj.review.hide)

    def test_user_that_flagged_the_review_cannot_moderate_it(self):
        moderation = self.make_moderation_object()
        self.factory.makeClickPackageReviewFlag(
            user=self.user, moderation=moderation)
        url = reverse(self.url_name, kwargs={
            'status': 'rejected', 'pk': moderation.pk})
        response = self.request_patch(url, json.dumps(self.data))
        self.assertEqual(response.status_code, httplib.BAD_REQUEST)
        data = json.loads(response.content)
        self.assertIn('User cannot modify this moderation.',
                      data['errors']['moderator'])

    def test_review_owner_cannot_moderate_its_flags(self):
        moderation = self.make_moderation_object()
        review = moderation.review
        review.reviewer = self.user
        review.save()
        url = reverse(self.url_name, kwargs={
            'status': 'rejected', 'pk': moderation.pk})
        response = self.request_patch(url, json.dumps(self.data))
        self.assertEqual(response.status_code, httplib.BAD_REQUEST)
        data = json.loads(response.content)
        self.assertIn('User cannot modify this moderation.',
                      data['errors']['moderator'])

    def test_returned_fields(self):
        encoder = DateTimeAwareJSONEncoder()

        def encode_date(date):
            return encoder.encode(date).strip('"')

        moderation = self.make_moderation_object()
        flag = self.factory.makeClickPackageReviewFlag(
            user=self.user, moderation=moderation)
        review = moderation.review
        reviewer = review.reviewer
        expected = {
            'date_created': encode_date(moderation.date_created),
            'date_moderated': None,
            'flags': [{
                'date_created': encode_date(flag.date_created),
                'description': flag.description,
                'summary': flag.summary,
                'user': {
                    'id': flag.user.pk,
                    'email': flag.user.email,
                    'first_name': flag.user.first_name,
                    'last_name': flag.user.last_name,
                    'get_full_name': flag.user.get_full_name()
                }
            }],
            'get_status_display': moderation.get_status_display(),
            'id': moderation.pk,
            'moderation_text': '',
            'review': {
                'id': review.pk,
                'date_created': encode_date(review.date_created),
                'date_deleted': None,
                'hide': review.hide,
                'language': review.language,
                'package_name': review.package_name(),
                'rating': review.rating,
                'review_text': review.review_text,
                'reviewer': {
                    'id': reviewer.pk,
                    'email': reviewer.email,
                    'first_name': reviewer.first_name,
                    'last_name': reviewer.last_name,
                    'get_full_name': reviewer.get_full_name()
                },
                'reviewer_displayname': review.reviewer_displayname(),
                'reviewer_username': review.reviewer_username(),
                'summary': review.summary,
                'usefulness_favorable': review.usefulness_favorable,
                'usefulness_total': review.usefulness_total,
                'version': review.version},
            'status': moderation.status}
        url = reverse(self.url_name, kwargs={
            'status': 'pending', 'pk': moderation.pk})
        response = self.request_get(url)
        data = json.loads(response.content)
        self.assertEqual(data, expected)
