"""Click Package Review model tests."""
import datetime

from django.contrib.auth.models import Permission
from django.core.exceptions import ValidationError
from mock import patch

from clickreviews.forms import ClickPackageReviewForm
from clickreviews.models import (
    ClickPackageFlag,
    ClickPackageModeration,
    ClickPackageReview,
    ClickPackageReviewFlag,
    ClickPackageReviewModeration,
)
from clickreviews.tests.factory import TestCaseWithFactory


class ClickPackageModelTestCase(TestCaseWithFactory):

    def test_update_stats(self):
        # Updating the stats calculates the average, total and histogram.
        click_package = self.factory.makeClickPackage()
        self.factory.makeClickPackageReview(
            click_package=click_package, rating="5")
        self.factory.makeClickPackageReview(
            click_package=click_package, rating="3")

        click_package.update_stats()

        self.assertEqual('4.0', click_package.ratings_average)
        self.assertEqual(2, click_package.ratings_total)
        self.assertEqual("[0, 0, 1, 0, 1]", click_package.histogram)

    def test_update_stats_ignores_hidden(self):
        # Updating the stats ignores hidden reviews.
        click_package = self.factory.makeClickPackage()
        self.factory.makeClickPackageReview(
            click_package=click_package, rating="5")
        self.factory.makeClickPackageReview(
            click_package=click_package, rating="3", hide=True)

        click_package.update_stats()

        self.assertEqual('5.0', click_package.ratings_average)
        self.assertEqual(1, click_package.ratings_total)
        self.assertEqual("[0, 0, 0, 0, 1]", click_package.histogram)

    def test_update_stats_ignores_incorrect(self):
        click_package = self.factory.makeClickPackage()
        self.factory.makeClickPackageReview(
            click_package=click_package, rating="5")
        self.factory.makeClickPackageReview(
            click_package=click_package, rating="6")
        self.factory.makeClickPackageReview(
            click_package=click_package, rating="-1")

        click_package.update_stats()

        self.assertEqual('5.0', click_package.ratings_average)
        self.assertEqual(1, click_package.ratings_total)
        self.assertEqual("[0, 0, 0, 0, 1]", click_package.histogram)

    def test_reviews_for_object(self):
        click_package = self.factory.makeClickPackage()
        self.factory.makeClickPackageReview(
            click_package=click_package, hide=True)
        review = self.factory.makeClickPackageReview(
            click_package=click_package, rating=3)
        deleted = self.factory.makeClickPackageReview(
            click_package=click_package)
        deleted.delete()
        self.assertEqual(list(click_package.reviews_for_object()), [review])


class ClickPackageReviewManagerTestCase(TestCaseWithFactory):

    def test_published(self):
        self.factory.makeClickPackageReview(hide=True)
        review = self.factory.makeClickPackageReview()
        deleted = self.factory.makeClickPackageReview()
        deleted.delete()
        self.assertEqual(
            list(ClickPackageReview.objects.published()), [review])


class ClickPackageReviewModelTestCase(TestCaseWithFactory):

    def test_max_lengths(self):
        """ Check max_length of 80 for summary and 5000 for text, per spec"""
        # We use a form to test the model here because Django doesn't enforce
        # max_length at the model level.  We could rely on the database to
        # complain, except sqlite doesn't :)

        user = self.factory.makeUser()
        click_package = self.factory.makeClickPackage()
        version = self.factory.makeVersion()
        form_data = {
            'reviewer': user.pk,
            'click_package': click_package.pk,
            'version': version,
            'rating': 4,
            'summary': 'a' * 82,
            'review_text': 'Up to standard',
            'language': 'en',
            'hide': False,
        }
        form = ClickPackageReviewForm(form_data)

        self.assertFailsValidation(form, 'summary')

        form_data.update({'summary': 'Nice', 'review_text': 'b' * 5010})
        form = ClickPackageReviewForm(form_data)

        self.assertFailsValidation(form, 'review_text')

        form_data.update({'version': 'x' * 101})
        form = ClickPackageReviewForm(form_data)

        self.assertFailsValidation(form, 'version')

    def assertFailsValidation(self, form, field):
        self.assertFalse(form.is_valid(), 'form should fail validation')
        self.assertTrue(field in form.errors)

    def test_is_awaiting_moderation_no_moderations(self):
        review = self.factory.makeClickPackageReview()
        review.moderation_set.all().delete()
        self.assertFalse(review.is_awaiting_moderation())

    def test_is_awaiting_moderation_with_pending_moderations(self):
        review = self.factory.makeClickPackageReview()
        self.factory.makeClickPackageReviewModeration(review=review)
        self.assertTrue(review.is_awaiting_moderation())

    def test_is_awaiting_moderation_with_no_pending_moderations(self):
        review = self.factory.makeClickPackageReview()
        self.factory.makeClickPackageReviewModeration(
            review=review, status=ClickPackageReviewModeration.APPROVED_STATUS)
        self.factory.makeClickPackageReviewModeration(
            review=review, status=ClickPackageReviewModeration.REJECTED_STATUS)
        self.assertFalse(review.is_awaiting_moderation())

    @patch('clickreviews.models.datetime')
    def test_delete(self, mock_datetime):
        expected = datetime.datetime.now()
        mock_datetime.now.return_value = expected
        review = self.factory.makeClickPackageReview(rating=3)
        assert not review.hide and review.date_deleted is None
        self.assertEqual(review.click_package.histogram, '[0, 0, 1, 0, 0]')
        review.delete()
        self.assertTrue(review.hide)
        self.assertEqual(review.date_deleted, expected)
        # Object is not really deleted.
        self.assertTrue(
            ClickPackageReview.objects.filter(pk=review.pk).exists())
        # Stats should be updated.
        self.assertEqual(review.click_package.histogram, '[0, 0, 0, 0, 0]')

    def test_reviewer_username_same_username_and_openid(self):
        expected = 'theuoid'
        user = self.factory.makeUser(username=expected, open_id=expected)
        review = self.factory.makeClickPackageReview(reviewer=user)
        self.assertEqual(review.reviewer_username(), expected)

    def test_reviewer_username_returns_openid(self):
        expected = 'theopenid'
        user = self.factory.makeUser(
            username='bogusfromlogin', open_id=expected)
        review = self.factory.makeClickPackageReview(reviewer=user)
        self.assertEqual(review.reviewer_username(), expected)

    def test_reviewer_username_fallbacks_to_username(self):
        expected = 'bogusfromlogin'
        user = self.factory.makeUser(username=expected, open_id=False)
        review = self.factory.makeClickPackageReview(reviewer=user)
        self.assertEqual(review.reviewer_username(), expected)


class BaseClickPackageModerationTestCase(TestCaseWithFactory):

    def __init__(self, *args, **kwargs):
        super(BaseClickPackageModerationTestCase, self).__init__(
            *args, **kwargs)
        # skip this base class to be run as a TestCase
        if self.__class__ == BaseClickPackageModerationTestCase:
            self.run = lambda self, *args, **kwargs: None

    model = None
    moderated_object_field_name = None

    def make_moderated_object(self):
        raise NotImplementedError

    def _make_moderated_object_kwarg(self, obj=None):
        if obj is None:
            obj = self.make_moderated_object()
        return {self.moderated_object_field_name: obj}

    def test_get_moderated_object(self):
        moderated_obj = self.make_moderated_object()
        moderation = self.model.objects.create(
            **self._make_moderated_object_kwarg(obj=moderated_obj))
        self.assertEqual(moderated_obj, moderation.get_moderated_object())

    def assertValidationError(self, ctx, field, message):
        errors = ctx.exception.message_dict
        self.assertIn(message, errors[field])

    def test_moderated_object_required(self):
        with self.assertRaises(ValidationError) as ctx:
            obj = self.model()
            obj.full_clean()
        self.assertValidationError(
            ctx, self.moderated_object_field_name,
            'This field cannot be null.')

    def test_default_status(self):
        obj = self.model(**self._make_moderated_object_kwarg())
        self.assertEqual(
            obj.status, self.model.PENDING_STATUS)

    def test_status_choices_invalid_type(self):
        with self.assertRaises(ValidationError) as ctx:
            obj = self.model(
                status='SUPERINVALID', **self._make_moderated_object_kwarg())
            obj.full_clean()
        self.assertValidationError(
            ctx, 'status', "'SUPERINVALID' value must be an integer.")

    def test_status_choices_invalid_value(self):
        with self.assertRaises(ValidationError) as ctx:
            obj = self.model(
                status=1000, **self._make_moderated_object_kwarg())
            obj.full_clean()
        self.assertValidationError(
            ctx, 'status', 'Value 1000 is not a valid choice.')

    @patch('core.models.datetime')
    def test_date_created_defaults_to_now(self, datetime_mock):
        now = datetime.datetime.utcnow()
        datetime_mock.datetime.now.return_value = now

        obj = self.model.objects.create(**self._make_moderated_object_kwarg())
        self.assertEqual(obj.date_created, now)


class ClickPackageModerationTestCase(BaseClickPackageModerationTestCase):

    model = ClickPackageModeration
    moderated_object_field_name = 'package'

    def make_moderated_object(self):
        return self.factory.makeClickPackage()

    def test_on_moderate(self):
        # Nothing yet
        pass

    def test_cannot_be_moderated_by_the_same_user_which_flagged(self):
        moderation = self.factory.makeClickPackageModeration()
        flag = self.factory.makeClickPackageFlag(moderation=moderation)
        can_moderate = Permission.objects.get(codename='can_moderate')
        flag.user.user_permissions.add(can_moderate)
        self.assertFalse(moderation.can_be_moderated_by(flag.user))

    def test_can_be_moderated_by_user(self):
        moderation = self.factory.makeClickPackageModeration()
        user = self.factory.makeUser()
        can_moderate = Permission.objects.get(codename='can_moderate')
        user.user_permissions.add(can_moderate)
        self.assertTrue(moderation.can_be_moderated_by(user))


class ClickPackageReviewModerationTestCase(BaseClickPackageModerationTestCase):

    model = ClickPackageReviewModeration
    moderated_object_field_name = 'review'

    def make_moderated_object(self):
        return self.factory.makeClickPackageReview()

    def test_on_moderate_approved(self):
        moderation = self.factory.makeClickPackageReviewModeration(
            status=ClickPackageReviewModeration.APPROVED_STATUS)
        moderation.on_moderate(moderation.review)
        review = ClickPackageReview.objects.get(pk=moderation.review.pk)
        self.assertFalse(review.hide)

    def test_on_moderate_pending(self):
        moderation = self.factory.makeClickPackageReviewModeration(
            status=ClickPackageReviewModeration.PENDING_STATUS)
        moderation.on_moderate(moderation.review)
        review = ClickPackageReview.objects.get(pk=moderation.review.pk)
        self.assertFalse(review.hide)

    def test_on_moderate_rejected(self):
        moderation = self.factory.makeClickPackageReviewModeration(
            status=ClickPackageReviewModeration.REJECTED_STATUS)
        moderation.on_moderate(moderation.review)
        review = ClickPackageReview.objects.get(pk=moderation.review.pk)
        self.assertTrue(review.hide)

    def test_cannot_be_moderated_by_the_same_user_which_flagged(self):
        flag = self.factory.makeClickPackageReviewFlag()
        can_moderate = Permission.objects.get(codename='can_moderate')
        flag.user.user_permissions.add(can_moderate)
        self.assertFalse(flag.moderation.can_be_moderated_by(flag.user))

    def test_can_be_moderated_by_user(self):
        moderation = self.factory.makeClickPackageReviewModeration()
        user = self.factory.makeUser()
        can_moderate = Permission.objects.get(codename='can_moderate')
        user.user_permissions.add(can_moderate)
        self.assertTrue(moderation.can_be_moderated_by(user))

    def test_cannot_be_moderated_by_review_owner(self):
        # user passes all conditions, except that he owns the review
        moderation = self.factory.makeClickPackageReviewModeration()
        can_moderate = Permission.objects.get(codename='can_moderate')
        user = moderation.review.reviewer
        user.user_permissions.add(can_moderate)
        self.assertFalse(moderation.can_be_moderated_by(user))


class BaseClickPackageFlagTestCase(TestCaseWithFactory):

    model = None

    def __init__(self, *args, **kwargs):
        super(BaseClickPackageFlagTestCase, self).__init__(
            *args, **kwargs)
        # skip this base class to be run as a TestCase
        if self.__class__ == BaseClickPackageFlagTestCase:
            self.run = lambda self, *args, **kwargs: None

    def assertValidationError(self, ctx, field, message):
        errors = ctx.exception.message_dict
        self.assertIn(message, errors[field])

    def make_moderation_object(self):
        raise NotImplemented

    def test_get_moderation_object(self):
        moderation_obj = self.make_moderation_object()
        flag = self.model.objects.create(
            summary='Summary',
            description='Description',
            moderation=moderation_obj,
            user=self.factory.makeUser())
        self.assertEqual(moderation_obj, flag.get_moderation_object())

    def test_summary_required(self):
        obj = self.model(description='Description')
        with self.assertRaises(ValidationError) as ctx:
            obj.full_clean()
        self.assertValidationError(
            ctx, 'summary', 'This field cannot be blank.')

    def test_summary_max_length(self):
        moderation = self.make_moderation_object()
        obj = self.model(
            summary='a' * 201,
            description='Description',
            moderation=moderation)
        with self.assertRaises(ValidationError) as ctx:
            obj.full_clean()
        self.assertValidationError(
            ctx, 'summary',
            'Ensure this value has at most 200 characters (it has 201).')

    def test_description_required(self):
        obj = self.model(summary='Summary')
        with self.assertRaises(ValidationError) as ctx:
            obj.full_clean()
        self.assertValidationError(
            ctx, 'description', 'This field cannot be blank.')

    @patch('core.models.datetime')
    def test_date_created_has_default(self, datetime_mock):
        now = datetime.datetime.utcnow()
        datetime_mock.datetime.now.return_value = now

        obj = self.model.objects.create(
            summary='Summary',
            description='Description',
            moderation=self.make_moderation_object(),
            user=self.factory.makeUser())

        # not providing a date shouldn't raise any errors
        obj.full_clean()

        self.assertEqual(obj.date_created, now)


class ClickPackageFlagTestCase(BaseClickPackageFlagTestCase):

    model = ClickPackageFlag

    def make_moderation_object(self):
        return self.factory.makeClickPackageModeration()


class ClickPackageReviewFlagTestCase(BaseClickPackageFlagTestCase):

    model = ClickPackageReviewFlag

    def make_moderation_object(self):
        return self.factory.makeClickPackageReviewModeration()
