import hmac
import shlex
import time
from hashlib import sha256

from django.conf import settings
from django.http import HttpResponse


MAX_AGE = getattr(settings, 'DELEGATED_SSO_AUTH_MAX_AGE', 60)
IDENTIFIER = 'DelegatedSSO'


def tokenize_authorization(value_string):
    """Tokenize the authorization header value.

    Return:
        A generator, which returns the authorization mechanism
        identifier as first item.  Subsequent values are two element
        tuples for each (key, value) pair.

    """
    lexer = shlex.shlex(value_string, posix=True)

    try:
        yield next(lexer)
    except ValueError:
        # If parsing failed, then return empty string as the mechanism
        # identifier and stop processing any further.
        yield ''
        raise StopIteration

    while True:
        key = next(lexer)

        # Stop processing at first malformed data
        equal = next(lexer)
        if equal != '=':
            raise StopIteration

        value = next(lexer)

        try:
            # Last item in a token group must be a comma.
            comma = next(lexer)  # drop ,
            if comma != ',':
                raise StopIteration
            yield (key, value)
        except StopIteration:
            # Except it's not mandatory for the last group of tokens.
            yield (key, value)
            raise StopIteration


def parse_authorization(tokens):
    """Return signature and timestamp from tokens.

    Args:
        tokens: Iterable of tuples containing (key, value) pairs.

    Return:
        A tuple containing signature and timestamp.

    """
    signature = None
    timestamp = None

    for key, value in tokens:
        # ignore all unexpected params
        if key == 'timestamp':
            try:
                float(value)
            except ValueError:
                pass
            else:
                timestamp = value
        elif key == 'signature':
            signature = value

    return signature, timestamp


def is_expired(timestamp, max_age=MAX_AGE):
    """Check if timestamp is not within the max_age."""
    timestamp = float(timestamp)
    return abs(time.time() - timestamp) > max_age


def is_signature_valid(signature, secret, uri, timestamp, body=None):
    """Return True if signature is valid."""
    # Safeguard in case body is not sent as the empty string.
    if body is None:
        body = ''
    message = "%s%s%s" % (uri, timestamp, body)
    calculated = hmac.new(secret, message, sha256).hexdigest()
    return calculated == signature


class DelegatedSSOAuthentication(object):
    """Delegated SSO authentication.

    This authentication mechanism allows piston resources to be used
    by external services sharing a secret key.

    The Authorization must be of the form:

    DelegatedSSO signature="",timestamp=""

    Args:
        secret: The shared secret between the consumer and the server.
        realm: A string used for the realm when asking authentication.

    Details:
        signature: HMAC SHA-256 signature of concatenating, the full
            URI, timestamp, and request's body.
        timestamp: Seconds from UNIX epoch, as returned by time.time().

    """

    def __init__(self, secret=None, realm='API'):
        self.secret = secret
        self.realm = realm

    def challenge(self):
        resp = HttpResponse("Authorization Required")
        resp['WWW-Authenticate'] = '%s realm="%s"' % (IDENTIFIER, self.realm)
        resp.status_code = 401
        return resp

    def is_authenticated(self, request):
        authenticated = False
        identifier = None
        authorization = request.META.get('HTTP_AUTHORIZATION')

        if authorization:
            tokens = tokenize_authorization(authorization)
            identifier = next(tokens)

        if identifier == IDENTIFIER:
            signature, timestamp = parse_authorization(tokens)
            uri = request.build_absolute_uri()
            # A bit of lazyness, checks from less to most intensive
            authenticated = (
                signature and timestamp and not is_expired(timestamp) and
                is_signature_valid(
                    signature, self.secret, uri, timestamp, request.body))

        return authenticated
