#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""SSO/OAuth authenticator for Piston."""

from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'SSOOAuthAuthentication',
]

from django.http import HttpResponse
from oauth import oauth
from piston.authentication import OAuthAuthentication

from core.utilities import (
    web_services,
    update_or_create_user_from_oauth,
)


class SSOOAuthAuthentication(OAuthAuthentication):
    """ This class is a Piston Authentication class.

    See http://bitbucket.org/jespern/django-piston/wiki/Documentation for
    more info.
    """
    def __init__(self, realm='API'):
        self.realm = realm

    def challenge(self):
        """ We define our own challenge so that we can provide our own realm
        and return a simple error message instead of rendering a template.
        """
        resp = HttpResponse("Authorization Required")
        resp['WWW-Authenticate'] = 'OAuth realm="%s"' % self.realm
        resp.status_code = 401
        return resp

    def is_authenticated(self, request):
        authorization = request.META.get('HTTP_AUTHORIZATION', None)
        query_string = request.META.get('QUERY_STRING', None)
        if not authorization and not query_string:
            return False

        auth_data = {
            'http_url': request.build_absolute_uri(),
            'http_method': request.method,
            'authorization': authorization,
            'query_string': query_string,
        }

        # check this is a valid SSO OAuth request
        success = web_services.validate_request(data=auth_data)

        headers = {}
        if authorization:
            headers = {'Authorization': authorization}

        if not success:
            return False

        oauth_request = oauth.OAuthRequest.from_request(
            http_method=auth_data['http_method'],
            http_url=auth_data['http_url'],
            headers=headers,
            query_string=auth_data.get('query_string', None),
        )

        openid = oauth_request.get_parameter('oauth_consumer_key')

        user, _ = update_or_create_user_from_oauth(openid)
        if user is not None:
            request.user = user
        else:
            success = False

        return success
