#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Decorators used in the Ratings and Reviews API."""

from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'dont_vary',
]

from functools import wraps

from django.conf import settings
from django.utils.cache import patch_vary_headers
from django.utils.decorators import available_attrs
from django.views.decorators.gzip import gzip_page


def dont_vary(func):
    """A decorator that removes any Vary header the response has."""
    @wraps(func)
    def wrapper(*args, **kwargs):
        response = func(*args, **kwargs)
        if 'vary' in response:
            del response['vary']
        return response
    return wrapper


def vary_only_on_headers(*headers):
    """A view decorator that forces the specified headers to the Vary
    header of the response (removing all previously set ones).

    The motivation to enforce headers and not use
    `django.views.decorators.vary.vary_on_headers` is that such
    decorator just adds to existing Vary headers and does not allow to
    override those automatically set in handlers (e.g 'Authorization'
    that's added automatically by piston, even in unauthenticated
    endpoints).

    Usage:

       @vary_only_on_headers('Cookie', 'Accept-language')
       def index(request):
           ...

       # This removes the Vary header from response entirely
       @vary_only_on_headers()
       def index(request):
           ...

    Note that the header names are not case-sensitive.
    """
    def decorator(func):
        @wraps(func, assigned=available_attrs(func))
        def inner_func(*args, **kwargs):
            response = dont_vary(func)(*args, **kwargs)
            if headers:
                patch_vary_headers(response, headers)
            return response
        return inner_func
    return decorator


def vary_only_on_accept(func):
    """Shortcut for vary_only_on_headers('Accept')"""
    return vary_only_on_headers('Accept')(func)


def gzip_content(func):
    """A decorator that gzip-encodes content, based on a setting."""
    @wraps(func)
    def wrapper(*args, **kwargs):
        if getattr(settings, 'ENABLE_GZIP_CONTENT', True):
            response = gzip_page(func)(*args, **kwargs)
        else:
            response = func(*args, **kwargs)
        return response
    return wrapper
