import urllib
import urlparse

from django.core.paginator import (
    EmptyPage,
    InvalidPage,
    PageNotAnInteger,
    Paginator
)


def _get_page_url(request, page_number_fn):
    absolute_uri = request.build_absolute_uri()
    uri_split = urlparse.urlsplit(absolute_uri)
    query_dict = urlparse.parse_qs(uri_split.query)

    try:
        page_number = page_number_fn()
        query_dict['page'] = [page_number]
        query = urllib.urlencode(query_dict, doseq=True)
        page_url = urlparse.SplitResult(uri_split.scheme,
                                        uri_split.netloc,
                                        uri_split.path,
                                        query,
                                        uri_split.fragment).geturl()
    except InvalidPage:
        page_url = None

    return page_url


def _get_previous_page_url(request, page):
    return _get_page_url(request, page.previous_page_number)


def _get_next_page_url(request, page):
    return _get_page_url(request, page.next_page_number)


def make_paginated_response(
        request, queryset, page_size=20, max_page_size=100):
    """Return paginated results with pagination data.

    Args:
        request: A request object.
        queryset: A QuerySet or iterable to be paginated.
        page_size: The number of results per page.
        max_page_size: The max number of results per page.
            Results are unlimited when None.

    Returns:
        A Dict including links to previous/next page, the object
        list count for current page, the total number of results
        and the results for the current page.
    """
    page_size = request.GET.get('page_size', page_size)
    if max_page_size is not None:
        page_size = min(page_size, max_page_size)

    paginator = Paginator(queryset, page_size)
    page = request.GET.get('page', 1)
    try:
        page = paginator.page(page)
    except PageNotAnInteger:
        page = paginator.page(1)
    except EmptyPage:
        page = paginator.page(paginator.num_pages)

    # 'django-rest-framework'-like serialization
    return {'previous': _get_previous_page_url(request, page),
            'next': _get_next_page_url(request, page),
            'count': len(page.object_list),
            'results': page.object_list}
