import datetime

import pytz

from django.db import models
from django.db.models import Avg, Count
from django.utils.translation import ugettext_lazy as _


class ModerationError(Exception):
    """Raised when BaseModeration.moderate is called incorrectly."""


class UserAlreadyFlaggedReview(Exception):
    """Raised when a user has already flagged a certain review once."""


class BaseModeration(models.Model):
    """Base class for moderated objects.

    Provides an unified API to record pending moderation and the
    action taken.

    During active moderation (moderate by default) a pending
    Moderation may be created when items are first saved.

    Afterwards, or during passive moderation (moderate when flagged),
    a ModerationObject may be created when an item is flagged.
    """
    PENDING_STATUS = 0
    APPROVED_STATUS = 1
    REJECTED_STATUS = 2
    MODERATION_CHOICES = (
        (APPROVED_STATUS, _('Approved')),
        (REJECTED_STATUS, _('Rejected')),
    )
    STATUS_CHOICES = MODERATION_CHOICES + (
        (PENDING_STATUS, _('Pending')),
    )

    status = models.SmallIntegerField(
        choices=STATUS_CHOICES, default=PENDING_STATUS)

    date_created = models.DateTimeField(
        default=lambda: datetime.datetime.now(pytz.utc),
        editable=False)
    date_moderated = models.DateTimeField(null=True, blank=True)
    moderator = models.ForeignKey('auth.User', null=True, blank=True)
    moderation_text = models.TextField(blank=True)

    class Meta:
        abstract = True

    def __unicode__(self):
        obj = self.get_moderated_object()
        return u'{0} for {1} {2}'.format(
            self.__class__.__name__, obj.__class__.__name__, obj.pk)

    def get_moderated_object(self):
        """Return the moderated object."""
        raise NotImplementedError

    def on_moderate(self, obj):
        """Action to perform after moderation change status."""
        raise NotImplementedError

    def can_be_moderated_by(self, user):
        """Return whether the user is allowed to moderate this object."""
        raise NotImplementedError

    def moderate(self, moderator, status, moderation_text):
        """Record the moderation of the review and hide/un-hide it."""
        self.moderator = moderator
        self.status = status
        self.moderation_text = moderation_text
        self.date_moderated = datetime.datetime.now(pytz.utc)
        self.save()

        obj = self.get_moderated_object()
        self.on_moderate(obj)


class ModerationManager(models.Manager):
    """Provide custom filters based on the status."""
    use_for_related_fields = True

    def pending(self):
        return self.get_query_set().filter(
            status=self.model.PENDING_STATUS)

    def approved(self):
        return self.get_query_set().filter(
            status=self.model.APPROVED_STATUS)

    def rejected(self):
        return self.get_query_set().filter(
            status=self.model.REJECTED_STATUS)


class BaseModerationFlag(models.Model):
    """Base class for flagging objects for moderation.

    Subclasses must add a FK to a `BaseModeration` subclass,
    preferably using the `moderation` property to hold the model
    field.  Finally `get_moderation_object` method should be
    implemented and point to the aforementioned FK.
    """

    summary = models.CharField(max_length=200)
    description = models.TextField()
    date_created = models.DateTimeField(
        default=lambda: datetime.datetime.now(pytz.utc),
        editable=False)
    user = models.ForeignKey('auth.User', null=True)

    class Meta:
        abstract = True

    # XXX: This wouldn't be needed if we agreed that every flag
    # subclass would point to a `BaseModeration` subclass using the
    # moderation property, but then `reviewsapp.ReviewModerationFlag`
    # is already using `review_moderation` for that and a refactor
    # doesn't give much gain ATM.
    def get_moderation_object(self):
        """Retrieve the moderation instance linked to this flag."""
        raise NotImplementedError

    def __unicode__(self):
        moderation = self.get_moderation_object()
        obj = moderation.get_moderated_object()
        return u'{0} for {1} {2}'.format(
            self.__class__.__name__, obj.__class__.__name__, obj.pk)


class ObjectWithReviewStats(models.Model):
    """Mixin for any item which can aggregate review statistics.

    Note: the item must provide a valid reviews_for_object method
    returning a result-set of reviews.
    """
    ratings_total = models.IntegerField(default=0)
    ratings_average = models.DecimalField(default=0, max_digits=3,
                                          decimal_places=2)
    histogram = models.CharField(max_length=64, blank=True, null=True)

    date_ratings_updated = models.DateTimeField(
        default=lambda: datetime.datetime.now(pytz.utc))

    def update_stats(self):
        reviews = self.reviews_for_object()
        count = reviews.values('rating').annotate(count=Count('rating'))
        histogram = [0] * 5
        for item in count:
            histogram[item['rating'] - 1] = item['count']
        self.histogram = str(histogram)

        stats = reviews.aggregate(average=Avg('rating'), total=Count('rating'))
        self.ratings_total = stats['total']

        # If there are no visible reviews, the average will be returned
        # as None.
        if stats['average'] is None:
            stats['average'] = 0

        self.ratings_average = str(stats['average'])
        self.date_ratings_updated = datetime.datetime.now(pytz.utc)
        self.save()

    class Meta:
        abstract = True
