from itertools import count

from django.contrib.auth.models import User
from django.test import TestCase
from django_openid_auth.models import UserOpenID
from oauth.oauth import (
    OAuthRequest,
    OAuthConsumer,
    OAuthToken,
    OAuthSignatureMethod_PLAINTEXT
)

from core.utilities import full_claimed_id


class BaseObjectFactory(object):

    counter = count()

    def getUniqueInteger(self):
        """Return a unique integer for the test run."""
        return self.counter.next()

    def getUniqueString(self, prefix=None):
        if prefix is None:
            prefix = "generic-string"
        return prefix + str(self.getUniqueInteger())

    def makeUser(self, username=None, email=None, password='test',
                 first_name=None, last_name=None, open_id=None,
                 is_admin=False):
        if username is None:
            username = self.getUniqueString(prefix='username')
        if email is None:
            email = self.getUniqueString(prefix='email') + '@example.com'
        if first_name is None:
            first_name = self.getUniqueString(prefix='firstname')
        if last_name is None:
            last_name = self.getUniqueString(prefix='lastname')

        user = User.objects.create_user(
            username=username, email=email, password=password)
        user.first_name = first_name
        user.last_name = last_name

        if is_admin:
            user.is_staff = True
            user.is_superuser = True
        user.save()

        # Create an openid record too.
        if open_id is None:
            open_id = self.getUniqueString(prefix='ident-')
        elif open_id is False:
            return user

        claimed_id = full_claimed_id(open_id)

        UserOpenID.objects.create(
            user=user, claimed_id=claimed_id, display_id=claimed_id)

        return user

    def _get_signed_oauth_request(
            self, request, token_key=None, token_secret=None,
            consumer_key=None, consumer_secret=None):
        if token_key is None:
            token_key = self.getUniqueString(prefix='token-key-')
        if token_secret is None:
            token_secret = self.getUniqueString(prefix='token-secret-')
        if consumer_key is None:
            consumer_key = self.getUniqueString(prefix='consumer-key-')
        if consumer_secret is None:
            consumer_secret = self.getUniqueString(prefix='consumer-secret-')
        consumer = OAuthConsumer(consumer_key, consumer_secret)
        token = OAuthToken(token_key, token_secret)
        oauth_request = OAuthRequest.from_consumer_and_token(
            consumer, token, http_url=request.build_absolute_uri())
        oauth_request.sign_request(
            OAuthSignatureMethod_PLAINTEXT(), consumer, token)
        return oauth_request

    def header_from_token(
            self, request, token_key=None, token_secret=None,
            consumer_key=None, consumer_secret=None):
        oauth_request = self._get_signed_oauth_request(
            request, token_key, token_secret, consumer_key, consumer_secret)
        headers = oauth_request.to_header('Lots of testing')
        return {'HTTP_AUTHORIZATION': headers['Authorization']}

    def querystring_from_token(
            self, request, token_key=None, token_secret=None,
            consumer_key=None, consumer_secret=None):
        oauth_request = self._get_signed_oauth_request(
            request, token_key, token_secret, consumer_key, consumer_secret)
        return oauth_request.to_postdata()


class TestCaseWithFactory(TestCase):

    def setUp(self):
        super(TestCaseWithFactory, self).setUp()
        self.factory = BaseObjectFactory()
