from operator import itemgetter

from django.db import connection
from django.test.utils import CaptureQueriesContext
from mock import patch

from core.utilities import web_services


class SSOTestCaseMixin(object):

    def patch_sso(self):
        p = patch('core.api.auth.sso.web_services.validate_request')
        self.mock_validate = p.start()
        self.mock_validate.return_value = True
        self.addCleanup(p.stop)

        p = patch('core.utilities.web_services.get_data_for_account')
        self.mock_get_data = p.start()
        self.mock_get_data.return_value = self.mock_data_for_account()
        self.addCleanup(p.stop)

        web_services.refresh_connections()

    def mock_data_for_account(self, **kwargs):
        """A reasonable return value for get_data_for_account."""
        data = {
            'email': 'foo@bar.com',
            'verified': True,
            'displayname': 'Foo',
        }
        data.update(**kwargs)
        return data


def assert_no_extra_queries_after(modify_data, f, *args, **kwargs):
    """Assert function performs equal query numbers after some change.

    Many views perform terribly because they perform extra queries
    for every item in some list. For example when looping over a
    list of users to present them, if a value is accessed through
    a foreign key then it may perform an extra query. If care
    is not taken then the view may perform terribly when
    operating on a large list. This is often worked around via
    .prefetch_related() or .select_related().

    This function helps in testing that this doesn't happen. The arguments
    to this function are a callable to modify data, and a function
    to call before and after (*args and **kwargs are passed
    through to the latter function). Both calls to the function
    should issue the same number of queries.

    For instance

    self.factory.makeUser()
    assert_no_extra_queries_after(
        self.factory.makeUser,
        self.client.get, '/users'):

    Will first request the view and record how many queries were
    done. It will then run the modify function, creating another user
    in this case, and then request the view again. If the second
    request doesn't issue the same number of queries then an
    AssertionError will be raised.
    """
    first = CaptureQueriesContext(connection)
    with first:
        f(*args, **kwargs)
    modify_data()
    second = CaptureQueriesContext(connection)
    with second:
        f(*args, **kwargs)
    if len(first) != len(second):
        message = ("Different number of queries for the "
                   "second execution. {} != {}".format(
                       len(first), len(second)))

        def format_queries(qs):
            return "\n\n".join(map(str, map(itemgetter('sql'), qs)))

        message += "\nFirst queries:\n"
        message += format_queries(first.captured_queries)
        message += "\n\n\nSecond queries:\n"
        message += format_queries(second.captured_queries)
        raise AssertionError(message)
