#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Ubuntu App Review auth tests."""
from __future__ import absolute_import

import datetime
import hmac
import httplib
import time
from hashlib import sha256

from django.conf import settings
from django.contrib.auth.models import User
from django.http import HttpResponse
from django.test import TestCase

from mock import Mock, patch
from piston.resource import Resource
from piston.handler import BaseHandler

from core.api.auth import (
    DelegatedSSOAuthentication,
    SSOOAuthAuthentication,
    delegated_sso,
)
from core.tests.factory import TestCaseWithFactory
from core.tests.helpers import SSOTestCaseMixin
from core.utilities import full_claimed_id


class MockRequest:
    def __init__(self, method='GET'):
        self.META = {'QUERY_STRING': ''}
        self.GET = {}
        self.POST = {}
        self.REQUEST = {}
        self.body = ''
        self.method = method
        self.environ = self.META

    def build_absolute_uri(self):
        return 'http://localhost:8000/api/1.0/foo'


class MockGetHandler(BaseHandler):
    allowed_methods = ('GET',)

    def read(self, request):
        self.called = True
        return 'Hi!'


class MockPostHandler(BaseHandler):
    allowed_methods = ('POST',)

    def create(self, request):
        self.called = True
        return HttpResponse(status=httplib.CREATED)


class BaseAuthenticationTestCase(SSOTestCaseMixin, TestCaseWithFactory):

    def _get_auth(self):
        raise NotImplementedError

    def setUp(self):
        super(BaseAuthenticationTestCase, self).setUp()
        self.auth = self._get_auth()
        self.user = self.factory.makeUser()
        self.openid = self.user.useropenid_set.get().claimed_id.split('/')[-1]
        self.get_resource = Resource(MockGetHandler,
                                     authentication=self.auth)
        self.post_resource = Resource(MockPostHandler,
                                      authentication=self.auth)
        self.patch_sso()


class SSOOAuthAuthenticationTestCase(BaseAuthenticationTestCase):

    def _get_auth(self):
        return SSOOAuthAuthentication(realm="Ratings and Reviews")

    def make_request_with_token(self, method='GET', token_key=None):
        request = MockRequest(method=method)
        headers = self.factory.header_from_token(
            request, consumer_key=self.openid, token_key=token_key)
        request.META.update(headers)
        return request

    def make_request_with_querystring_token(self, method='GET'):
        request = MockRequest(method=method)
        qs = self.factory.querystring_from_token(
            request, consumer_key=self.openid)
        request.META.update({'QUERY_STRING': qs})
        return request

    def test_no_auth_fails(self):
        request = MockRequest()
        assert not request.META.get('HTTP_AUTHORIZATION')
        assert not request.META.get('QUERY_STRING')

        response = self.get_resource(request)

        self.assertEqual(httplib.UNAUTHORIZED, response.status_code)
        self.assertFalse(self.mock_validate.called)

    def test_success(self):
        request = self.make_request_with_token()

        response = self.get_resource(request)

        auth_data = {
            'http_url': request.build_absolute_uri(),
            'http_method': request.method,
            'authorization': request.META.get('HTTP_AUTHORIZATION', None),
            'query_string': request.META.get('QUERY_STRING', None),
        }
        self.mock_validate.assert_called_with(data=auth_data)
        self.assertEqual(httplib.OK, response.status_code)

    def test_fetch_creds_from_sso(self):
        request = self.make_request_with_token()
        self.mock_get_data.return_value = self.mock_data_for_account()

        response = self.get_resource(request)

        self.assertTrue(self.mock_validate.called)
        self.assertEqual(httplib.OK, response.status_code)

    def test_post_data_doesnt_replace_parameters(self):
        """Check that POST data doesn't replace information from
        the Authentication header in the OAuthRequest parameter set.
        If the post request succeeds then the information from the header
        made it into the parameter list.
        """
        request = self.make_request_with_token(method='POST')
        request.POST = request.REQUEST = {'foo': 'bar', 'baz': 'troz'}

        response = self.post_resource(request)

        self.assertEqual(httplib.CREATED, response.status_code)

    def test_user_doesnt_exist(self):
        """Check that a user is created if none exists"""
        request = self.make_request_with_token()
        data_for_account = self.mock_data_for_account()
        self.mock_get_data.return_value = data_for_account
        self.user.delete()
        new_user = self.factory.makeUser()

        create_method = 'django.contrib.auth.models.UserManager.create_user'
        with patch(create_method) as mock_create_user:
            mock_create_user.return_value = new_user
            response = self.get_resource(request)

        mock_create_user.assert_called_with(
            self.openid, data_for_account['email'], password=None)
        self.assertEqual(httplib.OK, response.status_code)

    def test_user_is_not_authenticated(self):
        request = self.make_request_with_token()
        data_for_account = self.mock_data_for_account()
        self.mock_get_data.return_value = data_for_account
        self.user.delete()
        self.factory.makeUser()

        get_method = 'core.api.auth.sso.update_or_create_user_from_oauth'
        with patch(get_method) as mock_get_user:
            mock_get_user.return_value = (None, False)
            response = self.get_resource(request)

        self.assertEqual(httplib.UNAUTHORIZED, response.status_code)

    def test_invalid_token(self):
        """Check that a 401 is returned if an invalid token is provided.

        This test doesn't mock SSOOAuthAuthentication.validate_token to
        trigger OAuthError explicitly, to ensure that
        whatever exception gets raised it is caught and handled.
        """
        request = self.make_request_with_token()
        self.mock_validate.return_value = False

        response = self.get_resource(request)

        self.assertEqual(httplib.UNAUTHORIZED, response.status_code)

    def test_user_doesnt_exist_twice(self):
        """Check that an autocreated user persists on for a second call"""
        request = self.make_request_with_token()
        self.mock_get_data.return_value = self.mock_data_for_account()

        self.user.useropenid_set.all().delete()

        create_user = 'django.contrib.auth.models.UserManager.create_user'
        with patch(create_user) as mock_create_user:
            mock_create_user.return_value = self.user

            request = self.make_request_with_token()

            # First request...
            response = self.get_resource(request)
            request = MockRequest()
            request.META.update(self.factory.header_from_token(
                request, consumer_key=self.openid))

            response = self.get_resource(request)

        self.assertEqual(1, mock_create_user.call_count)
        self.assertEqual(httplib.OK, response.status_code)

    def test_user_is_created_with_right_id(self):
        """Check that autocreated users get the right (full) claimed_id"""
        request = self.make_request_with_token()
        data_for_account = self.mock_data_for_account()
        self.mock_get_data.return_value = data_for_account
        self.user.useropenid_set.all().delete()
        self.user.delete()
        new_user = self.factory.makeUser()
        new_user.useropenid_set.all().delete()

        create_user = 'django.contrib.auth.models.UserManager.create_user'
        with patch(create_user) as mock_create_user:
            mock_create_user.return_value = new_user
            self.get_resource(request)

        self.assertEqual(1, new_user.useropenid_set.count())
        openid = new_user.useropenid_set.get()
        expected_id = (settings.OPENID_SSO_SERVER_URL.strip('/') + '/+id/' +
                       self.openid)
        self.assertEqual(expected_id, openid.claimed_id)
        self.assertEqual(expected_id, openid.display_id)

    def test_full_claimed_id(self):
        """Simple test to check that the right claimed_id is calculated"""
        consumer_key = 'abcdef12'

        claimed_id = full_claimed_id(consumer_key)

        expected_id = (settings.OPENID_SSO_SERVER_URL.strip('/') + '/+id/' +
                       consumer_key)
        self.assertEqual(expected_id, claimed_id)

    def test_query_string_auth(self):
        request = self.make_request_with_querystring_token()

        response = self.get_resource(request)

        auth_data = {
            'http_url': request.build_absolute_uri(),
            'http_method': request.method,
            'authorization': request.META.get('HTTP_AUTHORIZATION', None),
            'query_string': request.META.get('QUERY_STRING', None),
        }
        self.mock_validate.assert_called_with(data=auth_data)
        self.assertEqual(httplib.OK, response.status_code)

    @patch('core.api.auth.sso.oauth.OAuthRequest.from_request')
    def test_oauth_request_with_authorization_header(
            self, mock_oauth_from_request):
        request = self.make_request_with_querystring_token()
        request.META.update({'HTTP_AUTHORIZATION': 'someauth'})

        ret = Mock()
        ret.get_parameter.side_effect = ['openid', 'token_key']
        mock_oauth_from_request.return_value = ret

        self.get_resource(request)

        auth_data = {
            'http_url': request.build_absolute_uri(),
            'http_method': request.method,
            'headers': {'Authorization': 'someauth'},
            'query_string': request.META.get('QUERY_STRING', None),
        }

        mock_oauth_from_request.assert_called_with(**auth_data)

    @patch('core.api.auth.sso.oauth.OAuthRequest.from_request')
    def test_oauth_request_with_empty_authorization_header(
            self, mock_oauth_from_request):
        request = self.make_request_with_querystring_token()
        request.META.update({'HTTP_AUTHORIZATION': ''})

        ret = Mock()
        ret.get_parameter.side_effect = ['openid', 'token_key']
        mock_oauth_from_request.return_value = ret

        self.get_resource(request)

        auth_data = {
            'http_url': request.build_absolute_uri(),
            'http_method': request.method,
            'headers': {},
            'query_string': request.META.get('QUERY_STRING', None),
        }

        mock_oauth_from_request.assert_called_with(**auth_data)

    @patch('core.api.auth.sso.oauth.OAuthRequest.from_request')
    def test_oauth_request_sans_authorization_header(
            self, mock_oauth_from_request):
        request = self.make_request_with_querystring_token()
        request.META.update({'HTTP_AUTHORIZATION': None})

        ret = Mock()
        ret.get_parameter.side_effect = ['openid', 'token_key']
        mock_oauth_from_request.return_value = ret

        self.get_resource(request)

        auth_data = {
            'http_url': request.build_absolute_uri(),
            'http_method': request.method,
            'headers': {},
            'query_string': request.META.get('QUERY_STRING', None),
        }

        mock_oauth_from_request.assert_called_with(**auth_data)

    def test_user_is_updated_from_sso_user_data_on_login_after_window(self):
        """Check that the data is updated from SSO."""
        request = self.make_request_with_token()
        request.META.update(self.factory.header_from_token(
            request, consumer_key=self.openid))
        self.mock_get_data.return_value = self.mock_data_for_account(
            displayname='Hazel Jennet', email='new@email.com')

        old_login_time = time.time() - settings.SSO_UPDATE_WINDOW - 10
        self.user.last_login = datetime.datetime.fromtimestamp(old_login_time)
        self.user.save()

        response = self.get_resource(request)

        # The original first and last names have been updated from SSO.
        user = User.objects.get(pk=self.user.pk)
        self.assertEqual('Hazel Jennet', user.get_full_name())
        self.assertEqual('new@email.com', user.email)
        self.assertEqual(httplib.OK, response.status_code)

    def test_existing_user_authenticates_even_with_missing_sso_data(self):
        self.user.first_name = 'Harry'
        self.user.last_name = 'Jones'
        self.user.email = 'old@email.com'
        self.user.save()

        request = self.make_request_with_token()
        request.META.update(self.factory.header_from_token(
            request, consumer_key=self.openid))
        self.mock_get_data.return_value = None

        response = self.get_resource(request)

        user = User.objects.get(pk=self.user.pk)
        self.assertEqual('Harry Jones', user.get_full_name())
        self.assertEqual('old@email.com', user.email)
        self.assertEqual(httplib.OK, response.status_code)

    def test_invalid_user_and_no_sso_data_is_unauthorized(self):
        request = self.make_request_with_querystring_token()
        self.user.delete()
        self.mock_get_data.return_value = None
        response = self.get_resource(request)
        self.assertEqual(httplib.UNAUTHORIZED, response.status_code)


class DelegatedSSOTokenizeAuthorization(TestCase):

    def test_ok(self):
        authorization = (
            'DelegatedSSO signature="abcdef0123456789",timestamp="123456"')
        tokens = delegated_sso.tokenize_authorization(authorization)
        self.assertEqual(list(tokens), ['DelegatedSSO',
                                        ('signature', 'abcdef0123456789'),
                                        ('timestamp', '123456')])

    def test_extra_comma_is_ok(self):
        authorization = (
            'DelegatedSSO signature="abcdef0123456789",timestamp="123456",')
        tokens = delegated_sso.tokenize_authorization(authorization)
        self.assertEqual(list(tokens), ['DelegatedSSO',
                                        ('signature', 'abcdef0123456789'),
                                        ('timestamp', '123456')])

    def test_tokenizer_stops_at_malformed_key_value(self):
        authorization = (
            'DelegatedSSO signature="abcdef0123456789",timestampWRONG"123456"')
        tokens = delegated_sso.tokenize_authorization(authorization)
        self.assertEqual(list(tokens), ['DelegatedSSO',
                                        ('signature', 'abcdef0123456789')])

    def test_tokenizer_stops_at_missing_comma(self):
        authorization = (
            'DelegatedSSO signature="abcdef0123456789" timestamp="123456"')
        tokens = delegated_sso.tokenize_authorization(authorization)
        self.assertEqual(list(tokens), ['DelegatedSSO',
                                        ('signature', 'abcdef0123456789')])

    def test_whitespace_is_ok(self):
        expected = ['DelegatedSSO',
                    ('signature', 'abcdef0123456789'),
                    ('timestamp', '123456')]
        authorization_a = (
            'DelegatedSSO signature="abcdef0123456789", timestamp="123456"')
        authorization_b = (
            'DelegatedSSO signature="abcdef0123456789",\ttimestamp="123456"')
        tokens_a = list(delegated_sso.tokenize_authorization(authorization_a))
        tokens_b = list(delegated_sso.tokenize_authorization(authorization_b))
        self.assertEqual(tokens_a, expected)
        self.assertEqual(tokens_b, expected)


class DelegatedSSOParseAuthorization(TestCase):

    def test_ok(self):
        expected = ('abcdef0123456789', '123456.1')
        tokens = [('signature', 'abcdef0123456789'),
                  ('timestamp', '123456.1')]
        parsed = delegated_sso.parse_authorization(tokens)
        self.assertEqual(parsed, expected)

    def test_extra_keys_are_ignored(self):
        expected = ('abcdef0123456789', '123456.1')
        tokens = [('signature', 'abcdef0123456789'),
                  ('timestamp', '123456.1'),
                  ('dummy', 'value')]
        parsed = delegated_sso.parse_authorization(tokens)
        self.assertEqual(parsed, expected)

    def test_missing_keys_are_none(self):
        expected = (None, None)
        parsed = delegated_sso.parse_authorization([])
        self.assertEqual(parsed, expected)


class DelegatedSSOIsExpired(TestCase):

    def setUp(self):
        super(DelegatedSSOIsExpired, self).setUp()
        self.time = 1411667041.559495
        p = patch('core.api.auth.delegated_sso.time.time',
                  return_value=self.time)
        p.start()
        self.addCleanup(p.stop)

    def test_is_expired_upper_limit(self):
        max_age = 60
        timestamp = self.time + max_age + 1
        self.assertTrue(delegated_sso.is_expired(timestamp, max_age))

    def test_is_expired_lower_limit(self):
        max_age = 60
        timestamp = self.time - max_age - 1
        self.assertTrue(delegated_sso.is_expired(timestamp, max_age))

    def test_is_not_expired_upper_limit(self):
        max_age = 60
        timestamp = self.time + (max_age - 1)
        self.assertFalse(delegated_sso.is_expired(timestamp, max_age))

    def test_is_not_expired_lower_limit(self):
        max_age = 60
        timestamp = self.time - (max_age - 1)
        self.assertFalse(delegated_sso.is_expired(timestamp, max_age))


class DelegatedSSOIsSignatureValid(TestCase):

    def setUp(self):
        super(DelegatedSSOIsSignatureValid, self).setUp()
        self.url = 'http://rnr.local/click/api/1.0/endpoint'
        self.timestamp = '123456'
        self.body = '{"key": "value"}'
        self.msg = self.url + self.timestamp + self.body
        self.secret = 's3cr37'
        self.signature = hmac.new(self.secret, self.msg, sha256).hexdigest()

    def test_is_signature_valid(self):
        is_valid = delegated_sso.is_signature_valid(
            self.signature, self.secret, self.url, self.timestamp, self.body)
        self.assertTrue(is_valid)

    def test_is_signature_not_valid_message(self):
        signature = hmac.new(self.secret, 'wrongmessage', sha256).hexdigest()
        is_valid = delegated_sso.is_signature_valid(
            signature, self.secret, self.url, self.timestamp, self.body)
        self.assertFalse(is_valid)

    def test_is_signature_not_valid_secret(self):
        is_valid = delegated_sso.is_signature_valid(
            self.signature, 'wrongsecret', self.url, self.timestamp, self.body)
        self.assertFalse(is_valid)

    def test_is_signature_not_valid_url(self):
        is_valid = delegated_sso.is_signature_valid(
            self.signature, self.secret, 'wrong', self.timestamp, self.body)
        self.assertFalse(is_valid)

    def test_is_signature_not_valid_timestamp(self):
        is_valid = delegated_sso.is_signature_valid(
            self.signature, self.secret, self.url, '567890', self.body)
        self.assertFalse(is_valid)

    def test_is_signature_not_valid_body(self):
        is_valid = delegated_sso.is_signature_valid(
            self.signature, self.secret, self.url, self.timestamp, 'wrongbody')
        self.assertFalse(is_valid)


class DelegatedSSOAuthenticationTestCase(BaseAuthenticationTestCase):

    def _get_auth(self):
        return DelegatedSSOAuthentication(secret=self.secret)

    def setUp(self):
        self.timestamp = time.time()
        self.url = 'http://rnr.local/click/api/1.0/endpoint'
        self.body = '{"key": "value"}'
        self.msg = "%s%s%s" % (self.url, self.timestamp, self.body)
        self.secret = 's3cr37'
        self.signature = hmac.new(self.secret, self.msg, sha256).hexdigest()
        super(DelegatedSSOAuthenticationTestCase, self).setUp()

    def make_request(self, method='GET', secret=None, url=None, signature=None,
                     timestamp=None, body=None, mechanism=None):
        request = MockRequest(method=method)

        if secret is None:
            secret = self.secret

        if url is None:
            url = self.url

        if timestamp is None:
            timestamp = self.timestamp

        if method.lower() == 'get':
            body = ''
        else:
            body = self.body if body is None else body

        request.body = body
        request.build_absolute_uri = lambda: url

        if signature is None:
            msg = "%s%s%s" % (url, timestamp, body)
            signature = hmac.new(secret, msg, sha256).hexdigest()

        if mechanism is None:
            mechanism = delegated_sso.IDENTIFIER

        auth = '%s signature="%s",timestamp="%s"' % (
            mechanism, signature, timestamp)

        request.META['HTTP_AUTHORIZATION'] = auth
        return request

    def test_success(self):
        request = self.make_request()
        response = self.get_resource(request)
        self.assertEqual(httplib.OK, response.status_code)

    def test_success_with_body(self):
        request = self.make_request(method='POST')
        response = self.post_resource(request)
        self.assertEqual(httplib.CREATED, response.status_code)

    def test_no_auth_fails(self):
        request = MockRequest()
        assert not request.META.get('HTTP_AUTHORIZATION')
        assert not request.META.get('QUERY_STRING')
        response = self.get_resource(request)
        self.assertEqual(httplib.UNAUTHORIZED, response.status_code)

    def test_good_data_with_wrong_mechanism_identifier_fails(self):
        request = self.make_request(mechanism='garbage')
        response = self.get_resource(request)
        self.assertEqual(httplib.UNAUTHORIZED, response.status_code)

    def test_wrong_timestamp_fails(self):
        request = self.make_request(signature=self.signature, timestamp=123456)
        response = self.get_resource(request)
        self.assertEqual(httplib.UNAUTHORIZED, response.status_code)

    def test_future_timestamp_fails(self):
        timestamp = time.time() + delegated_sso.MAX_AGE + 10
        request = self.make_request(timestamp=timestamp)
        response = self.get_resource(request)
        self.assertEqual(httplib.UNAUTHORIZED, response.status_code)

    def test_expired_timestamp_fails(self):
        timestamp = time.time() - delegated_sso.MAX_AGE - 10
        request = self.make_request(timestamp=timestamp)
        response = self.get_resource(request)
        self.assertEqual(httplib.UNAUTHORIZED, response.status_code)

    def test_wrong_body_fails(self):
        request = self.make_request('POST', self.signature, body='wrong')
        response = self.post_resource(request)
        self.assertEqual(httplib.UNAUTHORIZED, response.status_code)
