from django.conf import settings
from django.http import HttpResponse
from django.test import TestCase
from django.utils.cache import patch_vary_headers

from mock import patch

from core.api.decorators import (
    dont_vary,
    gzip_content,
    vary_only_on_accept,
    vary_only_on_headers,
)


class GzipContentTestCase(TestCase):
    """Test the gzip_content decorator."""

    def setUp(self):
        patcher = patch('core.api.decorators.gzip_page')
        self.mock_gzip = patcher.start()
        self.addCleanup(patcher.stop)

    @patch.multiple(settings, ENABLE_GZIP_CONTENT=True)
    def test_called_when_setting_enabled(self):
        """Test gzip is used when ENABLE_GZIP_CONTENT is True."""

        def test_func():
            return "response"

        decorated = gzip_content(test_func)
        decorated()

        self.mock_gzip.assert_called_once_with(test_func)

    @patch.multiple(settings, ENABLE_GZIP_CONTENT=False)
    def test_called_when_setting_disabled(self):
        """Test gzip is not used when ENABLE_GZIP_CONTENT is False."""

        @gzip_content
        def test_func():
            return "response"

        test_func()

        self.assertFalse(self.mock_gzip.called)

    @patch.multiple(settings, ENABLE_GZIP_CONTENT=True)
    def test_decorated_function_called_with_args_when_enabled(self):
        """Test gzip decorated function is called with all args."""

        def test_func(*args, **kwargs):
            return args, kwargs

        self.mock_gzip.return_value = test_func

        decorated = gzip_content(test_func)

        self.assertEqual(decorated(1, 2, a=3, b=4),
                         ((1, 2), {'a': 3, 'b': 4}))

    @patch.multiple(settings, ENABLE_GZIP_CONTENT=False)
    def test_decorated_function_called_with_args_when_disabled(self):
        """Test gzip decorated function is called with all args."""

        def test_func(*args, **kwargs):
            return args, kwargs

        decorated = gzip_content(test_func)

        self.assertEqual(decorated(1, 2, a=3, b=4),
                         ((1, 2), {'a': 3, 'b': 4}))


class DontVaryTestCase(TestCase):
    """Test the dont_vary decorator."""

    def test_strips_vary_headers_from_response(self):

        def test_func(request):
            response = HttpResponse("response")
            patch_vary_headers(response, ['Authorization', 'Accept'])
            return response

        decorated = dont_vary(test_func)

        response = test_func({})
        self.assertTrue(response.has_header('vary'))

        decorated_response = decorated({})
        self.assertFalse(decorated_response.has_header('vary'))


class VaryOnlyOnHeadersTestCase(TestCase):
    """Test the vary_only_on_headers decorator."""

    def test_adds_vary_headers_even_if_response_had_none(self):

        def test_func(request):
            response = HttpResponse("response")
            return response

        decorated = vary_only_on_headers(
            'Accept', 'Accept-Encoding')(test_func)

        response = test_func({})
        self.assertFalse(response.has_header('vary'))

        decorated_response = decorated({})
        self.assertTrue(decorated_response.has_header('vary'))
        self.assertEqual(decorated_response['vary'], 'Accept, Accept-Encoding')

    def test_strips_vary_headers_from_response(self):

        def test_func(request):
            response = HttpResponse("response")
            patch_vary_headers(response, ['Authorization', 'Cookie'])
            return response

        decorated = vary_only_on_headers()(test_func)

        response = test_func({})
        self.assertTrue(response.has_header('vary'))
        self.assertEqual(response['vary'], 'Authorization, Cookie')

        decorated_response = decorated({})
        self.assertFalse(decorated_response.has_header('vary'))

    def test_replaces_original_vary_headers_from_response(self):

        def test_func(request):
            response = HttpResponse("response")
            patch_vary_headers(response, ['Authorization', 'Cookie'])
            return response

        decorated = vary_only_on_headers(
            'Accept', 'Accept-Encoding')(test_func)

        response = test_func({})
        self.assertTrue(response.has_header('vary'))
        self.assertEqual(response['vary'], 'Authorization, Cookie')

        decorated_response = decorated({})
        self.assertTrue(decorated_response.has_header('vary'))
        self.assertEqual(decorated_response['vary'], 'Accept, Accept-Encoding')


class VaryOnlyOnAcceptTestCase(TestCase):
    """Test the vary_only_on_accept decorator."""

    def test_adds_vary_headers_even_if_response_had_none(self):

        def test_func(request):
            response = HttpResponse("response")
            return response

        decorated = vary_only_on_accept(test_func)

        response = test_func({})
        self.assertFalse(response.has_header('vary'))

        decorated_response = decorated({})
        self.assertTrue(decorated_response.has_header('vary'))
        self.assertEqual(decorated_response['vary'], 'Accept')

    def test_replaces_original_vary_headers_from_response(self):

        def test_func(request):
            response = HttpResponse("response")
            patch_vary_headers(response, ['Authorization', 'Cookie'])
            return response

        decorated = vary_only_on_accept(test_func)

        response = test_func({})
        self.assertTrue(response.has_header('vary'))
        self.assertEqual(response['vary'], 'Authorization, Cookie')

        decorated_response = decorated({})
        self.assertTrue(decorated_response.has_header('vary'))
        self.assertEqual(decorated_response['vary'], 'Accept')
