import urlparse

from django.test import TestCase
from mock import Mock

from core.api.pagination import make_paginated_response


class MakePaginatedResponseTestCase(TestCase):

    maxDiff = None

    def setUp(self):
        super(MakePaginatedResponseTestCase, self).setUp()
        self.request = self.make_request()

    def make_request(self, page_size=None, page=None, uri=None):
        request = Mock()
        request.GET = {}

        if page_size is not None:
            request.GET['page_size'] = page_size

        if page is not None:
            request.GET['page'] = page

        if uri is None:
            uri = 'http://someserver.org/api/objects/'

        request.build_absolute_uri = lambda location=None: uri
        return request

    def test_first_page(self):
        objects = range(100)
        request = self.make_request(page=1)
        response = make_paginated_response(request, objects)
        self.assertEqual(response, {
            'previous': None,
            'next': self.request.build_absolute_uri() + '?page=2',
            'count': 20,
            'results': range(20)})

    def test_no_page_gives_first_page(self):
        objects = range(100)
        response = make_paginated_response(self.request, objects)
        self.assertEqual(response, {
            'previous': None,
            'next': self.request.build_absolute_uri() + '?page=2',
            'count': 20,
            'results': range(20)})

    def test_next_page(self):
        objects = range(100)
        request = self.make_request(page=2)
        response = make_paginated_response(request, objects)
        self.assertEqual(response, {
            'previous': self.request.build_absolute_uri() + '?page=1',
            'next': self.request.build_absolute_uri() + '?page=3',
            'count': 20,
            'results': range(20, 40)})

    def test_last_page(self):
        objects = range(21)
        request = self.make_request(page=2)
        response = make_paginated_response(request, objects)
        self.assertEqual(response, {
            'previous': self.request.build_absolute_uri() + '?page=1',
            'next': None,
            'count': 1,
            'results': [20]})

    def test_unlimited_results(self):
        objects = range(1000)
        request = self.make_request(page_size=2000)
        response = make_paginated_response(
            request, objects, max_page_size=None)
        self.assertEqual(response, {
            'previous': None,
            'next': None,
            'count': 1000,
            'results': objects})

    def test_max_page_size(self):
        objects = range(1000)
        request = self.make_request(page_size=2000)
        response = make_paginated_response(
            request, objects, max_page_size=20)
        self.assertEqual(response, {
            'previous': None,
            'next': request.build_absolute_uri() + '?page=2',
            'count': 20,
            'results': range(20)})

    def test_default_page_size(self):
        objects = range(100)
        response = make_paginated_response(
            self.request, objects, page_size=50)
        self.assertEqual(response, {
            'previous': None,
            'next': self.request.build_absolute_uri() + '?page=2',
            'count': 50,
            'results': range(50)})

    def test_invalid_page_uses_first(self):
        objects = range(100)
        request = self.make_request(page='invalid')
        response = make_paginated_response(request, objects)
        self.assertEqual(response, {
            'previous': None,
            'next': request.build_absolute_uri() + '?page=2',
            'count': 20,
            'results': range(20)})

    def test_empty_page_uses_last_page(self):
        objects = range(100)
        # There are only 5 pages of 20 objects each, and user requests
        # the 200th page, she should get the last page instead.
        request = self.make_request(page=200)
        response = make_paginated_response(request, objects)
        self.assertEqual(response, {
            'previous': request.build_absolute_uri() + '?page=4',
            'next': None,
            'count': 20,
            'results': range(80, 100)})

    def test_next_and_previous_links_respect_query_params(self):
        objects = range(100)
        uri = 'http://someserver.org/api/objects/?a=1&b=2&b=3&c=4'
        request = self.make_request(page=2, uri=uri)
        response = make_paginated_response(request, objects)

        self.assertEqual(response['count'], 20)
        self.assertEqual(response['results'], range(20, 40))

        # existing query params must be in next and previous links
        expected_previous_qs = urlparse.parse_qs(
            urlparse.urlsplit(uri + '&page=1').query)
        expected_next_qs = urlparse.parse_qs(
            urlparse.urlsplit(uri + '&page=3').query)
        previous_qs = urlparse.parse_qs(
            urlparse.urlsplit(response['previous']).query)
        next_qs = urlparse.parse_qs(
            urlparse.urlsplit(response['next']).query)
        self.assertEqual(previous_qs, expected_previous_qs)
        self.assertEqual(next_qs, expected_next_qs)
