#   -*- coding: utf-8 -*-
#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010-2011 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the preflight integration."""

from __future__ import (
    absolute_import,
    with_statement,
)

__metaclass__ = type
__all__ = [
    'TestPreflight',
    'TestGetBZRRevno',
]

from mock import patch

from django.contrib.auth.models import Group

from core.preflight import RNRPreflight
from core.tests.factory import TestCaseWithFactory


class TestPreflight(TestCaseWithFactory):

    def _request_preflight(self, authenticated=True, user=None):
        if authenticated:
            if user is None:
                user = self.factory.makeUser()
                group = Group.objects.get(name='rnr-developers')
                user.groups.add(group)
            login = self.client.login(username=user.username, password='test')
            self.assertTrue(login, 'Could not log in')

        with patch('preflight.models.gather_checks') as mock_gather_checks:
            mock_gather_checks.return_value = []
            response = self.client.get('/preflight/')
        return response

    def test_anonymous(self):
        response = self._request_preflight(authenticated=False)

        self.assertEqual(404, response.status_code)

    def test_logged_in_no_group(self):
        user = self.factory.makeUser(username='test')

        response = self._request_preflight(user=user)

        self.assertEqual(404, response.status_code)

    def test_logged_in_wrong_group(self):
        g = Group.objects.create(name='someothergroup')
        user = self.factory.makeUser(username='test')
        user.groups.add(g)

        response = self._request_preflight(user=user)

        self.assertEqual(404, response.status_code)

    def test_success(self):
        response = self._request_preflight()

        self.assertEqual(200, response.status_code)
        libs = set([x['name'] for x in response.context[0]['versions']])
        expected = set([
            'bzr-revno', 'openid', 'Platform', 'Python', 'Django', 'piston',
            'launchpadlib', 'South', 'preflight',
        ])
        self.assertEqual(expected, libs)


class TestGetBZRRevno(TestCaseWithFactory):

    def setUp(self):
        super(TestGetBZRRevno, self).setUp()
        self.preflight = RNRPreflight()

    def test_success(self):
        self.assertIsNotNone(self.preflight.get_bzr_revno())

    def test_no_bzrlib_returns_none(self):
        patch_name = 'bzrlib.workingtree.WorkingTree.open_containing'
        with patch(patch_name) as mock_open_containing:
            mock_open_containing.side_effect = ImportError
            self.assertIsNone(self.preflight.get_bzr_revno())

    def test_no_branch_returns_none(self):
        from bzrlib.errors import NotBranchError
        patch_name = 'bzrlib.workingtree.WorkingTree.open_containing'
        with patch(patch_name) as mock_open_containing:
            mock_open_containing.side_effect = NotBranchError
            self.assertIsNone(self.preflight.get_bzr_revno())

    def test_no_working_tree_returns_none(self):
        from bzrlib.errors import NoWorkingTree
        patch_name = 'bzrlib.workingtree.WorkingTree.open_containing'
        with patch(patch_name) as mock_open_containing:
            mock_open_containing.side_effect = NoWorkingTree
            self.assertIsNone(self.preflight.get_bzr_revno())

    @patch('core.preflight.logger')
    @patch('bzrlib.workingtree.WorkingTree.open_containing')
    def test_bzrlib_working_tree_and_branch_but_error(
            self, mock_open_containing, mock_logger):
        mock_open_containing.side_effect = IOError
        self.assertIsNone(self.preflight.get_bzr_revno())
        mock_logger.exception.assert_called_with(
            "bzrlib and branch available but revno retrieval failed.")
