import os

from django.conf import settings
from django.core.cache import cache
from django.test import TestCase
from django.test.utils import override_settings
from launchpadlib.errors import HTTPError
from mock import patch, Mock
from piston_mini_client import PistonResponseObject

from core.tests.factory import TestCaseWithFactory
from core.utilities import (
    WebServices,
    full_claimed_id,
    update_or_create_user_from_oauth,
    web_services,
)


class SSOClientTestCase(TestCase):

    def setUp(self):
        super(SSOClientTestCase, self).setUp()

        p = patch('core.utilities.web_services._ssoclient')
        self.mock_sso_client = p.start()
        self.addCleanup(p.stop)

        self.mock_validate_request = self.mock_sso_client.validate_request
        self.mock_account_details = self.mock_sso_client.account_details
        self.mock_token_details = self.mock_sso_client.token_details

    def test_validate_request_ok(self):
        self.mock_validate_request.return_value = dict(is_valid=True)
        self.assertTrue(web_services.validate_request(data=dict()))

    def test_validate_request_not_ok(self):
        self.mock_validate_request.return_value = dict(is_valid=False)
        self.assertFalse(web_services.validate_request(data=dict()))

    def test_validate_request_failed(self):
        self.mock_validate_request.side_effect = ValueError()
        self.assertFalse(web_services.validate_request(data=dict()))

    def test_validate_not_json(self):
        self.mock_validate_request.return_value = 'this is not valid json'
        self.assertFalse(web_services.validate_request(data=dict()))

    def test_account_details(self):
        self.mock_account_details.return_value = 'data'

        response = web_services.get_data_for_account('open-id')
        self.mock_account_details.assert_called_with(
            'open-id', token=settings.SSO_V2_API_TOKEN_DATA)
        self.assertEqual(response, 'data')

    def test_account_details_fail(self):
        self.mock_account_details.side_effect = ValueError

        response = web_services.get_data_for_account('open-id')
        self.mock_account_details.assert_called_with(
            'open-id', token=settings.SSO_V2_API_TOKEN_DATA)
        self.assertEqual(response, None)


class VerifyPackageTestCase(TestCase):
    def setUp(self):
        super(VerifyPackageTestCase, self).setUp()
        # Force web_services to recreate the the Launchpad service root
        WebServices._launchpad_service = None

    def _make_mock_launchpad(self, binary_name='mypkg'):
        mock_ubuntu = Mock()
        mock_ubuntu.self_link = 'http://lpapi/ubuntu'

        mock_launchpad = Mock()
        mock_launchpad.distributions = dict(ubuntu=mock_ubuntu)
        app_review_board = Mock()
        mock_launchpad.people = {'app-review-board': app_review_board}

        mock_archive = Mock()
        mock_ubuntu.getArchive.return_value = mock_archive
        app_review_board.getPPAByName.return_value = mock_archive

        mock_binary = Mock()
        mock_binary.binary_package_name = binary_name
        mock_archive.getPublishedBinaries.return_value = [mock_binary]
        return (mock_launchpad, mock_ubuntu, mock_archive)

    def _verify_package(
            self, mock_lp_api, package_name, distro, series, arch_tag):
        with patch('core.utilities.Launchpad') as mock_launchpad:
            mock_launchpad.login_anonymously.return_value = mock_lp_api
            return WebServices().verify_packagename_in_repository(
                package_name, distro, series, arch_tag)

    def test_package_in_ubuntu(self):
        lp_api, ubuntu, archive = self._make_mock_launchpad()

        result = self._verify_package(
            lp_api, 'mypkg', 'ubuntu', 'lucid', 'i386')

        ubuntu.getArchive.assert_called_with(name='primary')
        archive.getPublishedBinaries.assert_called_with(
            binary_name='mypkg',
            distro_arch_series='http://lpapi/ubuntu/lucid/i386',
            status='Published', exact_match=True)
        self.assertTrue(result)

    def test_package_not_in_ubuntu(self):
        lp_api, ubuntu, archive = self._make_mock_launchpad(
            binary_name='other_pkg')

        result = self._verify_package(
            lp_api, 'mypkg', 'ubuntu', 'lucid', 'i386')

        ubuntu.getArchive.assert_called_with(name='primary')
        archive.getPublishedBinaries.assert_called_with(
            binary_name='mypkg',
            distro_arch_series='http://lpapi/ubuntu/lucid/i386',
            status='Published', exact_match=True)
        self.assertFalse(result)

    def test_package_in_partner(self):
        lp_api, ubuntu, archive = self._make_mock_launchpad()

        result = self._verify_package(
            lp_api, 'mypkg', 'canonical', 'lucid', 'i386')

        ubuntu.getArchive.assert_called_with(name='partner')
        archive.getPublishedBinaries.assert_called_with(
            binary_name='mypkg',
            distro_arch_series='http://lpapi/ubuntu/lucid/i386',
            status='Published', exact_match=True)
        self.assertTrue(result)

    def test_package_in_extras(self):
        lp_api, ubuntu, archive = self._make_mock_launchpad()

        result = self._verify_package(
            lp_api, 'mypkg', 'lp-ppa-app-review-board', 'lucid', 'i386')

        app_review_board = lp_api.people['app-review-board']
        app_review_board.getPPAByName.assert_called_with(name='ppa')
        archive.getPublishedBinaries.assert_called_with(
            binary_name='mypkg',
            distro_arch_series='http://lpapi/ubuntu/lucid/i386',
            status='Published', exact_match=True)
        self.assertTrue(result)

    def test_unknown_origin(self):
        lp_api, ubuntu, archive = self._make_mock_launchpad()

        result = self._verify_package(
            lp_api, 'mypkg', 'lp-ppa-unknown', 'lucid', 'i386')

        # Currently we only support primary, partner, extras and the
        # specified commercial origins.
        self.assertEqual(0, archive.getPublishedBinaries.call_count)
        self.assertFalse(result)

    def test_invalid_series_or_arch(self):
        # An invalid series or arch raises a 400 Bad Request.
        lp_api, ubuntu, archive = self._make_mock_launchpad()
        response = Mock()
        response.status = 400
        archive.getPublishedBinaries.side_effect = HTTPError(
            response, 'content')

        result = self._verify_package(
            lp_api, 'mypkg', 'ubuntu', 'lucid', 'i386')

        self.assertFalse(result)

    @patch('core.utilities.Launchpad')
    def test_doesnt_sign_in_each_time(self, mock_launchpad):
        lp_api, ubuntu, archive = self._make_mock_launchpad()
        mock_launchpad.login_anonymously.return_value = lp_api
        ws = WebServices()
        ws.verify_packagename_in_repository("foobar", 'ubuntu',
                                            "lucid", 'i386')
        self.assertTrue(mock_launchpad.login_anonymously.called)
        mock_launchpad.login_anonymously.reset_mock()

        ws.verify_packagename_in_repository("foobar", 'ubuntu',
                                            "lucid", 'i386')

        self.assertFalse(mock_launchpad.login_anonymously.called)

    # Patch at a lower level to ensure that the request will be made
    @patch('lazr.restfulclient._browser.Browser._request')
    def test_doesnt_fail_with_home_undefined(self, mock_request):
        mock_request.return_value = {}, ''
        if 'HOME' in os.environ:
            del os.environ['HOME']
        ws = WebServices()
        # SyntaxError is raised when we fail to parse our dummy wadl
        self.assertRaises(
            SyntaxError, ws.verify_packagename_in_repository,
            "foobar", "ubuntu", "lucid", 'i386')


class SCATestCase(TestCase):

    def get_test_file(self, file_name):
        return os.path.join(os.path.dirname(__file__), 'test_data',
                            file_name)

    @patch('httplib2.Http.request')
    def _request_sca_pubs_for(self, distroseries, arch_tag,
                              disable_cache, mock_request):
        data_file = self.get_test_file('software_center_apps.json')
        with open(data_file) as test_data:
            data = test_data.read()
            mock_request.return_value = ({'status': '200'}, data)

        if disable_cache:
            func = 'django.core.cache.cache.get'
            with patch(func) as mock_get:
                mock_get.return_value = False
                published_apps = WebServices().get_sca_publishings_for(
                    distroseries, arch_tag)
        else:
            published_apps = WebServices().get_sca_publishings_for(
                distroseries, arch_tag)
        return published_apps, mock_request

    def test_get_sca_publishings(self):
        published_apps, _ = self._request_sca_pubs_for(
            'maverick', 'i386', True)

        origins = [
            'lp-ppa-commercial-ppa-uploaders-brukkon',
            'lp-ppa-commercial-ppa-uploaders-fluendo-dvd',
            'lp-ppa-commercial-ppa-uploaders-fluendo-plugins',
            'lp-ppa-commercial-ppa-uploaders-fluendo-wmv-plugins',
            'lp-ppa-commercial-ppa-uploaders-illumination',
            'lp-ppa-commercial-ppa-uploaders-vendetta-online',
            'lp-ppa-commercial-ppa-uploaders-world-of-goo',
            'canonical',
            'lp-ppa-app-review-board',
        ]
        apps = [
            'brukkon', 'fluendo-dvd', 'gstreamer0.10-fluendo-plugins',
            'gstreamer0.10-fluendo-plugins-wmv', 'illumination',
            'vendetta-online', 'worldofgoo', 'acroread', 'qreator',
        ]
        expected = dict((k, set([v])) for k, v in zip(origins, apps))
        self.assertEqual(expected, published_apps)

    def test_retrieved_url(self):
        with override_settings(SCA_HOST_URL='http://example.com/'):
            ignored, mock_request = self._request_sca_pubs_for(
                'natty', 'amd64', True)

        mock_request.assert_called_with(
            'http://example.com/api/2.0/applications/any/ubuntu/any/any/',
            body='', headers={}, method='GET')

    def test_get_sca_publishings_cached(self):
        cache.clear()
        published_apps, mock_request = self._request_sca_pubs_for(
            'maverick', 'i386', False)
        self.assertEqual(1, mock_request.call_count)

        published_apps, mock_request = self._request_sca_pubs_for(
            'natty', 'i386', False)

        self.assertEqual(0, mock_request.call_count)
        apps = cache.get('sca_all_available_set')
        self.assertEqual(9, len(apps))
        for app in apps:
            self.assertTrue(isinstance(app, PistonResponseObject))

    def _request_sca_verify(self, pkgname, origin, distroseries, arch_tag):
        filename = self.get_test_file('software_center_apps.json')
        data = open(filename).read()
        with patch('httplib2.Http.request') as mock_request:
            mock_request.return_value = ({'status': 200}, data)
            ws = WebServices()
            return ws.sca_verify_packagename_in_repository(
                pkgname, origin, distroseries, arch_tag)

    def test_sca_verify_packagename(self):
        self.assertTrue(self._request_sca_verify(
            'brukkon', 'lp-ppa-commercial-ppa-uploaders-brukkon',
            'maverick', 'i386'))
        self.assertFalse(self._request_sca_verify(
            'fluendo-dvd', 'lp-ppa-commercial-ppa-uploaders-brukkon',
            'maverick', 'i386'))
        self.assertFalse(self._request_sca_verify(
            'brukkon', 'lp-ppa-commercial-ppa-uploaders-doesnt-exist',
            'maverick', 'i386'))


class FullClaimedIdTestCase(TestCase):

    def test_full_claimed_id(self):
        """Simple test to check that the right claimed_id is calculated"""
        consumer_key = 'abcdef12'

        claimed_id = full_claimed_id(consumer_key)

        self.assertTrue(claimed_id.startswith(settings.OPENID_SSO_SERVER_URL))
        self.assertTrue(claimed_id.endswith(consumer_key))


class UpdateOrCreateUserFromOauthTestCase(TestCaseWithFactory):

    def setUp(self):
        super(UpdateOrCreateUserFromOauthTestCase, self).setUp()
        self.user = self.factory.makeUser()
        p = patch('core.utilities.web_services')
        self.mock_web_services = p.start()
        self.addCleanup(p.stop)

        self.mock_get_data = self.mock_web_services.get_data_for_account
        self.mock_get_data.return_value = self.mock_data_for_account()

        web_services.refresh_connections()

    def mock_data_for_account(self, **kwargs):
        """A reasonable return value for get_data_for_account."""
        data = {
            'email': 'foo@bar.com',
            'verified': True,
            'displayname': 'Foo',
        }
        data.update(**kwargs)
        return data

    def assert_user_data(self, user, sso_user_data):
        first_name, _, last_name = sso_user_data['displayname'].rpartition(' ')
        email = sso_user_data['email']
        self.assertEqual(user.first_name, first_name)
        self.assertEqual(user.last_name, last_name)
        self.assertEqual(user.email, email)

    def test_missing_account_data_and_missing_user(self):
        self.mock_get_data.return_value = None
        user, created = update_or_create_user_from_oauth('inexistent')
        self.assertIsNone(user)
        self.assertFalse(created)

    def test_account_data_with_no_email_and_missing_user(self):
        sso_user_data = self.mock_data_for_account(email='')
        self.mock_get_data.return_value = sso_user_data
        user, created = update_or_create_user_from_oauth('inexistent')
        self.assertIsNone(user)
        self.assertFalse(created)

    def test_account_data_with_no_displayname_and_missing_user(self):
        sso_user_data = self.mock_data_for_account(displayname=None)
        self.mock_get_data.return_value = sso_user_data
        user, created = update_or_create_user_from_oauth('inexistent')
        self.assertIsNone(user)
        self.assertFalse(created)

    def test_account_data_unverified_and_missing_user(self):
        sso_user_data = self.mock_data_for_account(verified=False)
        self.mock_get_data.return_value = sso_user_data
        user, created = update_or_create_user_from_oauth('inexistent')
        self.assertIsNone(user)
        self.assertFalse(created)

    def test_account_data_and_missing_user(self):
        self.user.delete()
        sso_user_data = self.mock_data_for_account()
        self.mock_get_data.return_value = sso_user_data
        user, created = update_or_create_user_from_oauth('inexistent')
        self.assertIsNotNone(user)
        self.assertTrue(created)
        self.assert_user_data(user, sso_user_data)

    def test_account_data_and_existing_user(self):
        sso_user_data = self.mock_data_for_account(
            displayname='John Does', email='user@email.com')
        self.mock_get_data.return_value = sso_user_data
        open_id = 'existent'
        self.factory.makeUser(
            first_name='John', last_name='Does', email='user@email.com',
            open_id=open_id)
        user, created = update_or_create_user_from_oauth(open_id)
        self.assertIsNotNone(user)
        self.assertFalse(created)
        self.assert_user_data(user, sso_user_data)

    def test_missing_account_data_and_existing_user(self):
        self.mock_get_data.return_value = None
        open_id = 'existent'
        self.factory.makeUser(
            first_name='John', last_name='Does', email='user@email.com',
            open_id=open_id)
        user, created = update_or_create_user_from_oauth(open_id)
        self.assertIsNotNone(user)
        self.assertFalse(created)

    def test_account_data_updates_user_data_with_no_window(self):
        sso_user_data = self.mock_data_for_account(
            displayname='Juan Hace', email='juan@email.com')
        self.mock_get_data.return_value = sso_user_data
        open_id = 'existent'
        self.factory.makeUser(
            first_name='John', last_name='Does', email='user@email.com',
            open_id=open_id)
        user, created = update_or_create_user_from_oauth(
            open_id, update_window=None)
        self.assertIsNotNone(user)
        self.assertFalse(created)
        self.assert_user_data(user, sso_user_data)

    def test_account_data_is_not_updated_within_window(self):
        sso_user_data = self.mock_data_for_account(
            displayname='Juan Hace', email='juan@email.com')
        self.mock_get_data.return_value = sso_user_data
        open_id = 'existent'
        user = self.factory.makeUser(
            first_name='John', last_name='Does', email='user@email.com',
            open_id=open_id)
        logged_user, created = update_or_create_user_from_oauth(
            open_id, update_window=60 * 60 * 24)
        self.assertIsNotNone(logged_user)
        self.assertFalse(created)
        current_sso_user_data = self.mock_data_for_account(
            displayname=user.get_full_name(), email=user.email)
        self.assert_user_data(logged_user, current_sso_user_data)

    def test_unverified_account_data_doesnt_updates_user_data(self):
        sso_user_data = self.mock_data_for_account(
            verified=False, displayname='Juan Hace', email='juan@email.com')
        self.mock_get_data.return_value = sso_user_data
        open_id = 'existent'
        self.factory.makeUser(
            first_name='John', last_name='Does', email='user@email.com',
            open_id=open_id)
        user, created = update_or_create_user_from_oauth(
            open_id, update_window=None)
        self.assertIsNotNone(user)
        self.assertFalse(created)
        self.assertEqual(user.get_full_name(), 'John Does')
        self.assertEqual(user.email, 'user@email.com')

    def test_account_data_with_missing_email_doesnt_updates_user_data(self):
        sso_user_data = self.mock_data_for_account(
            displayname='Juan Hace', email='')
        self.mock_get_data.return_value = sso_user_data
        open_id = 'existent'
        self.factory.makeUser(
            first_name='John', last_name='Does', email='user@email.com',
            open_id=open_id)
        user, created = update_or_create_user_from_oauth(
            open_id, update_window=None)
        self.assertIsNotNone(user)
        self.assertFalse(created)
        self.assertEqual(user.get_full_name(), 'John Does')
        self.assertEqual(user.email, 'user@email.com')

    def test_account_data_with_missing_name_doesnt_updates_user_data(self):
        sso_user_data = self.mock_data_for_account(
            displayname=None, email='new@email.com')
        self.mock_get_data.return_value = sso_user_data
        open_id = 'existent'
        self.factory.makeUser(
            first_name='John', last_name='Does', email='user@email.com',
            open_id=open_id)
        user, created = update_or_create_user_from_oauth(
            open_id, update_window=None)
        self.assertIsNotNone(user)
        self.assertFalse(created)
        self.assertEqual(user.get_full_name(), 'John Does')
        self.assertEqual(user.email, 'user@email.com')
