import sys
import time

import logging
from httplib2 import ServerNotFoundError

from django.conf import settings
from django.contrib.auth.models import User
from django.core.cache import cache
from django.db import IntegrityError, transaction
from django.utils import six, timezone
from launchpadlib.launchpad import Launchpad
from launchpadlib.uris import lookup_service_root
from lazr.restfulclient.errors import HTTPError
from scaclient import SoftwareCenterAgentAPI
from ssoclient.v2.client import V2ApiClient


__all__ = [
    'WebServices',
    'WebServiceError',
    'full_claimed_id',
]


logger = logging.getLogger(__name__)


class WebServiceError(Exception):
    """Raised when we cannot connect to a web service."""


class WebServices:
    _launchpad_service = None
    _ssoclient = None

    def __init__(self):
        """Pre-create the identity provider webservice if it will be used."""
        if settings.PRELOAD_API_SERVICE_ROOTS:
            self._set_ssoclient()
            self._set_launchpad_service()

    def refresh_connections(self):
        """Ensure all http connections are dropped and restarted."""
        self._launchpad_service = None
        self._ssoclient = None

    @property
    def logger(self):
        return logging.getLogger('rnr.utilities')

    def _set_launchpad_service(self):
        service_root = lookup_service_root(settings.LP_SERVICE)
        launchpadlib_dir = settings.LAUNCHPADLIB_DIR
        try:
            WebServices._launchpad_service = Launchpad.login_anonymously(
                settings.LP_CONSUMER_NAME, service_root,
                launchpadlib_dir=launchpadlib_dir)
        except (HTTPError, ServerNotFoundError):
            self.logger.exception('Unable to create Launchpad API.')
            WebServices._launchpad_service = None

    def _set_ssoclient(self):
        self._ssoclient = V2ApiClient(settings.SSO_API_SERVICE_ROOT)

    @property
    def launchpad_service(self):
        if self._launchpad_service is None:
            self._set_launchpad_service()

        if self._launchpad_service is None:
            raise WebServiceError("The Launchpad service is not available.")

        return self._launchpad_service

    @property
    def ssoclient(self):
        if self._ssoclient is None:
            self._set_ssoclient()
        return self._ssoclient

    def validate_request(self, data):
        try:
            response = self.ssoclient.validate_request(data=data)
            result = response.get('is_valid', False)
        except (AttributeError, TypeError, KeyError, ValueError):
            result = False
        return result

    def get_data_for_account(self, openid_identifier):
        """Return SSO account details for the given openid_identifier."""
        data = None

        try:
            token = settings.SSO_V2_API_TOKEN_DATA
            data = self.ssoclient.account_details(
                openid_identifier, token=token)
        except:
            self.logger.exception('Cannot retrieve user data.')

        return data

    def verify_packagename_in_repository(
            self, pkgname, origin, distroseries, arch_tag):
        """Verify that a given package exists in a distroseries on LP.

        :param pkgname: The name of the package to check.
        :param origin: The repository origin identifier.
        :param distroseries: The name of distribution series.
        :param arch_tag: The machine architecture (i386, amd64...)
        :return: Whether the package was found or not.
        :rtype: bool
        """
        ubuntu = self.launchpad_service.distributions['ubuntu']
        if origin == 'ubuntu':
            archive = ubuntu.getArchive(name='primary')
        elif origin == 'canonical':
            archive = ubuntu.getArchive(name='partner')
        elif origin == 'lp-ppa-app-review-board':
            app_review_board = self.launchpad_service.people[
                'app-review-board']
            archive = app_review_board.getPPAByName(name='ppa')
        else:
            return False

        # We now have the archive, check for the distroseries and the
        # package.
        distro_arch_series = ubuntu.self_link + '/{0}/{1}'.format(
            distroseries, arch_tag)
        try:
            binaries = archive.getPublishedBinaries(
                binary_name=pkgname, distro_arch_series=distro_arch_series,
                status='Published', exact_match=True)
        except HTTPError as error:
            if error.response.status == 400:
                # A bad request is raised if the distro arch series we
                # indicate does not exist.  Obviously, this does not
                # verify!  Any other status is unexpected.
                return False
            raise
        for binary in binaries:
            if binary.binary_package_name == pkgname:
                return True
        return False

    def sca_verify_packagename_in_repository(self, pkgname, origin,
                                             distroseries, arch_tag):
        """Verify that a given package exists in a distroseries on SCA.

        :param pkgname: The name of the package to check.
        :param origin: The repository origin identifier.
        :param distroseries: The name of distribution series.
        :param arch_tag: The machine architecture (i386, amd64...)
        :return: Whether the package was found or not.
        :rtype: bool
        """
        sca_publishings = self.get_sca_publishings_for(distroseries, arch_tag)
        if origin in sca_publishings:
            # XXX 2011-03-11 michaeln bug=733170 PPA origins are
            # currently ambiguous.
            return pkgname in sca_publishings[origin]

        return False

    def get_sca_publishings_for(self, distroseries, arch_tag):
        cache_key = 'sca_all_available_set'
        apps = cache.get(cache_key)

        if not apps:
            api = SoftwareCenterAgentAPI(service_root="{0}api/2.0/"
                                         .format(settings.SCA_HOST_URL))
            apps = api.available_apps(lang='any', series='any', arch='any')
            # achuni 2012-06-12: Cache the unfiltered list so that we don't
            #                    need to store one list per distroseries.
            cache.set(cache_key, apps, settings.SCA_APPS_CACHE_TIMEOUT)

        pubs = {}
        for app in apps:
            if arch_tag not in app.series.get(distroseries, []):
                continue
            origin = 'ubuntu'
            if app.archive_root in settings.NON_PPA_ARCHIVE_ROOTS:
                origin = settings.NON_PPA_ARCHIVE_ROOTS[app.archive_root]
            elif app.archive_root:
                origin = 'lp-ppa-' + app.archive_id.replace('/', '-')
            if origin not in pubs:
                pubs[origin] = set()
            pubs[origin].add(app.package_name)

        return pubs


web_services = WebServices()


def full_claimed_id(consumer_key):
    return '%s/+id/%s' % (settings.OPENID_SSO_SERVER_URL.strip('/'),
                          consumer_key)


def _get_available_username(username):
    """Return a valid and available username."""
    # follow similar approach to django_openid_auth and web login
    i = User.objects.filter(username__startswith=username).count() + 1
    while True:
        if i > 1:
            username += str(i)
        try:
            User.objects.get(username__exact=username)
        except User.DoesNotExist:
            break
        i += 1
    return username


def _get_sso_account_data(account):
    is_valid = (account is not None and
                account.get('verified') and
                account.get('email') and
                account.get('displayname') is not None)

    if is_valid:
        email = account['email']
        first_name, _, last_name = account['displayname'].rpartition(" ")
        return email, first_name, last_name


@transaction.commit_on_success
def _get_or_create_user_from_oauth(consumer_key):
    claimed_id = full_claimed_id(consumer_key)
    created = False

    try:
        user = User.objects.get(useropenid__claimed_id=claimed_id)
    except User.DoesNotExist:
        user = None

    if user is None:
        account = web_services.get_data_for_account(
            openid_identifier=consumer_key)
        account_data = _get_sso_account_data(account)

        if account_data is None:
            logger.warn(
                'OAuthRequest was valid but could not get user data: %s',
                consumer_key)
        else:
            email, first_name, last_name = account_data
            username = _get_available_username(consumer_key)
            # mimic django get_or_create, handling a possible IntegrityError
            # if the user was created in the meantime
            try:
                sid = transaction.savepoint()
                user = User.objects.create_user(username, email, password=None)
                user.first_name = first_name
                user.last_name = last_name
                user.save()
                user.useropenid_set.create(
                    claimed_id=claimed_id, display_id=claimed_id)
                transaction.savepoint_commit(sid)
            except IntegrityError:
                transaction.savepoint_rollback(sid)
                exc_info = sys.exc_info()
                try:
                    user = User.objects.get(useropenid__claimed_id=claimed_id)
                except User.DoesNotExist:
                    # Re-raise the IntegrityError with its original traceback.
                    six.reraise(*exc_info)
            created = True

    return user, created


def update_or_create_user_from_oauth(
        consumer_key, update_window=settings.SSO_UPDATE_WINDOW):
    """Update or create user using SSO account data.

    This function tries to retrieve a `User` from the database using
    the `consumer_key`.  In case the user does not exist, the
    `consumer_key` is then used to fetch account data from SSO and the
    instance is created.  In the case a user already exists, data from
    SSO is only updated after `User.last_login` is older than the
    given `update_window`.  If `update_window` is `None`, SSO data is
    always updated.

    Returns:
        A two-element tuple, where the first element is the
        created/updated user or None and the last one is a boolean
        especifying whether the user was created or not.

    """
    user, created = _get_or_create_user_from_oauth(consumer_key)

    if user is not None:
        if update_window is None:
            should_update = True
        else:
            # Piggyback on `User.last_login` to decide account refresh.
            login_time = time.mktime(user.last_login.timetuple())
            should_update = (time.time() - login_time) > update_window

        if should_update:
            account = web_services.get_data_for_account(
                openid_identifier=consumer_key)
            account_data = _get_sso_account_data(account)
            if account_data is not None:
                user.email, user.first_name, user.last_name = account_data
                user.last_login = timezone.now()
                user.save()

    return user, created
