#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.conf import settings
from django.conf.urls import url, patterns
from django.views.decorators.cache import (
    cache_control,
    never_cache,
)

from piston.resource import Resource
from core.api.auth import SSOOAuthAuthentication
from core.api.decorators import (
    gzip_content,
    vary_only_on_accept,
    vary_only_on_headers,
)
from core.api.resource import CSRFExemptResource
from reviewsapp.api.handlers import (
    FlagReviewHandler,
    ReviewsStatsHandler,
    ServerStatusHandler,
    ShowReviewsHandler,
    ShowReviewsSinceHandler,
    SubmitReviewHandler,
    SubmitUsefulnessHandler,
    RetrieveReviewHandler,
    ShowUsefulnessHandler,
    DeleteReviewHandler,
    ModifyReviewHandler,
)

auth = SSOOAuthAuthentication(realm="Ratings and Reviews")


# The server status resource should never be cached.
class OptionsResource(Resource):
    """A resource that supports the OPTIONS method."""
    callmap = {
        'GET': 'read',
        'POST': 'create',
        'PUT': 'update',
        'DELETE': 'delete',
        'OPTIONS': 'options',
    }
server_status_resource = OptionsResource(handler=ServerStatusHandler)
server_status_resource.__name__ = "OptionsResource"
server_status_resource = never_cache(server_status_resource)

# Although the review stats resource will be cached by the cache
# middleware, we update the cache-control so that upstream caches will
# keep it around for much longer (CACHE_REVIEW_STATS_SECONDS).
reviews_stats_resource = Resource(handler=ReviewsStatsHandler)
cached_resource = cache_control(max_age=settings.CACHE_REVIEW_STATS_SECONDS)
reviews_stats_resource.__name__ = "Resource"
# We need to also vary on Accept-Encoding because we also might
# provide this content gzip-encoded.  Not doing so could bring a
# situation where:
#  1) Client A, supports gzip, states it in the request and content is
#     served, returned gzip-encoded content is now cached.
#  2) But then Client B, who *does not* support gzip, asks for stats.
#     Since the content was cached gzip-encoded and Vary was not set
#     to include Accept-Encoding, that causes the server to return
#     gzipped content to a non supporting client. Super bad.
reviews_stats_resource = vary_only_on_headers('Accept', 'Accept-Encoding')(
    gzip_content(cached_resource(reviews_stats_resource)))

# The show reviews resource is cached.
show_reviews_resource = Resource(handler=ShowReviewsHandler)
show_reviews_resource.__name__ = "Resource"
show_reviews_resource = vary_only_on_accept(show_reviews_resource)

show_reviews_since_resource = Resource(handler=ShowReviewsSinceHandler)
show_reviews_since_resource.__name__ = "Resource"
show_reviews_since_resource = vary_only_on_accept(show_reviews_since_resource)

retrieve_review_resource = Resource(handler=RetrieveReviewHandler)
retrieve_review_resource.__name__ = "Resource"
retrieve_review_resource = vary_only_on_accept(retrieve_review_resource)

# The show usefulness resource is cached.
show_usefulness_resource = Resource(handler=ShowUsefulnessHandler)
show_usefulness_resource.__name__ = "Resource"
show_usefulness_resource = vary_only_on_accept(show_usefulness_resource)


# The following POST handlers have POST data and will not be cached.
# The .__name__ hack is needed to support Django 1.1
submit_review_resource = CSRFExemptResource(handler=SubmitReviewHandler,
                                            authentication=auth)
modify_review_resource = CSRFExemptResource(handler=ModifyReviewHandler,
                                            authentication=auth)
flag_review_resource = CSRFExemptResource(handler=FlagReviewHandler,
                                          authentication=auth)
submit_usefulness_resource = CSRFExemptResource(
    handler=SubmitUsefulnessHandler, authentication=auth)
delete_review_resource = CSRFExemptResource(handler=DeleteReviewHandler,
                                            authentication=auth)


def djson13_url(regex, view, kwargs=None, name=None, prefix=''):
    """Define urls forcing the 'default_emitter' to be 'djson13'.

    This must be used to ensure that all endpoints return JSON data
    encoded with Django 1.3 style dates for keeping compatibility with
    older clients.  For more details on this, see:
    https://bitbucket.org/jespern/django-piston/wiki/Documentation#!emitters

    """
    if kwargs is None:
        kwargs = {}
    kwargs.update({'emitter_format': 'djson13'})
    return url(regex, view, kwargs, name, prefix)


urlpatterns = patterns(
    '',
    # get status of the service (usually just "ok", might be "read-only")
    # to ensure we inform the user when he wants to write a review or
    # send a moderation request
    #  GET /1.0/server-status/
    djson13_url(r'^1.0/server-status/$',
                server_status_resource, name='rnr-server-status'),

    # gets the ratings for a given distroseries
    #  GET /1.0/review-stats
    djson13_url(
        r'^1.0/review-stats/(?P<origin>[-\.\w]+)/(?P<distroseries>\w+)/$',
        reviews_stats_resource, name='rnr-api-reviews-stats'),
    djson13_url(
        r'^1.0/review-stats/(?P<origin>[-\.\w]+)/(?P<distroseries>\w+)/'
        r'updates-last-(?P<days>[1,3,7])-days/$',
        reviews_stats_resource, name='rnr-api-reviews-stats'),

    # The following urls are kept for backwards compatibility with old
    # versions of rnr-client (rnrclient <= r33). Once the bulk of
    # natty clients have been updated, we can remove these two and the
    # associated tests.
    djson13_url(r'^1.0/review-stats/$', reviews_stats_resource,
                name='rnr-api-reviews-stats'),
    djson13_url(r'^1.0/review-stats/updates-last-(?P<days>[1,3,7])-days/$',
                reviews_stats_resource, name='rnr-api-reviews-stats'),

    # gets the reviews (and ratings) for a given pkg/app
    djson13_url(r'^1.0/reviews/filter/(?P<language>\w+)/(?P<origin>[-\.\w]+)/'
                r'(?P<distroseries>[-\.\w]+)/(?P<version>[-\w+.:~]+)/'
                r'(?P<package_name>[\w%+.;-]+)/$',
                show_reviews_resource, name='rnr-api-reviews-for-package'),
    djson13_url(r'^1.0/reviews/filter/(?P<language>\w+)/(?P<origin>[-\.\w]+)/'
                r'(?P<distroseries>[-\.\w]+)/(?P<version>[-\w+.:~]+)/'
                r'(?P<package_name>[\w%+.;-]+)/page/(?P<page>\d+)/$',
                show_reviews_resource, name='rnr-api-reviews-for-package'),
    djson13_url(
        r'^1.0/reviews/filter/(?P<language>\w+)/(?P<origin>[-\.\w]+)/'
        r'(?P<distroseries>[-\.\w]+)/(?P<version>[-\w+.:~]+)/'
        r'(?P<package_name>[\w%+.;-]+)/page/(?P<page>\d+)/(?P<sort>\w+)/$',
        show_reviews_resource, name='rnr-api-reviews-for-package'),

    # gets latest reviews, based on the supplied date
    # GET /1.0/reviews/since/2012-01-02/
    djson13_url(r'^1.0/reviews/since/(?P<since>\d{4}-\d{2}-\d{2})/$',
                show_reviews_since_resource, name='rnr-api-reviews-since'),

    # vote on the usefulness of a review
    #  POST /1.0/reviews/405/usefulness/False/
    djson13_url(r'^1.0/reviews/(?P<review_id>\d+)/recommendations/$',
                submit_usefulness_resource, name='rnr-api-recommendations'),

    # get the filtered usefulness records using parameters for filtering
    #  GET /1.0/reviews/usefulness/?username=aaronp
    #  GET /1.0/reviews/usefulness/?review_id=100
    #  GET /1.0/reviews/usefulness/?username=aaronp&review_id=100
    djson13_url(r'^1.0/usefulness/$',
                show_usefulness_resource, name='rnr-api-usefulness-filter'),

    # gets a single review for a given review id
    # GET /1.0/reviews/100/
    djson13_url(r'^1.0/reviews/(?P<review_id>\d+)/$',
                retrieve_review_resource, name='rnr-api-review-for-review-id'),

    # submits a new review for a given pkg/app
    #  POST /1.0/reviews/
    djson13_url(r'^1.0/reviews/$',
                submit_review_resource, name='rnr-api-reviews'),

    # modifies an existing review
    # PUT /1.0/reviews/modify/100
    djson13_url(r'^1.0/reviews/modify/(?P<review_id>\d+)/$',
                modify_review_resource, name='rnr-api-modify-review'),

    # flag inappropriate reviews
    #  POST /1.0/reviews/1/+report-review/
    djson13_url(r'^1.0/reviews/(?P<review_id>\d+)/flags/$',
                flag_review_resource),

    # delete own reviews
    #  POST /1.0/reviews/delete/100
    djson13_url(r'^1.0/reviews/delete/(?P<review_id>\d+)/$',
                delete_review_resource),
)
