#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2012 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time

from functools import wraps

from reviewsapp.middleware.txstatsd_client import statsd_client as client


def timed_call(view):
    """A decorator for callables that measures time with txstatsd.

    Use it like this::

        @timed_call
        def view(request, *args):
            code

    or::

        @timed_call
        def my_callable(*args):
            code

    It will generate a timing metric (durations and rates) in a path
    that looks like this for Django views::

        <module_path>.<func_name>.<result_code>

    or this for other callables::

        <module_path>.<func_name>

    or::

        <module_path>.<func_name>.error

    if an exception was raised.

    The meter is namespaced thus::

        <environ>.<service_group>.<service_name>.<instance_id>
    """
    @wraps(view)
    def wrapper(*args, **kwargs):
        client.connect()
        start_time = time.time()
        try:
            result = view(*args, **kwargs)
        except:
            key = view.__module__ + "." + view.func_name + ".error"
            client.timing(key, time.time() - start_time)
            raise
        else:
            status_code = getattr(result, "status_code", None)
            if status_code is not None:
                key = ".".join([
                    view.__module__, view.func_name, str(status_code)])
                client.timing(key, time.time() - start_time)
            else:
                key = '.'.join([view.__module__, view.func_name, 'success'])
                client.timing(
                    key,
                    time.time() - start_time)

        return result
    return wrapper
