#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Django models."""

from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'Architecture',
    'Repository',
    'Review',
    'ReviewModeration',
    'ReviewModerationFlag',
    'SoftwareItem',
    'SoftwareItemInOrigin',
    'SoftwareItemInRepository',
]

import pytz

from datetime import datetime

from django.contrib.auth.models import (
    Group,
    Permission,
    User,
)
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models.signals import post_syncdb
from django.utils.translation import ugettext_lazy as _
from core.fields import (
    DebVersionField,
    IntegerRangeField,
)
from core.models import (
    BaseModeration,
    BaseModerationFlag,
    ModerationError,
    ModerationManager,
    ObjectWithReviewStats,
    UserAlreadyFlaggedReview,
)
from reviewsapp.utilities import wilson_score


class SoftwareItem(ObjectWithReviewStats):
    """The software item, i.e. package_name/application_name."""
    package_name = models.SlugField(max_length=100, unique=True)

    def __unicode__(self):
        return u'Package {0}'.format(self.package_name)

    def reviews_for_object(self):
        """See `ObjectWithReviewStats."""
        return Review.objects.filter(softwareitem=self, hide=False,
                                     rating__gt=0, rating__lte=5)

    class Meta:
        app_label = 'reviewsapp'


class SoftwareItemInRepository(ObjectWithReviewStats):
    """Enables us to record statistics for items in a repo."""

    softwareitem = models.ForeignKey('reviewsapp.SoftwareItem')
    repository = models.ForeignKey('reviewsapp.Repository')

    @property
    def softwareitem_in_origin(self):
        item_in_origin, created = SoftwareItemInOrigin.objects.get_or_create(
            softwareitem=self.softwareitem,
            origin=self.repository.origin)
        return item_in_origin

    def reviews_for_object(self):
        """See `ObjectWithReviewStats."""
        return Review.objects.filter(
            softwareitem=self.softwareitem, repository=self.repository,
            hide=False, rating__gt=0, rating__lte=5)

    def update_stats(self):
        """Overridden to ensure that all stats are updated at once."""
        super(SoftwareItemInRepository, self).update_stats()
        self.softwareitem.update_stats()
        self.softwareitem_in_origin.update_stats()

    class Meta:
        app_label = 'reviewsapp'
        unique_together = (('softwareitem', 'repository'),)


class SoftwareItemInOrigin(ObjectWithReviewStats):
    """Enables us to record statistics for items in a repo."""

    softwareitem = models.ForeignKey('reviewsapp.SoftwareItem')
    origin = models.SlugField(max_length=100, db_index=True)

    def reviews_for_object(self):
        """See `ObjectWithReviewStats."""
        return Review.objects.filter(
            softwareitem=self.softwareitem, repository__origin=self.origin,
            hide=False, rating__gt=0, rating__lte=5)

    class Meta:
        app_label = 'reviewsapp'
        unique_together = (('softwareitem', 'origin'),)


class Repository(models.Model):
    """The repository a package comes from.

    That includes the origin (e.g. ubuntu) and the distro release
    (e.g. lucid).
    """
    # Ubuntu, ppa-ppa-name, ...
    origin = models.SlugField(max_length=100)
    # karmic, lucid
    distroseries = models.SlugField(max_length=25)

    def __unicode__(self):
        return u'{0}/{1}'.format(self.distroseries, self.origin)

    class Meta:
        verbose_name_plural = 'Repositories'
        app_label = 'reviewsapp'


class ReviewModeration(BaseModeration):
    """Track moderation of package reviews."""

    review = models.ForeignKey('Review')

    objects = ModerationManager()

    class Meta:
        app_label = 'reviewsapp'

    def get_moderated_object(self):
        return self.review

    def moderate(self, moderator, status, moderation_text):
        # XXX: This ModerationError exception approach doesn't look
        # good, if any validation should happen that must be covered
        # at the form level.  To be refactored later on.
        if not self.can_be_moderated_by(moderator):
            raise ModerationError(
                "This item cannot be moderated by '{0}'.".format(
                    moderator.username))

        if self.status != self.PENDING_STATUS:
            raise ModerationError(
                "Only pending moderations can be moderated.")

        if status == self.PENDING_STATUS:
            raise ModerationError(
                "You must approve or reject to moderate.")

        super(ReviewModeration, self).moderate(
            moderator, status, moderation_text)

    def on_moderate(self, review):
        # Hide or unhide the related review as appropriate.
        should_hide = (self.status == ReviewModeration.REJECTED_STATUS)
        if review.hide != should_hide:
            review.hide = should_hide
            review.save()
            review.softwareitem_in_repository.update_stats()

    def can_be_moderated_by(self, user):
        if self.review.reviewer == user:
            return False

        if self.reviewmoderationflag_set.filter(user=user):
            return False

        return True


class ReviewModerationFlag(BaseModerationFlag):
    """An object that has been flagged for moderation."""
    review_moderation = models.ForeignKey(ReviewModeration)

    class Meta:
        ordering = ['-date_created']
        app_label = 'reviewsapp'


class Architecture(models.Model):
    """Represents a machine architecture.

    Note: This is only included as a model (rather than a char attribute
    on Review) so that we can dynamically update the list of valid
    architectures (either by adding them via the admiin interface or by
    checking with Launchpad).
    """
    tag = models.CharField(unique=True, max_length=16)

    class Meta:
        app_label = 'reviewsapp'

    def __unicode__(self):
        return self.tag


class Review(models.Model):
    """A specific review."""
    softwareitem = models.ForeignKey(SoftwareItem)
    version = DebVersionField(max_length=100, db_index=True)
    architecture = models.ForeignKey(Architecture)
    repository = models.ForeignKey(Repository)
    reviewer = models.ForeignKey(User)
    date_created = models.DateTimeField(
        default=lambda: datetime.now(pytz.utc),
        editable=False)
    rating = IntegerRangeField(min_value=1, max_value=5)
    summary = models.CharField(max_length=80)
    review_text = models.CharField(max_length=5000)
    language = models.SlugField(max_length=10)
    hide = models.BooleanField(
        'Hide review from the UI', db_index=True, default=False)
    date_deleted = models.DateTimeField(blank=True, null=True)
    # The following get updated when a usefulness vote is added
    usefulness_total = models.IntegerField(default=0)
    usefulness_favorable = models.IntegerField(default=0)
    usefulness_percentage = models.IntegerField(default=0)
    usefulness_wilson_rating = models.FloatField(default=0.0)
    # XXX 2011-03-18 michaeln Remove application_name eventually.
    # After removing the app_name attribute of software item, we add it
    # here on the review so that we're not losing any information being
    # passed by the client.
    app_name = models.CharField(max_length=100, default='', blank=True)

    def __unicode__(self):
        return u'Review for {0}'.format(self.softwareitem.package_name)

    @property
    def softwareitem_in_repository(self):
        """Helper property until we switch the foreignkey."""
        item_repo, created = SoftwareItemInRepository.objects.get_or_create(
            softwareitem=self.softwareitem, repository=self.repository)
        return item_repo

    def package_name(self):
        return self.softwareitem.package_name

    def origin(self):
        return self.repository.origin

    def distroseries(self):
        return self.repository.distroseries

    def package_app_name(self):
        app_name = self.app_name
        if app_name:
            return self.package_name() + "/" + app_name
        else:
            return self.package_name()

    def reviewer_username(self):
        return self.reviewer.username

    def reviewer_displayname(self):
        return self.reviewer.get_full_name()

    def is_awaiting_moderation(self):
        """Return true if the review has any pending moderations"""
        moderations = self.reviewmoderation_set.filter(
            status=ReviewModeration.PENDING_STATUS).count()
        if moderations == 0:
            return False
        return True

    def delete(self):
        """Delete a review submitted by the user"""
        self.hide = True
        self.date_deleted = datetime.now(pytz.utc)
        self.save()
        self.softwareitem_in_repository.update_stats()
        return self

    def flag(self, user, summary, description):
        """Flag an object as possibly requiring moderation.

        @param user: The user flagging the object.
        @param summary: A summary identifying the reason for flagging the
            item such as 'Spam', or 'Offensive language'.
        @param description: A description of the reason for flagging the
            item.

        # XXX michaeln 2010-11-29 Extract moderation functionality.
        This method may be added automatically when registering an
        item.
        """
        # Ensure that this user hasn't already flagged this item.
        previous_flags_for_user = ReviewModerationFlag.objects.filter(
            review_moderation__review=self,
            user=user)
        if previous_flags_for_user.count() > 0:
            raise UserAlreadyFlaggedReview()

        review_moderation, created = self.reviewmoderation_set.get_or_create(
            status=ReviewModeration.PENDING_STATUS)
        flagged_object = review_moderation.reviewmoderationflag_set.create(
            user=user, summary=summary, description=description)

        # If this object hasn't been approved previously, then we hide
        # it.
        if self.reviewmoderation_set.approved().count() == 0:
            self.hide = True
            self.save()
            self.softwareitem_in_repository.update_stats()

        return flagged_object

    def update_wilson_rating(self):
        self.usefulness_wilson_rating = wilson_score(
            self.usefulness_favorable, self.usefulness_total)
        self.save()

    def update_usefulness_aggregates(self):
        total = Usefulness.objects.filter(review=self).count()
        useful = Usefulness.objects.filter(review=self, useful=True).count()
        if total == 0:
            percentage = 0
        else:
            percentage = useful / float(total)

        self.usefulness_total = total
        self.usefulness_favorable = useful
        self.usefulness_percentage = int(percentage * 100)
        self.save()

    class Meta:
        app_label = 'reviewsapp'


class Usefulness(models.Model):
    review = models.ForeignKey(Review, db_index=True)
    user = models.ForeignKey(User)
    useful = models.BooleanField(db_index=True, default=False)

    def username(self):
        return self.user.username

    def review_id(self):
        return self.review.id

    class Meta:
        app_label = 'reviewsapp'


class RNRSettings(models.Model):
    PASSIVE_MODERATION = 'passive'
    ACTIVE_MODERATION = 'active'
    BLACKLIST_MODERATION = 'blcklst'
    MODERATION_CHOICES = (
        (PASSIVE_MODERATION, 'Passive - automated approval'),
        (ACTIVE_MODERATION, 'Active - manual approval'),
        (BLACKLIST_MODERATION, 'Automated moderation based on blacklist'),
    )

    moderation_mode = models.CharField(
        default=PASSIVE_MODERATION, max_length=8, choices=MODERATION_CHOICES)

    blacklist_words = models.TextField(
        default='', help_text=_("Enter one word per line."))

    @classmethod
    def get_settings(cls):
        # There is only ever a singleton setting object.
        try:
            settings = RNRSettings.objects.get()
        except RNRSettings.DoesNotExist:
            settings = RNRSettings.objects.create()

        return settings

    @classmethod
    def set_moderation_mode(cls, mode):
        settings = cls.get_settings()
        settings.moderation_mode = mode
        settings.save()

    @classmethod
    def should_moderate_text(cls, *text):
        settings = cls.get_settings()
        if settings.moderation_mode == cls.ACTIVE_MODERATION:
            return True
        elif settings.moderation_mode == cls.PASSIVE_MODERATION:
            return False

        blacklist_words = settings.blacklist_words.split('\n')
        combined_texts = "\n".join(text)
        for blacklist_word in blacklist_words:
            if blacklist_word in combined_texts:
                return True

        return False

    @classmethod
    def get_moderation_summary_and_desc(cls):
        moderation_mode = cls.get_settings().moderation_mode
        if moderation_mode == cls.ACTIVE_MODERATION:
            return (
                'Automatically Flagged',
                'This review was automatically flagged because the rnr '
                'server was operating in active moderation mode')
        elif moderation_mode == cls.BLACKLIST_MODERATION:
            return (
                'Contains Blacklisted Word',
                'This review was flagged because the server was operating '
                'in blacklist mode and the review contained a blacklisted '
                'word.')
        else:
            return None, None

    class Meta:
        app_label = 'reviewsapp'
        verbose_name_plural = 'RNR Settings'


# GroupPermissions can't go in fixtures because Permissions are stored as
# model metadata, so our Permission's primary key can change without warning
def post_syncdb_handler(sender, **kwargs):
    if sender.__name__ == 'reviewsapp.models':
        ct = ContentType.objects.get_for_model(ReviewModeration)
        perm, created = Permission.objects.get_or_create(
            codename='can_moderate',
            name="Can resolve moderation requests",
            content_type=ct,
        )
        group, created = Group.objects.get_or_create(name='rnr-moderators')
        if perm not in group.permissions.all():
            group.permissions.add(perm)
            group.save()
        group, created = Group.objects.get_or_create(name='rnr-developers')

post_syncdb.connect(post_syncdb_handler)
