#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Ubuntu App Review test helpers."""

from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'WSGIInterceptedTestCase',
    'mock_data_for_account',
]

from contextlib import contextmanager
from unittest import TestCase

from django.db import close_old_connections

from django.conf import settings
from django.core import signals
from wsgi_intercept import add_wsgi_intercept, remove_wsgi_intercept
from wsgi_intercept.httplib2_intercept import install, uninstall
from wsgi_intercept.urllib2_intercept import install_opener, uninstall_opener


class WSGIInterceptedTestCase(TestCase):
    def setUp(self, callbacks=None):
        super(WSGIInterceptedTestCase, self).setUp()
        if callbacks is None:
            callbacks = {}

        # Django 1.6 changed the way db connections are managed in the
        # request cycle, now it cleans up connections at request start
        # and end by detecting if the are usable or not.  Because the
        # current connection in tests generates outside of the patched
        # wsgi request, such connections are closed and therefore
        # tests fail. This code prevents that:
        signals.request_started.disconnect(close_old_connections)
        signals.request_finished.disconnect(close_old_connections)
        self.addCleanup(
            signals.request_started.connect, close_old_connections)
        self.addCleanup(
            signals.request_finished.connect, close_old_connections)

        self.intercepted = []
        install_opener()
        install()

        for key, callback in callbacks.items():
            if key == 'default':
                continue
            host, port = key
            add_wsgi_intercept(host, port, callback)
            self.intercepted.append((host, port))

    def tearDown(self):
        for host, port in self.intercepted:
            remove_wsgi_intercept(host, port)
        uninstall_opener()
        uninstall()
        super(WSGIInterceptedTestCase, self).tearDown()


# Original snippet from http://djangosnippets.org/snippets/2156/
class SettingDoesNotExist:
    pass


def switch_settings(**kwargs):
    """Helper method that updates settings and returns old settings."""
    old_settings = {}
    for key, new_value in kwargs.items():
        old_value = getattr(settings, key, SettingDoesNotExist)
        old_settings[key] = old_value

        if new_value is SettingDoesNotExist:
            delattr(settings, key)
        else:
            setattr(settings, key, new_value)

    return old_settings


@contextmanager
def patch_settings(**kwargs):
    old_settings = switch_settings(**kwargs)
    yield
    switch_settings(**old_settings)
# end snippet


@contextmanager
def disable_csrf(**kwargs):
    middleware_classes = settings.MIDDLEWARE_CLASSES
    settings.MIDDLEWARE_CLASSES = [x for x in middleware_classes
                                   if 'csrf' not in x.lower()]
    yield
    settings.MIDDLEWARE_CLASSES = middleware_classes


def mock_data_for_account(consumer, token, **kwargs):
    """ A credible return value for get_data_for_account. """
    data = {
        'preferred_email': 'foo@bar.com',
        'username': 'foobar',
        'displayname': 'foo bar',
        'consumer_secret': consumer.secret,
        'token': token.token,
        'token_secret': token.token_secret,
        'name': token.name,
    }
    data.update(kwargs)
    return data
