from django.http import HttpResponse
from django.test import TestCase

from mock import patch, ANY

from reviewsapp.decorators import timed_call
from reviewsapp.middleware.txstatsd_client import StatsdClient


class TestTimedCall(TestCase):
    """Test the timed_call decorator."""

    def setUp(self):
        self.patcher = patch('reviewsapp.decorators.client',
                             spec=StatsdClient)
        self.mock_client = self.patcher.start()

    def tearDown(self):
        self.patcher.stop()

    def test_timed_call_with_callable(self):
        """Ensure that the timed_call decorator logs the expected metrics
        when used on a callable that isn't a Django view.
        """
        @timed_call
        def test_func():
            return True

        test_func()

        self.mock_client.timing.assert_called_once_with(
            '{0}.{1}.success'.format(
                test_func.__module__, test_func.func_name),
            ANY)

    def test_timed_call_with_error(self):
        """Ensure that the timed_call decorator logs the expected metrics
        when used on a callable that isn't a Django view and raises an error.
        """
        @timed_call
        def test_func():
            raise StandardError()

        self.assertRaises(StandardError, test_func)

        self.mock_client.timing.assert_called_once_with(
            '{0}.{1}.error'.format(
                test_func.__module__, test_func.func_name),
            ANY)

    def test_timed_call_with_django_view(self):
        """Ensure that the timed_call decorator logs the expected metrics when
        decorating a Django view.
        """
        @timed_call
        def test_view(request):
            """Simple view that returns a stub response object containing a
            status_code.
            """
            return HttpResponse("Response", status=200)

        test_view(object())

        self.mock_client.timing.assert_called_once_with(
            '{0}.{1}.200'.format(
                test_view.__module__, test_view.func_name),
            ANY)
