#   -*- coding: utf-8 -*-
#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Ubuntu App Review handlers tests."""

from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'FlagReviewHandlerTestCase',
    'ReviewsStatsHandlerTestCase',
    'ServerStatusHandlerTestCase',
    'ShowReviewsHandlerTestCase',
    'ShowReviewsSinceHandlerTestCase',
    'SubmitReviewHandlerTestCase',
    'ModifyReviewHandlerTestCase',
    'UsefulnessHandlerTestCase',
    'ShowUsefulnessHandlerTestCase',
    'DeleteReviewHandlerTestCase',
]

import httplib

import pytz

from django.http import Http404
from datetime import date, datetime, timedelta
from urllib import quote_plus

from mock import Mock, patch

from django.conf import settings
from django.core.cache import cache
from django.core.urlresolvers import (
    NoReverseMatch,
    reverse,
)
from django.test import TestCase
from django.test.client import RequestFactory
from django.utils import simplejson
from django.utils.unittest import skipUnless
from piston.emitters import Emitter
from piston.handler import typemapper

from reviewsapp.api.handlers import (
    Django13JSONEncoder,
    Django13JSONEmitter,
    FlagReviewHandler,
    ServerStatusHandler,
    ShowReviewsHandler,
    SubmitUsefulnessHandler,
    ShowUsefulnessHandler,
    DeleteReviewHandler,
    ModifyReviewHandler,
)
from reviewsapp.models import (
    Repository,
    Review,
    ReviewModerationFlag,
    SoftwareItem,
)
from reviewsapp.tests.factory import TestCaseWithFactory
from reviewsapp.tests.factory import RequestTestObject
from reviewsapp.tests.helpers import patch_settings
from reviewsapp.tests.matchers import (
    ContainsStatsForPackage,
)
from reviewsapp.utilities import (
    cache_key_for_url,
    is_using_postgres,
    wilson_score,
)


def _get_url_kwargs_dict_for_review(review):
    # this is the quoting we expect
    appname = quote_plus(quote_plus(
        ";" + review.app_name.encode("utf-8")))

    pkgname = review.softwareitem.package_name
    return dict(language=review.language,
                origin=review.repository.origin,
                distroseries=review.repository.distroseries,
                package_name="%s%s" % (pkgname, appname),
                version=review.version)


class Django13JSONEncoderTestCase(TestCase):

    def test_datetime_format(self):
        dt = datetime(2014, 8, 25, 12, 30, 25, 10)
        dt_str = Django13JSONEncoder().default(dt)
        self.assertEqual(dt_str, '2014-08-25 12:30:25')

    def test_date_format(self):
        d = date(2014, 8, 25)
        d_str = Django13JSONEncoder().default(d)
        self.assertEqual(d_str, '2014-08-25')

    def test_time_format(self):
        t = datetime(2014, 8, 25, 12, 30, 25, 10).time()
        time_str = Django13JSONEncoder().default(t)
        self.assertEqual(time_str, '12:30:25')


class Django13JSONEmitterTestCase(TestCase):

    def test_json_encoding_datetime_format(self):
        dt = datetime(2014, 8, 25, 12, 30, 25, 10)
        payload = {'datetime': dt}
        emitter = Django13JSONEmitter(payload, typemapper, lambda r: None)
        request = RequestFactory()
        request.GET = {}
        rendered = emitter.render(request)
        self.assertEqual(
            rendered, '{\n    "datetime": "2014-08-25 12:30:25"\n}')

    def test_json_encoding_date_format(self):
        d = date(2014, 8, 25)
        payload = {'date': d}
        emitter = Django13JSONEmitter(payload, typemapper, lambda r: None)
        request = RequestFactory()
        request.GET = {}
        rendered = emitter.render(request)
        self.assertEqual(
            rendered, '{\n    "date": "2014-08-25"\n}')

    def test_json_encoding_time_format(self):
        t = datetime(2014, 8, 25, 12, 30, 25, 10).time()
        payload = {'time': t}
        emitter = Django13JSONEmitter(payload, typemapper, lambda r: None)
        request = RequestFactory()
        request.GET = {}
        rendered = emitter.render(request)
        self.assertEqual(
            rendered, '{\n    "time": "12:30:25"\n}')

    def test_valid_jsonp_callback(self):
        """Valid JSONP callback should be added to response."""
        t = datetime(2014, 8, 25, 12, 30, 25, 10).time()
        payload = {'time': t}
        emitter = Django13JSONEmitter(payload, typemapper, lambda r: None)
        request = RequestFactory()
        request.GET = {'callback': 'cb'}
        rendered = emitter.render(request)
        self.assertEqual(
            rendered, 'cb({\n    "time": "12:30:25"\n})')

    def test_invalid_jsonp_callback(self):
        """Invalid JSONP callback be ignored silently."""
        t = datetime(2014, 8, 25, 12, 30, 25, 10).time()
        payload = {'time': t}
        emitter = Django13JSONEmitter(payload, typemapper, lambda r: None)
        request = RequestFactory()
        request.GET = {'callback': 'invalid()'}
        rendered = emitter.render(request)
        self.assertEqual(
            rendered, '{\n    "time": "12:30:25"\n}')

    def test_is_registered(self):
        expected = (Django13JSONEmitter, 'application/json; charset=utf-8')
        self.assertEqual(Emitter.EMITTERS['djson13'], expected)


class ServerStatusHandlerTestCase(TestCase):
    def test_read(self):
        ss_handler = ServerStatusHandler()
        self.assertEqual('ok', ss_handler.read(None))

    def test_options(self):
        response = self.client.options(reverse('rnr-server-status'))

        self.assertEqual(response.content, '')
        self.assertEqual(response['allow'], 'GET,OPTIONS')

    def test_request_not_cached(self):
        # Requests for server status are never cached.
        response = self.client.get(reverse('rnr-server-status'))

        self.assertEqual('max-age=0', response['cache-control'])

        # A subsequent call will hit the view.
        with patch('reviewsapp.api.handlers.ServerStatusHandler.read') as (
                mock_read):
            response = self.client.get(reverse('rnr-server-status'))
            self.assertTrue(mock_read.called)


class ReviewsStatsHandlerTestCase(TestCaseWithFactory):

    def clear_cache(self):
        cache._cache.clear()
        cache._expire_info.clear()

    def _request_stats(self, last_days=0, data=None, origin='any',
                       distroseries='any', use_old_api=False):
        data = data or {}
        if use_old_api:
            kwargs = {}
        else:
            kwargs = dict(origin=origin, distroseries=distroseries)
        if last_days > 0:
            kwargs['days'] = last_days
        url = reverse('rnr-api-reviews-stats', kwargs=kwargs)
        return self.client.get(url, data)

    def test_read(self):
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'),
            repository=self.factory.makeRepository('ubuntu', 'hardy'),
            language='de')
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_bar'),
            repository=self.factory.makeRepository('ubuntu', 'natty'),
            language='de')

        self.clear_cache()

        response = self._request_stats()
        reviews = simplejson.loads(response.content)

        self.assertEqual(2, len(reviews))
        self.assertThat(reviews, ContainsStatsForPackage(
            'package_foo', average=5, count=1, hist='[0, 0, 0, 0, 1]'))
        self.assertThat(reviews, ContainsStatsForPackage(
            'package_bar', average=5, count=1, hist='[0, 0, 0, 0, 1]'))

    def test_read_for_repo(self):
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'),
            repository=self.factory.makeRepository('ubuntu', 'hardy'),
            language='de')
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_bar'),
            repository=self.factory.makeRepository('ubuntu', 'natty'),
            language='de')

        response = self._request_stats(origin='ubuntu', distroseries='hardy')
        reviews = simplejson.loads(response.content)

        self.assertEqual(1, len(reviews))
        self.assertThat(reviews, ContainsStatsForPackage(
            'package_foo', average=5, count=1, hist='[0, 0, 0, 0, 1]'))

    def test_read_for_origin(self):
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'),
            repository=self.factory.makeRepository('ubuntu', 'hardy'),
            language='de')
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_bar'),
            repository=self.factory.makeRepository('canonical', 'natty'),
            language='de')

        response = self._request_stats(origin='ubuntu', distroseries='any')
        reviews = simplejson.loads(response.content)

        self.assertEqual(1, len(reviews))
        assertion = ContainsStatsForPackage(
            'package_foo', average=5, count=1, hist='[0, 0, 0, 0, 1]')
        self.assertThat(reviews, assertion)

    def test_request_cached_with_extra_headers(self):
        # Requests for server status are never cached.
        response = self._request_stats()

        self.assertEqual(
            'max-age={0}'.format(settings.CACHE_REVIEW_STATS_SECONDS),
            response['cache-control'])

        # A subsequent call will not hit the view.
        read_handler = 'reviewsapp.api.handlers.ReviewsStatsHandler.read'
        with patch(read_handler) as mock_read:
            response = self._request_stats()
            self.assertFalse(mock_read.called)

    def test_cannot_request_with_get_params(self):
        # Requesting the un-cached response (by adding get params) is
        # not allowed.
        response = self._request_stats(data=dict(foo='bar'))

        self.assertEqual(400, response.status_code)

    def test_updates_last_days(self):
        # Only items which have updated stats are returned.
        # 4 days old stats
        self.factory.makeSoftwareItem(
            date_ratings_updated=datetime.now(pytz.utc) - timedelta(4))
        stat_2_days_old = self.factory.makeSoftwareItem(
            date_ratings_updated=datetime.now(pytz.utc) - timedelta(2))
        stat_0_days_old = self.factory.makeSoftwareItem()

        response = self._request_stats(last_days=3)

        response_decoded = simplejson.loads(response.content)
        self.assertEqual(2, len(response_decoded))
        self.assertThat(response_decoded, ContainsStatsForPackage(
            stat_2_days_old.package_name, average='3.5', count=2,
            hist="[0, 0, 1, 1, 0]"))
        self.assertThat(response_decoded, ContainsStatsForPackage(
            stat_0_days_old.package_name, average='3.5', count=2,
            hist="[0, 0, 1, 1, 0]"))

    def test_updates_last_days_matches_only_specified_days(self):
        # The handler cannot be used to generate random updates.
        # The following are all valid.
        kwargs = dict(origin='ubuntu', distroseries='natty')
        for days in (1, 3, 7):
            kwargs['days'] = days
            reverse('rnr-api-reviews-stats', kwargs=kwargs)

        # But other options for days will not hit the handler.
        for days in (2, 4, 8):
            kwargs['days'] = days
            self.assertRaises(
                NoReverseMatch, reverse, 'rnr-api-reviews-stats',
                kwargs=kwargs)

    def test_stats_old_api(self):
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'),
            repository=self.factory.makeRepository('ubuntu', 'hardy'))

        response = self._request_stats(use_old_api=True)
        reviews = simplejson.loads(response.content)

        self.assertEqual(1, len(reviews))
        assertion = ContainsStatsForPackage(
            'package_foo', average=5, count=1, hist="[0, 0, 0, 0, 1]")
        self.assertThat(reviews, assertion)

    def test_stats_old_api_with_days(self):
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'),
            repository=self.factory.makeRepository('ubuntu', 'hardy'))

        response = self._request_stats(last_days=1, use_old_api=True)
        reviews = simplejson.loads(response.content)

        self.assertEqual(1, len(reviews))
        assertion = ContainsStatsForPackage(
            'package_foo', average=5, count=1, hist="[0, 0, 0, 0, 1]")
        self.assertThat(reviews, assertion)

    def test_read_across_origins(self):
        # It is possible to request natty apps from various origins
        # (ubuntu, a ppa, or canonical etc.)
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'),
            repository=self.factory.makeRepository('ubuntu', 'natty'),
            language='de')
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_bar'),
            repository=self.factory.makeRepository('canonical', 'natty'),
            language='de')

        response = self._request_stats(origin='any', distroseries='natty')
        reviews = simplejson.loads(response.content)

        self.assertEqual(2, len(reviews))
        assertion = ContainsStatsForPackage(
            'package_foo', average=5, count=1, hist="[0, 0, 0, 0, 1]")
        self.assertThat(reviews, assertion)
        assertion = ContainsStatsForPackage(
            'package_bar', average=5, count=1, hist="[0, 0, 0, 0, 1]")
        self.assertThat(reviews, assertion)

    def test_histogram_data_can_be_disabled(self):
        review = self.factory.makeReview()
        with patch_settings(ENABLE_HISTOGRAM_DATA=False):
            response = self._request_stats(origin='any', distroseries='any')
            reviews = simplejson.loads(response.content)
            self.assertThat(reviews, ContainsStatsForPackage(
                review.softwareitem.package_name, average=5, count=1))
        self.assertEquals(settings.ENABLE_HISTOGRAM_DATA, True)

    def test_vary_headers(self):
        response = self._request_stats(origin='any', distroseries='any')
        self.assertEqual(200, response.status_code)
        self.assertTrue(response.has_header('vary'))
        vary = response['vary']
        self.assertEqual('Accept, Accept-Encoding', vary)


class ShowUsefulnessHandlerTestCase(TestCaseWithFactory):
    def test_read_for_single_parameter(self):
        '''test that a single usefulness vote by a user is returned when
         searching for that user'''
        review = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_bar'))
        user = self.factory.makeUser()
        usefulness = self.factory.makeUsefulness(review, user)

        handler = ShowUsefulnessHandler()

        request1 = RequestTestObject({'username': user.username})
        request2 = RequestTestObject({'review_id': review.id})

        results1 = handler.read(request1)
        results2 = handler.read(request2)

        self.assertEqual(1, len(results1))
        self.assertEqual(usefulness, results1[0])
        self.assertEqual(usefulness, results2[0])

    def test_read_filtering(self):
        '''test that a passed parameters results in all relevant items
        being returned, but no unwanted items'''
        review1 = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_bar'))
        review2 = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('foo_bar'))
        user1 = self.factory.makeUser()
        user2 = self.factory.makeUser()
        usefulness1 = self.factory.makeUsefulness(review1, user1)
        usefulness2 = self.factory.makeUsefulness(review1, user2)
        usefulness3 = self.factory.makeUsefulness(review2, user1)

        handler = ShowUsefulnessHandler()

        request1 = RequestTestObject({'username': user1.username})
        request2 = RequestTestObject({'review_id': review1.id})
        request3 = RequestTestObject(
            {'username': user1.username, 'review_id': review2.id}
        )

        results1 = handler.read(request1)
        results2 = handler.read(request2)
        results3 = handler.read(request3)

        # Test username filter results
        self.assertEqual(2, len(results1))
        self.assertIn(usefulness1, results1)
        self.assertIn(usefulness3, results1)
        self.assertNotIn(usefulness2, results1)

        # Test review_id filter results
        self.assertEqual(2, len(results2))
        self.assertIn(usefulness1, results2)
        self.assertIn(usefulness2, results2)
        self.assertNotIn(usefulness3, results2)

        # Test usefulness AND review_id filter results
        self.assertEqual(1, len(results3))
        self.assertNotIn(usefulness1, results3)
        self.assertNotIn(usefulness2, results3)
        self.assertIn(usefulness3, results3)

    def test_read_for_no_parameters_gives_400_error(self):
        '''test that a http 400 error is returned if no valid parameter
        is passed'''
        handler = ShowUsefulnessHandler()
        request1 = RequestTestObject()
        request2 = RequestTestObject({'something': 'nothing'})
        results1 = handler.read(request1)
        results2 = handler.read(request2)
        self.assertEqual(httplib.BAD_REQUEST, results1.status_code)
        self.assertEqual(httplib.BAD_REQUEST, results2.status_code)

    def test_vary_headers(self):
        url = reverse('rnr-api-usefulness-filter')
        response = self.client.get(url + '?username=test')
        self.assertEqual(200, response.status_code)
        self.assertTrue(response.has_header('vary'))
        self.assertEqual(response['vary'], 'Accept')


class RetrieveReviewHandlerHandlerTestCase(TestCaseWithFactory):

    def _request_review(self, review_id=None):
        url = reverse('rnr-api-review-for-review-id', args=(review_id,))
        return self.client.get(url)

    def test_not_found(self):
        response = self._request_review(0)
        self.assertEqual(404, response.status_code)

    def test_ok(self):
        review = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'))
        response = self._request_review(review.pk)
        self.assertEqual(200, response.status_code)

    def test_vary_headers(self):
        review = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'))
        url = reverse('rnr-api-usefulness-filter')
        response = self.client.get(url, {'review_id': review.pk})
        self.assertEqual(200, response.status_code)
        self.assertTrue(response.has_header('vary'))
        self.assertEqual(response['vary'], 'Accept')


class ShowReviewsHandlerTestCase(TestCaseWithFactory):

    _GET_REQUEST = Mock()
    _GET_REQUEST.GET = {}

    def test_review_for_package(self):
        review = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'))
        handler = ShowReviewsHandler()

        reviews = handler.read(self._GET_REQUEST, 'any', 'any', 'any', 'any',
                               'package_foo')

        self.assertEqual(1, len(reviews))
        self.assertEqual(review, reviews[0])

    def test_review_for_app(self):
        # We no-longer differentiate for app_names, returning all
        # reviews for the package instead.
        software_item = self.factory.makeSoftwareItem('package_foo')
        review1 = self.factory.makeReview(software_item=software_item,
                                          app_name='app_foo')
        review2 = self.factory.makeReview(software_item=software_item,
                                          app_name='AnotherApp')
        handler = ShowReviewsHandler()

        reviews = handler.read(self._GET_REQUEST, 'any', 'any', 'any', 'any',
                               'package_foo;app_foo')

        self.assertEqual(set([review1, review2]), set(reviews))

    def test_review_for_app_url(self):
        # test utf8 encoded app
        review = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'),
            app_name=u'Déjà Dup')
        url = reverse(
            'rnr-api-reviews-for-package',
            kwargs=_get_url_kwargs_dict_for_review(review))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        reviews = simplejson.loads(response.content)
        self.assertEqual(reviews[0]["app_name"], review.app_name)
        self.assertEqual(reviews[0]["package_name"], review.package_name())

        # now repeat with blender that contains a ()
        review = self.factory.makeReview(
            software_item=review.softwareitem,
            app_name='Blender (windowed)')
        url = reverse(
            'rnr-api-reviews-for-package',
            kwargs=_get_url_kwargs_dict_for_review(review))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        reviews = simplejson.loads(response.content)
        self.assertEqual(reviews[0]["app_name"], review.app_name)
        self.assertEqual(reviews[0]["package_name"], review.package_name())

        # now repeat with a fake app that contains a ";"
        review = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo2'),
            app_name=u'TestApp with;!"§$%&()=?ß/äluü')
        url = reverse(
            'rnr-api-reviews-for-package',
            kwargs=_get_url_kwargs_dict_for_review(review))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        reviews = simplejson.loads(response.content)
        self.assertEqual(reviews[0]["app_name"], review.app_name)
        self.assertEqual(reviews[0]["package_name"], review.package_name())

    def test_review_for_language(self):
        package = self.factory.makeSoftwareItem('package_foo')
        review1 = self.factory.makeReview(software_item=package,
                                          language='en')
        review2 = self.factory.makeReview(software_item=package,
                                          language='de')
        handler = ShowReviewsHandler()

        reviews = handler.read(self._GET_REQUEST, 'en', 'any', 'any', 'any',
                               'package_foo')

        self.assertEqual(1, len(reviews))
        self.assertIn(review1, reviews)
        self.assertNotIn(review2, reviews)

    def test_review_for_any_language(self):
        package = self.factory.makeSoftwareItem('package_foo')
        repository = self.factory.makeRepository('ubuntu', 'hardy')
        review1 = self.factory.makeReview(
            software_item=package,
            language='en', repository=repository, version='1.0')
        review2 = self.factory.makeReview(
            software_item=package,
            language='de', repository=repository, version='1.0')
        handler = ShowReviewsHandler()

        reviews = handler.read(self._GET_REQUEST, 'any', 'ubuntu', 'hardy',
                               '1.0', 'package_foo')

        self.assertEqual(2, len(reviews))
        self.assertIn(review1, reviews)
        self.assertIn(review2, reviews)

    def test_review_from_distroseries(self):
        package = self.factory.makeSoftwareItem('package_foo')
        review1 = self.factory.makeReview(
            software_item=package,
            repository=self.factory.makeRepository('ubuntu', 'hardy'))
        review2 = self.factory.makeReview(
            software_item=package,
            repository=self.factory.makeRepository('ubuntu', 'lucid'))
        handler = ShowReviewsHandler()

        reviews = handler.read(self._GET_REQUEST, 'any', 'any', 'hardy', 'any',
                               'package_foo')

        self.assertEqual(1, len(reviews))
        self.assertIn(review1, reviews)
        self.assertNotIn(review2, reviews)

    def test_review_from_any_distroseries(self):
        package = self.factory.makeSoftwareItem('package_foo')
        review1 = self.factory.makeReview(
            software_item=package,
            repository=self.factory.makeRepository('ubuntu', 'hardy'),
            language='en', version='+-1')
        review2 = self.factory.makeReview(
            software_item=package,
            repository=self.factory.makeRepository('ubuntu', 'lucid'),
            language='en', version='+-1')
        handler = ShowReviewsHandler()

        reviews = handler.read(self._GET_REQUEST, 'en', 'ubuntu', 'any', '+-1',
                               'package_foo')

        self.assertEqual(2, len(reviews))
        self.assertIn(review1, reviews)
        self.assertIn(review2, reviews)

    def test_review_from_origin(self):
        package = self.factory.makeSoftwareItem('package_foo')
        review1 = self.factory.makeReview(
            software_item=package,
            repository=self.factory.makeRepository('ubuntu', 'hardy'))
        review2 = self.factory.makeReview(
            software_item=package,
            repository=self.factory.makeRepository('debian', 'lenny'))
        handler = ShowReviewsHandler()

        reviews = handler.read(self._GET_REQUEST, 'any', 'ubuntu', 'any',
                               'any', 'package_foo')

        self.assertEqual(1, len(reviews))
        self.assertIn(review1, reviews)
        self.assertNotIn(review2, reviews)

    def test_review_from_any_origin(self):
        package = self.factory.makeSoftwareItem('package_foo')
        review1 = self.factory.makeReview(
            software_item=package,
            repository=self.factory.makeRepository('ubuntu', 'hardy'),
            language='en', version='vanilla')
        review2 = self.factory.makeReview(
            software_item=package,
            repository=self.factory.makeRepository('debian', 'hardy'),
            language='en', version='vanilla')
        handler = ShowReviewsHandler()

        reviews = handler.read(self._GET_REQUEST, 'en', 'any', 'hardy',
                               'vanilla', 'package_foo')

        self.assertEqual(2, len(reviews))
        self.assertIn(review1, reviews)
        self.assertIn(review2, reviews)

    def test_review_for_version(self):
        package = self.factory.makeSoftwareItem('package_foo')
        review1 = self.factory.makeReview(software_item=package,
                                          version='0.2')
        review2 = self.factory.makeReview(software_item=package,
                                          version='0.3')
        handler = ShowReviewsHandler()

        reviews = handler.read(self._GET_REQUEST, 'any', 'any', 'any', '0.2',
                               'package_foo')

        self.assertEqual(1, len(reviews))
        self.assertIn(review1, reviews)
        self.assertNotIn(review2, reviews)

    def test_review_for_any_version(self):
        package = self.factory.makeSoftwareItem('package_foo')
        repository = self.factory.makeRepository('ubuntu', 'hardy')
        review1 = self.factory.makeReview(
            software_item=package,
            language='en', repository=repository, version='1.0')
        review2 = self.factory.makeReview(
            software_item=package,
            language='en', repository=repository, version='2.0')
        handler = ShowReviewsHandler()

        reviews = handler.read(self._GET_REQUEST, 'en', 'ubuntu', 'hardy',
                               'any', 'package_foo')

        self.assertEqual(2, len(reviews))
        self.assertIn(review1, reviews)
        self.assertIn(review2, reviews)

    def _create_reviews_for_version_tests(self, pkg_name='pkg_foo'):
        package = self.factory.makeSoftwareItem(pkg_name)
        review_02 = self.factory.makeReview(
            software_item=package, version="0.2", summary="Review0.2")
        review_01_helpful = self.factory.makeReview(
            software_item=package, version="0.1", summary="Review0.1-helpful")
        review_02_helpful = self.factory.makeReview(
            software_item=package, version="0.2", summary="Review0.2-helpful")
        review_01 = self.factory.makeReview(
            software_item=package, version="0.1", summary="Review0.1")
        # Make sure the helpful reviews really are helpful.
        self.factory.makeUsefulness(review=review_01_helpful)
        self.factory.makeUsefulness(review=review_02_helpful)
        for review in (review_01_helpful, review_02_helpful):
            review.update_usefulness_aggregates()
            review.update_wilson_rating()

        return {
            '0.1': review_01,
            '0.1-helpful': review_01_helpful,
            '0.2': review_02,
            '0.2-helpful': review_02_helpful,
        }

    def test_review_sort_helpful(self):
        reviews = self._create_reviews_for_version_tests()
        showHandler = ShowReviewsHandler()

        response = showHandler.read(self._GET_REQUEST, 'any', 'any', 'any',
                                    'any', 'pkg_foo', '1', 'helpful')

        self.assertEqual(4, len(response))
        expected = [
            reviews['0.2-helpful'],
            reviews['0.2'],
            reviews['0.1-helpful'],
            reviews['0.1'],
        ]
        self.assertEquals(expected, list(response))

    def test_default_sort(self):
        self._create_reviews_for_version_tests()
        showHandler = ShowReviewsHandler()
        response1 = showHandler.read(self._GET_REQUEST, 'any', 'any', 'any',
                                     'any', 'pkg_foo', '1', 'helpful')
        response2 = showHandler.read(self._GET_REQUEST, 'any', 'any', 'any',
                                     'any', 'pkg_foo', '1')

        self.assertEquals(list(response1), list(response2))

    def test_review_sort_newest(self):
        reviews = self._create_reviews_for_version_tests()
        showHandler = ShowReviewsHandler()

        # sort by newest, useful votes irrelevant
        response = showHandler.read(self._GET_REQUEST, 'any', 'any', 'any',
                                    'any', 'pkg_foo', '1', 'newest')

        self.assertEqual(4, len(reviews))
        expected = [
            reviews['0.2-helpful'],
            reviews['0.2'],
            reviews['0.1'],
            reviews['0.1-helpful'],
        ]
        self.assertEquals(expected, list(response))

    @skipUnless(is_using_postgres(), "Required PostgreSQL")
    def test_debversion_sorting(self):
        package = self.factory.makeSoftwareItem('pkg_foo')
        review1 = self.factory.makeReview(
            software_item=package, version="1.0-1~lucid1",
            summary="pre-release-1.0-1~lucid1")
        review2 = self.factory.makeReview(
            software_item=package, version="1.0-1",
            summary="release-1.0-1")
        showHandler = ShowReviewsHandler()
        reviews = showHandler.read(self._GET_REQUEST, 'any', 'any', 'any',
                                   'any', 'pkg_foo', '1')

        self.assertEquals(review2, reviews[0])
        self.assertEquals(review1, reviews[1])

    def _make_5_reviews_for_package(self, package_name):
        package = self.factory.makeSoftwareItem(package_name)
        for count in range(5):
            self.factory.makeReview(software_item=package)

    def _reviews_url_for_package(self, package_name, origin='any',
                                 distro_series='any', language='any',
                                 version='any',
                                 page=None):
        kwargs = {
            'language': language,
            'origin': origin,
            'distroseries': distro_series,
            'package_name': package_name,
            'version': version,
        }
        if page is not None:
            kwargs['page'] = page

        return reverse('rnr-api-reviews-for-package', kwargs=kwargs)

    def test_only_first_pagination_returned(self):
        # The default request will return only the first batch of
        # reviews.
        self._make_5_reviews_for_package('package_foo')

        with patch_settings(API_REVIEWS_PAGINATION=3):
            response = self.client.get(self._reviews_url_for_package(
                'package_foo'))

        response_decoded = simplejson.loads(response.content)
        self.assertEqual(3, len(response_decoded))

    def test_request_second_page(self):
        # Subsequent pages of results are returned when requested.
        self._make_5_reviews_for_package('package_foo')

        with patch_settings(API_REVIEWS_PAGINATION=3):
            response = self.client.get(self._reviews_url_for_package(
                'package_foo', page=2))

        response_decoded = simplejson.loads(response.content)
        self.assertEqual(2, len(response_decoded))

    def test_request_cached(self):
        # Requests for reviews are cached.
        url = self._reviews_url_for_package('package_foo')
        response = self.client.get(url)

        self.assertEqual(
            'max-age={0}'.format(settings.CACHE_MIDDLEWARE_SECONDS),
            response['cache-control'])
        key = cache_key_for_url(url)
        self.assertTrue(key in cache)

        # A subsequent call will not hit the view.
        read_func = 'reviewsapp.api.handlers.ShowReviewsHandler.read'
        with patch(read_func) as mock_read:
            response = self.client.get(
                self._reviews_url_for_package('package_foo'))
            self.assertFalse(mock_read.called)

    def test_review_from_id(self):
        review = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_bar'))
        url = reverse(
            'rnr-api-review-for-review-id',
            kwargs={'review_id': review.id})
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        result_review = simplejson.loads(response.content)
        self.assertEqual(result_review["package_name"], review.package_name())
        self.assertEqual(result_review["id"], review.id)

    def test_review_from_origin_with_period(self):
        package = self.factory.makeSoftwareItem('package_foo')
        self.factory.makeReview(
            software_item=package,
            repository=self.factory.makeRepository('myppa.official', 'hardy'),
            language='en', version='vanilla')
        self.factory.makeReview(
            software_item=package,
            repository=self.factory.makeRepository('myppa.official', 'lucid'),
            language='en', version='vanilla')

        response = self.client.get(self._reviews_url_for_package(
            'package_foo', origin='myppa.official'))
        self.assertEqual(response.status_code, 200)
        response_decoded = simplejson.loads(response.content)

        self.assertEqual(2, len(response_decoded))

    def test_reviews_without_proper_quoting(self):
        # bug=802411 We get 20 requests per day from older USC clients
        # that have not quoted the app name. The example below is from
        # oops 2038rangpur120226. We want to ensure that we do indeed
        # find the correct result rather than 500ing.
        review = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('usb-creator-kde'),
            language='ru', version='0.1',
            app_name=u"Создание+загрузочного+диска")
        # We use a placeholder for the app_name, otherwise the call to
        # the reverse url resolver below does its own unicode
        # conversions on the string and doesn't find a match.
        package_name = "usb-creator-kde;app_name_placeholder"

        example_kwargs = dict(
            language='ru',
            origin='any',
            distroseries='any',
            package_name=package_name,
            version="0.1",
        )
        url = reverse(
            'rnr-api-reviews-for-package',
            kwargs=example_kwargs)
        url = url.replace('app_name_placeholder', review.app_name)

        response = self.client.get(url)

        self.assertEqual(200, response.status_code)
        response_decoded = simplejson.loads(response.content)
        self.assertEqual(1, len(response_decoded))
        self.assertEqual(
            'usb-creator-kde', response_decoded[0]['package_name'])


class ShowReviewsSinceHandlerTestCase(TestCaseWithFactory):

    def _reviews_since_url(self, since_str):
        return '/reviews/api/1.0/reviews/since/{0}/'.format(since_str)

    def test_only_reviews_since_specified_date_are_shown(self):
        now = datetime.now(pytz.utc)
        for i in range(1, 11):
            self.factory.makeReview(date_created=now.replace(day=i))

        response = self.client.get(
            self._reviews_since_url(now.strftime("%Y-%m-05")))
        response_decoded = simplejson.loads(response.content)

        since = now.replace(day=5)
        for review in response_decoded:
            date_created = datetime.strptime(
                review['date_created'], "%Y-%m-%d %H:%M:%S")
            date_created = date_created.replace(tzinfo=pytz.utc)
            self.assertTrue(date_created.date() >= since.date())

    def test_wrong_date_is_returns_bad_request_status_code(self):
        response = self.client.get(
            self._reviews_since_url('2000-02-31'))

        self.assertEqual(response.status_code, 400)

    def test_you_can_only_request_for_up_to_30_days_in_the_past(self):
        req_date = datetime.now(pytz.utc) - timedelta(days=31)

        response = self.client.get(
            self._reviews_since_url(req_date.strftime("%Y-%m-%d")))

        self.assertEqual(response.status_code, 400)

    def test_vary_headers(self):
        now = datetime.now(pytz.utc)
        for i in range(1, 11):
            self.factory.makeReview(date_created=now.replace(day=i))
        response = self.client.get(
            self._reviews_since_url(now.strftime("%Y-%m-05")))

        self.assertEqual(200, response.status_code)
        self.assertTrue(response.has_header('vary'))
        self.assertEqual(response['vary'], 'Accept')


class SubmitReviewHandlerTestCase(TestCaseWithFactory):

    def setUp(self):
        super(SubmitReviewHandlerTestCase, self).setUp()
        self.required_data = {
            'package_name': 'inkscape',
            'summary': 'Excellent',
            'version': '0.47.0-2ubuntu2',
            'review_text': 'Plain brilliant piece of software',
            'rating': 5,
            'distroseries': 'lucid',
            'origin': 'ubuntu',
            'language': 'en',
            'arch_tag': 'i386',
        }
        self.factory.makeArch('i386')

    def _post_new_review(self, data, user=None, verify_result=True):
        # Post a review as an authenticated user.
        url = reverse('rnr-api-reviews')

        if user is None:
            user = self.factory.makeUser()
        self.client.login(username=user.username, password='test')

        # Ensure we don't hit the network for our tests.
        is_authed_fn = ('core.api.auth.SSOOAuthAuthentication.'
                        'is_authenticated')
        lp_verify_fn = ('core.utilities.WebServices.'
                        'verify_packagename_in_repository')
        sca_verify_fn = ('core.utilities.WebServices.'
                         'sca_verify_packagename_in_repository')
        with patch(is_authed_fn) as mock_is_authenticated:
            mock_is_authenticated.return_value = True
            with patch(lp_verify_fn) as mock_lp_verify_method:
                mock_lp_verify_method.return_value = verify_result
                with patch(sca_verify_fn) as mock_sca_verify_method:
                    mock_sca_verify_method.return_value = verify_result
                    response = self.client.post(
                        url, data=data, content_type='application/json')
        return response

    def test_bogus_data(self):
        # Submitting bogus data results in a bad request.
        response = self._post_new_review(simplejson.dumps({'foo': 'bar'}))

        self.assertEqual(httplib.BAD_REQUEST, response.status_code)

    def test_invalid_rating(self):
        self.required_data['rating'] = 6

        response = self._post_new_review(simplejson.dumps(self.required_data))

        self.assertEqual(httplib.BAD_REQUEST, response.status_code)

    def test_missing_args(self):
        # If required args are not included it is a bad request.
        for key in self.required_data:
            data = self.required_data.copy()
            del data[key]

            response = self._post_new_review(simplejson.dumps(data))

            self.assertEqual(httplib.BAD_REQUEST, response.status_code)

    def test_create(self):
        # Posting the required json data creates and returns the review.
        response = self._post_new_review(simplejson.dumps(self.required_data))

        self.assertEqual(httplib.OK, response.status_code)
        self.assertEqual('application/json; charset=utf-8',
                         response['content-type'])
        response_dict = simplejson.loads(response.content)
        self.assertEqual('inkscape', response_dict['package_name'])

    def test_creates_repository_and_software_item(self):
        # Submitting a review for a software item or repository of which
        # we don't yet have a record will create those records (after
        # validating that the corresponding binary exists on Launchpad).
        required_data = self.required_data
        required_data['distroseries'] = 'doesntexist'
        self.assertEqual(0, Repository.objects.filter(
            origin='ubuntu', distroseries='doesntexist').count())
        self.assertEqual(0, SoftwareItem.objects.filter(
            package_name='inkscape').count())

        self._post_new_review(simplejson.dumps(self.required_data))

        self.assertEqual(1, Repository.objects.filter(
            origin='ubuntu', distroseries='doesntexist').count())
        self.assertEqual(1, SoftwareItem.objects.filter(
            package_name='inkscape').count())

    def test_bad_repository(self):
        # If lp_verify returns false a bad request status is returned,
        # and no new repositories are created.
        required_data = self.required_data
        required_data['distroseries'] = 'doesntexist'
        self.assertEqual(0, Repository.objects.filter(
            origin='ubuntu', distroseries='doesntexist').count())
        self.assertEqual(0, SoftwareItem.objects.filter(
            package_name='inkscape').count())

        response = self._post_new_review(
            simplejson.dumps(self.required_data), verify_result=False)

        self.assertEqual(httplib.BAD_REQUEST, response.status_code)
        self.assertEqual(0, Repository.objects.filter(
            origin='ubuntu', distroseries='doesntexist').count())
        self.assertEqual(0, SoftwareItem.objects.filter(
            package_name='inkscape').count())

    def test_max_length_summary(self):
        # Posting with a summary > 80 chars results in a bad request.
        data = self.required_data
        data['summary'] = 'a' * 82

        response = self._post_new_review(simplejson.dumps(data))

        self.assertEqual(httplib.BAD_REQUEST, response.status_code)
        errors = simplejson.loads(response.content)['errors']
        self.assertEqual(['summary'], errors.keys())

    def test_max_length_text(self):
        # Posting with a review_text > 5000 chars results in a bad
        # request.
        data = self.required_data
        data['review_text'] = 'a' * 5027

        response = self._post_new_review(simplejson.dumps(data))

        self.assertEqual(httplib.BAD_REQUEST, response.status_code)
        errors = simplejson.loads(response.content)['errors']
        self.assertEqual(['review_text'], errors.keys())

    def test_stats_updated(self):
        # Posting a review updates the associated stats.
        software_item = self.factory.makeSoftwareItem(
            package_name=self.required_data['package_name'],
            ratings_total=0, ratings_average=0)

        self._post_new_review(simplejson.dumps(self.required_data))

        software_item = SoftwareItem.objects.get(id=software_item.id)
        self.assertEqual(1, software_item.ratings_total)
        self.assertEqual(5, software_item.ratings_average)

    def test_unknown_architecture(self):
        # Posting a review of an unknown architecture is a bad request.
        self.required_data['arch_tag'] = 'unknown'

        response = self._post_new_review(simplejson.dumps(self.required_data))

        errors = simplejson.loads(response.content)['errors']
        self.assertEqual(1, len(errors))
        self.assertTrue('arch_tag' in errors)

    def test_multiple_reviews_per_arch(self):
        # Posting multiple reviews for the same architecture shouldn't fail
        for i in range(3):
            response = self._post_new_review(
                simplejson.dumps(self.required_data))
            self.assertEqual(200, response.status_code)

    def test_submit_review_invalidates_cache(self):
        # Submitting a new review invalidates the show reviews cache for
        # the package.
        self.factory.makeSoftwareItem(
            package_name=self.required_data['package_name'],
            ratings_total=0, ratings_average=0)
        url_keys = ('language', 'origin', 'distroseries', 'version',
                    'package_name')
        url_kwargs = dict((key, self.required_data[key]) for key in url_keys)
        url = reverse('rnr-api-reviews-for-package', kwargs=url_kwargs)
        self.client.get(url)
        key = cache_key_for_url(url)
        self.assertTrue(key in cache)

        self._post_new_review(simplejson.dumps(self.required_data))

        self.assertFalse(key in cache)

    def test_submit_same_app_and_distroseries_twice_fails(self):
        # A user can only submit a review for an app once.
        user = self.factory.makeUser()
        first = self._post_new_review(simplejson.dumps(self.required_data),
                                      user=user)
        second = self._post_new_review(simplejson.dumps(self.required_data),
                                       user=user)

        self.assertEqual(httplib.OK, first.status_code)
        self.assertEqual(httplib.BAD_REQUEST, second.status_code)
        errors = simplejson.loads(second.content)['errors']
        self.assertEqual(['__all__'], errors.keys())

    def test_submit_same_app_differente_series_ok(self):
        # A user can submit multiple reviews for the same app in
        # different distroseries.
        user = self.factory.makeUser()
        first = self._post_new_review(simplejson.dumps(self.required_data),
                                      user=user)
        data_natty = self.required_data
        data_natty.update(distroseries='natty')
        second = self._post_new_review(simplejson.dumps(data_natty),
                                       user=user)

        self.assertEqual(httplib.OK, first.status_code)
        self.assertEqual(httplib.OK, second.status_code)


class ModifyReviewHandlerTestCase(TestCaseWithFactory):

    post_data = {'rating': 4, 'summary': 'summary', 'review_text': 'text'}
    bad_post_data = {'summary': 'summary', 'review_text': 'text'}
    oversize_post_data = {'rating': 4, 'summary': 'a' * 82,
                          'review_text': 'text'}

    def _modify_review_id(self, review_id, post_data, user=None):
        if user is None:
            user = self.factory.makeUser()
        request = Mock()
        request.POST = post_data
        request.user = user
        handler = ModifyReviewHandler()
        return handler.update(request, review_id)

    def test_modify_review(self):
        review = self.factory.makeReview()
        response = self._modify_review_id(review.id, self.post_data,
                                          review.reviewer)
        Review.objects.get(id=review.id)
        self.assertEqual(review.id, response.id)
        self.assertEqual(self.post_data['rating'], response.rating)
        self.assertEqual(self.post_data['summary'], response.summary)
        self.assertEqual(self.post_data['review_text'], response.review_text)

    def test_non_existent_review(self):
        review = self.factory.makeReview()
        self.assertRaises(Http404, self._modify_review_id, review.id + 1,
                          self.post_data, review.reviewer)

    def test_wrong_user(self):
        review = self.factory.makeReview()
        wrong_user = self.factory.makeUser(username='modifier')
        self.assertRaises(Http404, self._modify_review_id, review.id,
                          self.post_data, wrong_user)

    def test_invalid_data_modify_fails(self):
        review = self.factory.makeReview()
        # Missing parameter
        response = self._modify_review_id(review.id, self.bad_post_data,
                                          review.reviewer)
        self.assertEqual(httplib.BAD_REQUEST, response.status_code)
        # Overlength field
        response = self._modify_review_id(review.id, self.oversize_post_data,
                                          review.reviewer)
        self.assertEqual(httplib.BAD_REQUEST, response.status_code)

    def test_review_awaiting_moderation_fails(self):
        review_moderation = self.factory.makeReviewModeration()
        review = review_moderation.review
        response = self._modify_review_id(review.id, self.post_data,
                                          review.reviewer)
        self.assertEqual(httplib.FORBIDDEN, response.status_code)

    def test_modify_time_limits(self):
        old_time = (datetime.now(pytz.utc) -
                    timedelta(settings.MODIFY_WINDOW_MINUTES + 60))
        review = self.factory.makeReview(date_created=old_time)
        response = self._modify_review_id(
            review.id, self.post_data, review.reviewer)
        self.assertEqual(httplib.FORBIDDEN, response.status_code)

    def test_stats_update_on_modify(self):
        software_item = self.factory.makeSoftwareItem(
            package_name='compiz-core',
            ratings_total=0,
            ratings_average=0)
        review = self.factory.makeReview(software_item=software_item, rating=2)
        self.factory.makeReview(
            software_item=software_item, rating=4)
        software_item = SoftwareItem.objects.get(id=software_item.id)

        # 2 reviews with ratings of 2 + 4 = 6 (average of 3)
        self.assertEqual(2, software_item.ratings_total)
        self.assertEqual(3, software_item.ratings_average)

        self._modify_review_id(review.id, self.post_data, review.reviewer)
        software_item = SoftwareItem.objects.get(id=software_item.id)

        # 2 reviews with ratings of 4 + 4 = 8 (average of 4)
        self.assertEqual(2, software_item.ratings_total)
        self.assertEqual(4, software_item.ratings_average)


class DeleteReviewHandlerTestCase(TestCaseWithFactory):
    def _delete_review_id(self, review_id, user=None):
        if user is None:
            user = self.factory.makeUser()
        request = Mock()
        request.POST = None
        request.user = user
        handler = DeleteReviewHandler()
        return handler.create(request, review_id)

    def test_delete_review(self):
        review = self.factory.makeReview()
        response = self._delete_review_id(review.id, review.reviewer)
        self.assertEqual(review.id, response.id)
        self.assertEqual(True, response.hide)
        self.assertNotEqual(None, response.date_deleted)

    def test_non_existent_review(self):
        review = self.factory.makeReview()
        self.assertRaises(
            Http404, self._delete_review_id, review.id + 1, review.reviewer)

    def test_already_deleted(self):
        review = self.factory.makeReview()

        response_good = self._delete_review_id(review.id, review.reviewer)
        self.assertEqual(review, response_good)
        # Try to delete a second time and check it raises a 404 error
        self.assertRaises(
            Http404, self._delete_review_id, review.id, review.reviewer)

    def test_wrong_user(self):
        review = self.factory.makeReview()
        wrong_user = self.factory.makeUser(username='deleter')
        self.assertRaises(
            Http404, self._delete_review_id, review.id, wrong_user)

    def test_delete_updates_stats(self):
        software_item = self.factory.makeSoftwareItem(
            package_name='compiz-core',
            ratings_total=0,
            ratings_average=0)
        self.factory.makeReview(software_item=software_item, rating=5)
        review = self.factory.makeReview(software_item=software_item,
                                         rating=3)
        software_item = SoftwareItem.objects.get(id=software_item.id)

        # 2 reviews with ratings of 5 + 3 = 8 (average of 4)
        self.assertEqual(2, software_item.ratings_total)
        self.assertEqual(4, software_item.ratings_average)

        self._delete_review_id(review.id, review.reviewer)
        software_item = SoftwareItem.objects.get(id=software_item.id)

        # 1 review with rating 5 = average of 5
        self.assertEqual(1, software_item.ratings_total)
        self.assertEqual(5, software_item.ratings_average)


class FlagReviewHandlerTestCase(TestCaseWithFactory):
    def _flag_review_id(self, review_id, user=None, data=None):
        if user is None:
            user = self.factory.makeUser()
        if data is None:
            data = dict(reason='A reason', text='explanation of the flag')
        request = Mock()
        request.POST = data
        request.user = user
        handler = FlagReviewHandler()
        return handler.create(request, review_id)

    def test_bogus_data(self):
        review = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'),
            app_name='app_foo',
            repository=self.factory.makeRepository('ubuntu', 'hardy'),
            language='en')

        response = self._flag_review_id(review.id, data={'foo': 'bazzinga'})

        self.assertEqual(httplib.BAD_REQUEST, response.status_code)

    def test_bad_review_id(self):
        response = self._flag_review_id(1234)

        self.assertEqual(httplib.NOT_FOUND, response.status_code)

    def test_create(self):
        review = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package-foo'),
            app_name=u'Déjà Dup',
            repository=self.factory.makeRepository('ubuntu', 'hardy'),
            language='en')
        flagger = self.factory.makeUser()

        flag = self._flag_review_id(review.id, user=flagger)

        self.assertTrue(isinstance(flag, ReviewModerationFlag))
        self.assertEqual(flagger, flag.user)
        self.assertEqual(review, flag.review_moderation.review)

    def test_flag_review_invalidates_cache(self):
        # Flagging a new review invalidates the show reviews cache for
        # the package only if the review visibility changes.
        review = self.factory.makeReview()
        url_kwargs = dict(
            language=review.language,
            origin=review.repository.origin,
            distroseries=review.repository.distroseries,
            package_name=review.softwareitem.package_name,
            version=review.version,
        )
        url = reverse('rnr-api-reviews-for-package', kwargs=url_kwargs)
        self.client.get(url)
        key = cache_key_for_url(url)
        self.assertTrue(key in cache)

        self._flag_review_id(review.id)

        self.assertFalse(key in cache)


class UsefulnessHandlerTestCase(TestCaseWithFactory):

    def test_vote_on_usefulness(self):
        review = self.factory.makeReview()
        self.assertEqual(review.usefulness_total, 0)
        self.assertEqual(review.usefulness_favorable, 0)
        self.assertEqual(review.usefulness_percentage, 0)
        self.assertEqual(review.usefulness_wilson_rating, 0)
        request = Mock()
        request.user = self.factory.makeUser()

        response = SubmitUsefulnessHandler().create(request, review.id)

        self.assertEquals(response.status_code, 201)
        review = Review.objects.get(id=review.id)
        self.assertEqual(review.usefulness_total, 1)
        self.assertEqual(review.usefulness_favorable, 1)
        self.assertEqual(review.usefulness_percentage, 100)
        self.assertAlmostEqual(
            review.usefulness_wilson_rating, wilson_score(1, 1))

    def test_usefulness_in_review(self):
        review = self.factory.makeReview()
        request = Mock()
        request.user = self.factory.makeUser()
        handler = SubmitUsefulnessHandler()
        response = handler.create(request, review.id)
        self.assertEquals(response.status_code, 201)
        url = reverse(
            'rnr-api-reviews-for-package',
            kwargs=_get_url_kwargs_dict_for_review(review))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        reviews = simplejson.loads(response.content)
        self.assertEqual(reviews[0]["usefulness_total"], 1)
        self.assertEqual(reviews[0]["usefulness_favorable"], 1)

    def test_one_vote_per_user(self):
        review = self.factory.makeReview()
        request = Mock()
        request.user = self.factory.makeUser()
        handler = SubmitUsefulnessHandler()
        response = handler.create(request, review.id)
        self.assertEquals(response.status_code, 201)
        review = Review.objects.get(id=review.id)
        self.assertEqual(review.usefulness_total, 1)
        self.assertEqual(review.usefulness_favorable, 1)
        self.assertEqual(review.usefulness_percentage, 100)
        # same user
        response = handler.create(request, review.id)
        self.assertEquals(response.status_code, 200)
        self.assertEquals(response.content, '"Not modified"')
        review = Review.objects.get(id=review.id)
        self.assertEqual(review.usefulness_total, 1)
        self.assertEqual(review.usefulness_favorable, 1)
        self.assertEqual(review.usefulness_percentage, 100)
        # same user
        request.POST = {'useful': 'False'}
        response = handler.create(request, review.id)
        self.assertEquals(response.status_code, 200)
        self.assertEquals(response.content, '"Updated"')
        review = Review.objects.get(id=review.id)
        self.assertEqual(review.usefulness_total, 1)
        self.assertEqual(review.usefulness_favorable, 0)
        self.assertEqual(review.usefulness_percentage, 0)

    def test_vote_aggregation(self):
        review = self.factory.makeReview()
        request = Mock()
        request.user = self.factory.makeUser()
        handler = SubmitUsefulnessHandler()
        response = handler.create(request, review.id)
        self.assertEquals(response.status_code, 201)
        review = Review.objects.get(id=review.id)
        self.assertEqual(review.usefulness_total, 1)
        self.assertEqual(review.usefulness_favorable, 1)
        self.assertEqual(review.usefulness_percentage, 100)
        self.assertAlmostEqual(
            review.usefulness_wilson_rating, wilson_score(1, 1))
        # new user
        request.user = self.factory.makeUser()
        response = handler.create(request, review.id)
        self.assertEquals(response.status_code, 201)
        review = Review.objects.get(id=review.id)
        self.assertEqual(review.usefulness_total, 2)
        self.assertEqual(review.usefulness_favorable, 2)
        self.assertEqual(review.usefulness_percentage, 100)
        self.assertAlmostEqual(
            review.usefulness_wilson_rating, wilson_score(2, 2))
        # another new user
        request.user = self.factory.makeUser()
        request.POST = {'useful': 'False'}
        response = handler.create(request, review.id)
        self.assertEquals(response.status_code, 201)
        review = Review.objects.get(id=review.id)
        self.assertEqual(review.usefulness_total, 3)
        self.assertEqual(review.usefulness_favorable, 2)
        self.assertEqual(review.usefulness_percentage, 66)
        self.assertAlmostEqual(
            review.usefulness_wilson_rating, wilson_score(2, 3))

    def _post_usefulness_as_user(self, user, useful=True):
        review = self.factory.makeReview()
        url = reverse('rnr-api-recommendations', args=[review.id])
        self.client.login(username=user.username, password='test')
        is_authenticated_method = (
            'core.api.auth.SSOOAuthAuthentication.is_authenticated')
        with patch(is_authenticated_method) as mock_is_authenticated:
            mock_is_authenticated.return_value = True
            response = self.client.post(url, data={'useful': useful})
        return response

    def test_request_not_cached(self):
        # Requests for server status are never cached.
        user = self.factory.makeUser()

        response = self._post_usefulness_as_user(user)

        self.assertFalse(response.has_header('cache-control'))

        # A subsequent call will hit the view.
        create_method = (
            'reviewsapp.api.handlers.SubmitUsefulnessHandler.create')
        with patch(create_method) as mock_create:
            response = self._post_usefulness_as_user(user)
            self.assertTrue(mock_create.called)
