#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for our middleware classes"""

from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'GraphiteMiddlewareTestCase',
]

from django.contrib.auth.models import Permission
from django.core.urlresolvers import reverse
from mock import patch

from reviewsapp.tests.factory import TestCaseWithFactory
from reviewsapp.tests.helpers import patch_settings


class GraphiteMiddlewareTestCase(TestCaseWithFactory):

    def setUp(self):
        super(GraphiteMiddlewareTestCase, self).setUp()
        self.factory.makeReviewModeration()

    def login_as_moderator(self):
        user = self.factory.makeUser()
        can_moderate = Permission.objects.get(codename='can_moderate')
        user.user_permissions.add(can_moderate)
        self.client.login(username=user.username, password='test')

    def assert_stats_sent(self, keys, mock_send):
        for key in keys:
            self.assert_stat_sent(key, mock_send)

    def assert_stat_sent(self, key, mock_send):
        keys = [x[0][0] for x in mock_send.call_args_list]
        found = False
        for k in keys:
            if key in k:
                found = True
                break
        self.assertTrue(found)

    @patch('txstatsd.client.UdpStatsDClient.write')
    def test_stats_are_sent(self, mock_send):

        with patch_settings(ENABLE_GRAPHITE_STATISTICS=True):
            self.client.get(reverse('rnr-reviewmoderation-list'))

        keys = ['.requests', '.views.moderation.reviewmoderation_list',
                '.status_codes.302']

        self.assert_stats_sent(keys, mock_send)

    @patch('reviewsapp.models.ModerationManager.filter')
    @patch('txstatsd.client.UdpStatsDClient.write')
    def test_exception_stats_are_sent(self, mock_send, mock_filter):
        mock_filter.side_effect = ZeroDivisionError()
        self.login_as_moderator()

        with patch_settings(ENABLE_GRAPHITE_STATISTICS=True):
            self.assertRaises(
                ZeroDivisionError,
                self.client.get, reverse('rnr-reviewmoderation-list'))

        key = '.exceptions.ZeroDivisionError'
        self.assert_stat_sent(key, mock_send)
