#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Ubuntu App Review model tests."""

from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'ArchitectureTestCase',
    'ReviewModelTestCase',
    'ReviewModerationTestCase',
    'ReviewFlagTestCase',
    'RNRSettingsTestCase',
    'SoftwareItemInOriginTestCase',
    'SoftwareItemInRepositoryTestCase',
    'SoftwareItemTestCase',
]

from django.db import IntegrityError

from core.models import ModerationError, UserAlreadyFlaggedReview
from reviewsapp.forms import ReviewForm
from reviewsapp.models import (
    Architecture,
    Review,
    ReviewModeration,
    ReviewModerationFlag,
    RNRSettings,
    SoftwareItem,
    SoftwareItemInOrigin,
    SoftwareItemInRepository,
)
from reviewsapp.tests.factory import TestCaseWithFactory
from reviewsapp.utilities import (
    is_using_postgres,
    wilson_score,
)


class ArchitectureTestCase(TestCaseWithFactory):

    def test_architecture_unique(self):
        # An architecture must be unique.
        arch_tag = self.factory.getUniqueString(prefix='arch-')
        Architecture.objects.create(tag=arch_tag)

        self.assertRaises(
            IntegrityError, Architecture.objects.create, tag=arch_tag)


class ReviewModelTestCase(TestCaseWithFactory):

    def test_max_lengths(self):
        """ Check max_length of 80 for summary and 5000 for text, per spec"""
        # We use a form to test the model here because Django doesn't enforce
        # max_length at the model level.  We could rely on the database to
        # complain, except sqlite doesn't :)

        user = self.factory.makeUser()
        software_item = self.factory.makeSoftwareItem()
        version = self.factory.makeVersion()
        repo = self.factory.makeRepository()
        form_data = {
            'reviewer': user.pk,
            'softwareitem': software_item.pk,
            'repository': repo.pk,
            'version': version,
            'rating': 4,
            'summary': 'a' * 82,
            'review_text': 'Up to standard',
            'language': 'en',
            'hide': False,
        }
        form = ReviewForm(form_data)

        self.assertFailsValidation(form, 'summary')

        form_data.update({'summary': 'Nice', 'review_text': 'b' * 5010})
        form = ReviewForm(form_data)

        self.assertFailsValidation(form, 'review_text')

        form_data.update({'version': 'x' * 101})
        form = ReviewForm(form_data)

        self.assertFailsValidation(form, 'version')

    def assertFailsValidation(self, form, field):
        self.assertFalse(form.is_valid(), 'form should fail validation')
        self.assertTrue(field in form.errors)

    def test_pending_moderation(self):
        review = self.factory.makeReview()
        result = review.is_awaiting_moderation()
        self.assertEqual(False, result)

        review_moderation = self.factory.makeReviewModeration()
        review = review_moderation.review
        result = review.is_awaiting_moderation()
        self.assertEqual(True, result)

    def test_update_wilson(self):
        review = self.factory.makeReview()
        # mock 3 useful votes out of 5 votes total, withouy associated
        # wilson_rating update
        review.usefulness_favorable = 3
        review.usefulness_total = 5
        review.usefulness_wilson_rating = 0.0

        review.update_wilson_rating()
        self.assertEqual(wilson_score(3, 5), review.usefulness_wilson_rating)

    def test_update_usefulness(self):
        review = self.factory.makeReview()
        self.factory.makeUsefulness(review=review, useful=True)
        self.factory.makeUsefulness(review=review, useful=False)
        review.usefulness_favorable = 0
        review.usefulness_total = 0
        review.usefulness_percentage = 0

        review.update_usefulness_aggregates()

        self.assertEqual(2, review.usefulness_total)
        self.assertEqual(1, review.usefulness_favorable)
        self.assertEqual(50, review.usefulness_percentage)


class ReviewFlagTestCase(TestCaseWithFactory):

    def test_flag_object_returns_reviewmoderationflag(self):
        # The new flag is returned when Review.flag() is called.
        review = self.factory.makeReview()

        flag = review.flag(self.factory.makeUser(), 'summary', 'description')

        self.assertTrue(isinstance(flag, ReviewModerationFlag))

    def test_flag_object_creates_new_pending_moderation(self):
        # Flagging an object creates a new pending moderation
        # if a pending moderation doesn't exist already.
        review_moderation = self.factory.makeReviewModeration(
            status=ReviewModeration.APPROVED_STATUS)
        review = review_moderation.review
        flagger = self.factory.makeUser()
        self.assertEqual(1, review.reviewmoderation_set.all().count())

        review.flag(
            flagger, 'Offensive language', "The word 'blue' is offensive.")

        self.assertEqual(2, review.reviewmoderation_set.all().count())

    def test_flag_object_appends_to_existing_pending_moderation(self):
        # Flagging an object that already has a pending moderation will
        # just create a new flag for the existing pending moderation.
        review_moderation = self.factory.makeReviewModeration()
        review = review_moderation.review
        flagger = self.factory.makeUser()
        self.assertEqual(1, review.reviewmoderation_set.all().count())

        review.flag(
            flagger, 'Offensive language', "The word 'blue' is offensive.")

        self.assertEqual(1, review.reviewmoderation_set.all().count())

    def test_flag_object_records_details(self):
        review = self.factory.makeReview()
        flagger = self.factory.makeUser()

        review.flag(
            flagger, 'Offensive language', "The word 'blue' is offensive.")

        pending_moderations = review.reviewmoderation_set.pending()
        flags = pending_moderations[0].reviewmoderationflag_set.all()
        self.assertEqual(1, flags.count())
        flag = flags[0]
        self.assertEqual('Offensive language', flag.summary)
        self.assertEqual("The word 'blue' is offensive.", flag.description)
        self.assertEqual(flagger, flag.user)

    def test_flag_hides_unmoderated_items(self):
        # Flagging an item which has not previously been approved
        # will hide that item.
        review = self.factory.makeReview()
        stats_updated_orig = review.softwareitem.date_ratings_updated

        self.assertFalse(review.hide)
        flagger = self.factory.makeUser()

        review.flag(
            flagger, 'Offensive language', "The word 'blue' is offensive.")

        self.assertTrue(review.hide)
        # The stats were updated for the software item.
        softwareitem = SoftwareItem.objects.get(id=review.softwareitem.id)
        self.assertNotEquals(
            stats_updated_orig, softwareitem.date_ratings_updated)

    def test_flag_does_not_hide_approved_items(self):
        # Flagging an item which has been previously approved does not
        # hide that item.
        review_moderation = self.factory.makeReviewModeration(
            status=ReviewModeration.APPROVED_STATUS, num_flags=0)
        review = review_moderation.review
        stats_updated_orig = review.softwareitem.date_ratings_updated
        flagger = self.factory.makeUser()
        self.assertFalse(review.hide)

        review.flag(
            flagger, 'Offensive language', "The word 'blue' is offensive.")

        self.assertFalse(review.hide)
        # The stats were not updated for the software item.
        self.assertEquals(
            stats_updated_orig, review.softwareitem.date_ratings_updated)

    def test_multiple_flag_ordering(self):
        # When an object is flagged multiple times default ordering is
        # newest first.
        review = self.factory.makeReview()
        flagger1 = self.factory.makeUser()
        flagger2 = self.factory.makeUser()

        review.flag(
            flagger1, 'Offensive language', "The word 'blue' is offensive.")
        review.flag(
            flagger2, 'Spam links',
            "The link to http://example.com is spam.")

        pending_moderation = review.reviewmoderation_set.pending()[0]
        flagged = pending_moderation.reviewmoderationflag_set.all()
        flag_summaries = list(flagged.values_list('summary', flat=True))
        expected = [u'Spam links', u'Offensive language']
        self.assertEqual(expected, flag_summaries)

    def test_flag_once_only_for_user(self):
        # It is not possible for the same person to flag an object twice.
        review = self.factory.makeReview()
        flagger = self.factory.makeUser()

        review.flag(
            flagger, 'Offensive language', "The word 'blue' is offensive.")

        self.assertRaises(
            UserAlreadyFlaggedReview,
            review.flag, flagger, 'Offensive language',
            "The word 'blue' is offensive.")


class ReviewModerationTestCase(TestCaseWithFactory):

    def test_default_status_pending(self):
        # The default status for a moderated object is pending.
        review_moderation = ReviewModeration.objects.create(
            review=self.factory.makeReview())

        self.assertEqual(
            ReviewModeration.PENDING_STATUS, review_moderation.status)

    def test_moderate_updates_moderation_fields(self):
        # The moderate method sets moderation fields appropriately.
        review_moderation = ReviewModeration.objects.create(
            review=self.factory.makeReview())
        moderator = self.factory.makeUser()

        review_moderation.moderate(
            moderator, ReviewModeration.APPROVED_STATUS, "Review text")

        self.assertEqual(moderator, review_moderation.moderator)
        self.assertEqual(ReviewModeration.APPROVED_STATUS,
                         review_moderation.status)
        self.assertEqual("Review text", review_moderation.moderation_text)
        self.assertIsNot(None, review_moderation.date_moderated)

    def test_moderate_approval_unhides_review(self):
        # Moderating with approval unhides a review.
        review_moderation = self.factory.makeReviewModeration(num_flags=1)
        review = review_moderation.review
        stats_updated_orig = review.softwareitem.date_ratings_updated
        self.assertTrue(review.hide)

        review_moderation.moderate(
            self.factory.makeUser(), ReviewModeration.APPROVED_STATUS,
            "whatever")

        review = Review.objects.get(id=review.id)
        self.assertFalse(review.hide)
        # The stats were updated for the software item.
        self.assertNotEquals(
            stats_updated_orig, review.softwareitem.date_ratings_updated)

    def test_moderate_rejection_hides_review(self):
        # Moderating with rejection unhides a review.
        review_moderation = self.factory.makeReviewModeration(num_flags=0)
        review = review_moderation.review
        stats_updated_orig = review.softwareitem.date_ratings_updated
        self.assertFalse(review.hide)

        review_moderation.moderate(
            self.factory.makeUser(), ReviewModeration.REJECTED_STATUS,
            "whatever")

        review = Review.objects.get(id=review.id)
        # The stats were updated for the software item.
        self.assertNotEquals(
            stats_updated_orig, review.softwareitem.date_ratings_updated)
        self.assertTrue(review.hide)

    def test_moderate_non_pending_moderation_errors(self):
        # A moderation that is not pending cannot be moderated.
        review_moderation = self.factory.makeReviewModeration(
            status=ReviewModeration.REJECTED_STATUS)

        self.assertRaises(
            ModerationError,
            review_moderation.moderate, self.factory.makeUser(),
            ReviewModeration.REJECTED_STATUS, "Second try")

    def test_moderate_to_pending_errors(self):
        # The status cannot be left PENDING during moderate.
        review_moderation = self.factory.makeReviewModeration()

        self.assertRaises(
            ModerationError,
            review_moderation.moderate, self.factory.makeUser(),
            ReviewModeration.PENDING_STATUS, "Leave pending")

    def test_moderate_own_review(self):
        # It is not possible to moderate your own review that has been
        # flagged.
        review_moderation = self.factory.makeReviewModeration()
        reviewer = review_moderation.review.reviewer

        self.assertFalse(review_moderation.can_be_moderated_by(reviewer))

    def test_moderate_own_flag(self):
        # It is not possible to moderate a review which you have flagged
        # yourself.
        review_moderation = self.factory.makeReviewModeration()
        flagger = self.factory.makeUser()
        review_moderation.review.flag(flagger, "Inappropriate", "This is bad.")

        self.assertFalse(review_moderation.can_be_moderated_by(flagger))

    def test_moderate_not_allowed_errors(self):
        # Moderating a review by a moderator who is not allowed will
        # error.
        review_moderation = self.factory.makeReviewModeration()
        reviewer = review_moderation.review.reviewer

        self.assertRaises(
            ModerationError,
            review_moderation.moderate, reviewer,
            ReviewModeration.APPROVED_STATUS, "Great.")


class SoftwareItemTestCase(TestCaseWithFactory):

    def test_update_stats(self):
        # Updating the stats calculates the average, total and histogram.
        software_item = self.factory.makeSoftwareItem()
        self.factory.makeReview(
            software_item=software_item, rating="5")
        self.factory.makeReview(
            software_item=software_item, rating="3")

        software_item.update_stats()

        self.assertEqual('4.0', software_item.ratings_average)
        self.assertEqual(2, software_item.ratings_total)
        self.assertEqual("[0, 0, 1, 0, 1]", software_item.histogram)

    def test_update_stats_ignores_hidden(self):
        # Updating the stats ignores hidden reviews.
        software_item = self.factory.makeSoftwareItem()
        self.factory.makeReview(
            software_item=software_item, rating="5")
        self.factory.makeReview(
            software_item=software_item, rating="3", hide=True)

        software_item.update_stats()

        self.assertEqual('5.0', software_item.ratings_average)
        self.assertEqual(1, software_item.ratings_total)
        self.assertEqual("[0, 0, 0, 0, 1]", software_item.histogram)

    def test_update_stats_ignores_incorrect(self):
        software_item = self.factory.makeSoftwareItem()
        self.factory.makeReview(
            software_item=software_item, rating="5")
        self.factory.makeReview(
            software_item=software_item, rating="6")
        self.factory.makeReview(
            software_item=software_item, rating="-1")

        software_item.update_stats()

        self.assertEqual('5.0', software_item.ratings_average)
        self.assertEqual(1, software_item.ratings_total)
        self.assertEqual("[0, 0, 0, 0, 1]", software_item.histogram)


class RNRSettingsTestCase(TestCaseWithFactory):

    def test_get_settings_creates_settings(self):
        self.assertEqual(0, RNRSettings.objects.count())

        RNRSettings.get_settings()

        self.assertEqual(1, RNRSettings.objects.count())

    def test_get_settings_does_not_create_second_setting(self):
        RNRSettings.get_settings()
        self.assertEqual(1, RNRSettings.objects.count())

        RNRSettings.get_settings()

        self.assertEqual(1, RNRSettings.objects.count())

    def test_set_moderation_mode(self):
        self.assertEqual(
            RNRSettings.PASSIVE_MODERATION,
            RNRSettings.get_settings().moderation_mode)

        RNRSettings.set_moderation_mode(RNRSettings.BLACKLIST_MODERATION)

        self.assertEqual(
            RNRSettings.BLACKLIST_MODERATION,
            RNRSettings.get_settings().moderation_mode)

    def test_should_moderate_text_true_for_active(self):
        RNRSettings.set_moderation_mode(RNRSettings.ACTIVE_MODERATION)

        self.assertTrue(RNRSettings.should_moderate_text('foo'))

    def test_should_moderate_text_false_for_passive(self):
        RNRSettings.set_moderation_mode(RNRSettings.PASSIVE_MODERATION)

        self.assertFalse(RNRSettings.should_moderate_text('foo'))

    def _set_blacklist(self, blacklist):
        RNRSettings.set_moderation_mode(RNRSettings.BLACKLIST_MODERATION)
        settings = RNRSettings.get_settings()
        settings.blacklist_words = "\n".join(blacklist)
        settings.save()

    def test_should_moderate_true_for_blacklisted_word(self):
        self._set_blacklist(['foi', 'foo', 'baz'])

        self.assertTrue(RNRSettings.should_moderate_text(
            'There once was a', 'boy with a little ', 'foo drum.'))

    def test_should_moderate_false_for_non_blacklisted_word(self):
        self._set_blacklist(['foi', 'foo', 'baz'])

        self.assertFalse(RNRSettings.should_moderate_text(
            'There once was a boy with a little toy drum.'))

    def test_get_moderation_summary_and_desc_passive(self):
        RNRSettings.set_moderation_mode(RNRSettings.PASSIVE_MODERATION)

        summary, description = RNRSettings.get_moderation_summary_and_desc()

        self.assertEqual(None, summary)
        self.assertEqual(None, description)

    def test_get_moderation_summary_and_desc_active(self):
        RNRSettings.set_moderation_mode(RNRSettings.ACTIVE_MODERATION)

        summary, description = RNRSettings.get_moderation_summary_and_desc()

        self.assertEqual('Automatically Flagged', summary)
        self.assertTrue(
            description.startswith('This review was automatically'))

    def test_get_moderation_summary_and_desc_blacklist(self):
        RNRSettings.set_moderation_mode(RNRSettings.BLACKLIST_MODERATION)

        summary, description = RNRSettings.get_moderation_summary_and_desc()

        self.assertEqual('Contains Blacklisted Word', summary)
        self.assertTrue(
            description.startswith('This review was flagged because the'))


class SoftwareItemInRepositoryTestCase(TestCaseWithFactory):

    def test_uniqueness(self):
        # Django 1.1 doesn't raise Integrity errors for sqlite.
        if is_using_postgres():
            item = self.factory.makeSoftwareItemInRepository()

            self.assertRaises(
                IntegrityError,
                self.factory.makeSoftwareItemInRepository,
                software_item=item.softwareitem, repository=item.repository)

    def test_update_stats(self):
        item = self.factory.makeSoftwareItemInRepository()
        for rating in [3, 4, 5]:
            self.factory.makeReview(
                software_item=item.softwareitem,
                repository=item.repository, rating=rating)
        # One other review which should not be included.
        self.factory.makeReview(
            software_item=item.softwareitem, rating=5)

        item.update_stats()

        item_reloaded = SoftwareItemInRepository.objects.get(id=item.id)
        self.assertEqual(3, item_reloaded.ratings_total)
        self.assertEqual(4, item_reloaded.ratings_average)
        self.assertEqual("[0, 0, 1, 1, 1]", item_reloaded.histogram)

    def test_update_stats_triggers_other_updates(self):
        # Calling update_stats on the SoftwareItemInRepository ensures
        # that the other aggregate stats are updated also.
        item_in_repo = self.factory.makeSoftwareItemInRepository()
        item_in_origin = self.factory.makeSoftwareItemInOrigin(
            origin=item_in_repo.repository.origin,
            software_item=item_in_repo.softwareitem)
        for rating in [3, 4, 5]:
            self.factory.makeReview(
                software_item=item_in_repo.softwareitem,
                repository=item_in_repo.repository)

        item_in_repo.update_stats()

        item_in_repo = SoftwareItemInRepository.objects.get(
            id=item_in_repo.id)
        item_in_origin = SoftwareItemInOrigin.objects.get(
            id=item_in_origin.id)
        self.assertEqual(3, item_in_repo.ratings_total)
        self.assertEqual(3, item_in_origin.ratings_total)
        self.assertEqual(3, item_in_repo.softwareitem.ratings_total)


class SoftwareItemInOriginTestCase(TestCaseWithFactory):

    def test_uniqueness(self):
        # Django 1.1 doesn't raise Integrity errors for sqlite.
        if is_using_postgres():
            item = self.factory.makeSoftwareItemInOrigin()

            self.assertRaises(
                IntegrityError,
                self.factory.makeSoftwareItemInOrigin,
                software_item=item.softwareitem, origin=item.origin)

    def test_update_stats(self):
        repo1 = self.factory.makeRepository(origin='ubuntu')
        repo2 = self.factory.makeRepository(origin='ubuntu')
        repo3 = self.factory.makeRepository(origin='canonical')

        item = self.factory.makeSoftwareItemInOrigin(origin='ubuntu')
        self.factory.makeReview(
            software_item=item.softwareitem, repository=repo1, rating=5)
        self.factory.makeReview(
            software_item=item.softwareitem, repository=repo2, rating=3)
        self.factory.makeReview(
            software_item=item.softwareitem, repository=repo3, rating=1)
        item.update_stats()

        item_reloaded = SoftwareItemInOrigin.objects.get(id=item.id)
        self.assertEqual(2, item_reloaded.ratings_total)
        self.assertEqual(4, item_reloaded.ratings_average)
        self.assertEqual("[0, 0, 1, 0, 1]", item_reloaded.histogram)
