#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the RnR api client module."""

from __future__ import (
    absolute_import,
    with_statement,
)

__metaclass__ = type
__all__ = [
    'RnRFlagReviewTestCase',
    'RnRGetReviewTestCase',
    'RnRGetReviewsTestCase',
    'RnRReviewStatsTestCase',
    'RnRServerStatusTestCase',
    'RnRSubmitReviewTestCase',
    'RnRModifyReviewTestCase',
    'RnRDeleteReviewTestCase',
    'UsefulnessAPITestCase',
]

from datetime import timedelta

from decimal import Decimal
from django.conf import settings
from django.contrib.auth.models import User
from django.core.cache import cache
from django.core.handlers.wsgi import WSGIHandler
from django_openid_auth.models import UserOpenID
from mock import patch

from piston_mini_client.validators import ValidationException
from piston_mini_client.auth import OAuthAuthorizer
from rnrclient import RatingsAndReviewsAPI, ReviewRequest

from reviewsapp.models import (
    ReviewModeration,
    ReviewModerationFlag,
    RNRSettings,
    Usefulness,
)
from reviewsapp.tests.helpers import (
    patch_settings,
    WSGIInterceptedTestCase,
)
from reviewsapp.tests.factory import (
    TestCaseWithFactory,
    TxTestCaseWithFactory,
)
from reviewsapp.utilities import invalidate_paginated_reviews_for


class RnRClientTestCaseBase(WSGIInterceptedTestCase):

    def setUp(self):
        callbacks = {('localhost', 8000): WSGIHandler}
        super(RnRClientTestCaseBase, self).setUp(callbacks)

    def make_api_for_user(
            self, user=None, mock_get_data_returns=None, mock_verify=None):
        if user is None:
            user = self.factory.makeUser()

        consumer_id = 'consumer_key'
        email = 'foo@bar.com'

        if user is not None and not user.is_anonymous():
            # Set correct consumer id and email in case of valid user
            try:
                claimed_id = user.useropenid_set.latest('pk').claimed_id
                consumer_id = claimed_id.rsplit('/', 1)[1]
            except (UserOpenID.DoesNotExist, IndexError):
                pass
            email = user.email

        auth = OAuthAuthorizer(
            'token', 'secret', consumer_id, 'consumer_secret')

        if mock_get_data_returns is None:
            mock_get_data_returns = {
                'displayname': 'Some Body',
                'email': email,
                'verified': True,
            }

        # SSO will validate the request
        patcher = patch('core.api.auth.sso.web_services.validate_request')
        validate_request = patcher.start()
        validate_request.return_value = True
        self.addCleanup(patcher.stop)

        patcher = patch('core.utilities.web_services.get_data_for_account')
        get_data_for_account = patcher.start()
        get_data_for_account.return_value = mock_get_data_returns
        self.addCleanup(patcher.stop)

        if mock_verify is None:
            patcher = patch('reviewsapp.forms.web_services')
            mock_ws = patcher.start()
            self.addCleanup(patcher.stop)
            mock_ws.verify_packagename_in_repository.return_value = True

        return RatingsAndReviewsAPI(auth=auth)


class RnRClientTestCase(RnRClientTestCaseBase, TestCaseWithFactory):
    pass


class RnRClientTxTestCase(RnRClientTestCaseBase, TxTestCaseWithFactory):
    pass


class RnRServerStatusTestCase(RnRClientTestCase):
    def test_server_status(self):
        api = RatingsAndReviewsAPI()

        response = api.server_status()

        self.assertEqual('ok', response)


class RnRReviewStatsTestCase(RnRClientTxTestCase):

    def clear_cache(self):
        cache.clear()

    def get_stats_without_cache(self, **kwargs):
        cache.clear()
        return RatingsAndReviewsAPI().review_stats(**kwargs)

    def make_app_with_ratings(
            self, package_name, ratings=[], origin='ubuntu'):
        repo = self.factory.makeRepository(origin=origin)
        software_item = self.factory.makeSoftwareItem(
            package_name=package_name)
        for rating in ratings:
            self.factory.makeReview(
                software_item=software_item, rating=rating,
                repository=repo)
        return software_item

    def test_review_stats_single_package(self):
        self.make_app_with_ratings('SuperCoolApp', [3, 4, 5])

        stats = self.get_stats_without_cache()

        self.assertEquals(1, len(stats))  # one per software item
        self.assertEquals(3, stats[0].ratings_total)
        self.assertEquals(Decimal('4'), Decimal(stats[0].ratings_average))

    def test_review_stats_multiple_packages(self):
        self.make_app_with_ratings('SuperCoolApp', [3, 4, 5])
        self.make_app_with_ratings('SortaCoolApp', [1, 2, 3])

        stats = self.get_stats_without_cache()

        self.assertEquals(2, len(stats))  # one per software item
        self.assertEquals(3, stats[1].ratings_total)
        self.assertEquals(Decimal('2'), Decimal(stats[1].ratings_average))

        # Check stats for pkg 1 are still the same
        self.assertEquals(3, stats[0].ratings_total)
        self.assertEquals(Decimal('4'), Decimal(stats[0].ratings_average))

    def test_review_stats_updates_only(self):
        # Apps with stats updated in the last N days only are returned.
        self.make_app_with_ratings('SuperCoolApp', [3, 4, 5])
        sorta_cool_app = self.make_app_with_ratings('SortaCoolApp', [1, 2, 3])
        # SortaCoolApp had its ratings last updated 5 days ago.
        sorta_cool_app.date_ratings_updated = (
            sorta_cool_app.date_ratings_updated - timedelta(5))
        sorta_cool_app.save()

        self.assertEquals(2, len(self.get_stats_without_cache()))
        self.assertEquals(1, len(self.get_stats_without_cache(days=1)))
        self.assertEquals(1, len(self.get_stats_without_cache(days=3)))
        self.assertEquals(2, len(self.get_stats_without_cache(days=7)))

    def test_review_stats_vary_headers(self):
        """Check Vary headers is set to Accept and Accept-Encoding."""
        self.clear_cache()
        self.make_app_with_ratings('my-app', [3, 4, 5])
        self.make_app_with_ratings('my-other-app', [1, 5])
        response = self.client.get(
            '/reviews/api/1.0/review-stats/',
            HTTP_ACCEPT_ENCODING='gzip')
        self.assertEqual(response['vary'], 'Accept, Accept-Encoding')

    def test_review_stats_ppa_origin(self):
        self.make_app_with_ratings(
            'test-package', [3, 4, 5], origin='lp-ppa-app-review-board')

        stats = self.get_stats_without_cache(origin='lp-ppa-app-review-board')

        self.assertEquals(1, len(stats))

    def test_review_stats_are_gzipped(self):
        self.clear_cache()
        self.make_app_with_ratings('my-app', [3, 4, 5])
        self.make_app_with_ratings('my-other-app', [1, 5])
        response = self.client.get(
            '/reviews/api/1.0/review-stats/',
            HTTP_ACCEPT_ENCODING='gzip')
        self.assertEqual('gzip', response['content-encoding'])

    def test_stats_arent_gzipped_if_gzipping_is_disabled(self):
        self.clear_cache()
        self.make_app_with_ratings('my-app', [3, 4, 5])
        self.make_app_with_ratings('my-other-app', [1, 5])
        with patch_settings(ENABLE_GZIP_CONTENT=False):
            response = self.client.get(
                '/reviews/api/1.0/review-stats/',
                HTTP_ACCEPT_ENCODING='gzip')
        self.assertEqual(None, response.get('content-encoding', None))


class RnRGetReviewsTestCase(RnRClientTestCase):
    def test_get_reviews_for_package(self):
        review = self.factory.makeReview()
        api = RatingsAndReviewsAPI()

        reviews = api.get_reviews(
            language=review.language,
            origin=review.repository.origin,
            distroseries=review.repository.distroseries,
            packagename=review.package_name()
        )

        self.assertEqual(1, len(reviews))
        self.assertEqual(review.package_name(), reviews[0].package_name)
        self.assertEqual(review.app_name, reviews[0].app_name)
        self.assertEqual(review.summary, reviews[0].summary)
        self.assertEqual(review.version, reviews[0].version)
        self.assertEqual(review.reviewer_username(),
                         reviews[0].reviewer_username)
        self.assertEqual(review.reviewer_displayname(),
                         reviews[0].reviewer_displayname)
        self.assertEqual(review.origin(), reviews[0].origin)
        self.assertEqual(review.distroseries(), reviews[0].distroseries)

    def test_get_reviews_for_extras_pkg(self):
        repo = self.factory.makeRepository(origin='lp-ppa-app-review-board')
        review = self.factory.makeReview(repository=repo)
        api = RatingsAndReviewsAPI()

        reviews = api.get_reviews(
            language=review.language,
            origin=review.repository.origin,
            distroseries=review.repository.distroseries,
            packagename=review.package_name()
        )

        self.assertEqual(1, len(reviews))

    def test_get_reviews_for_app(self):
        review = self.factory.makeReview(app_name='something')
        api = RatingsAndReviewsAPI()

        reviews = api.get_reviews(
            language=review.language,
            origin=review.repository.origin,
            distroseries=review.repository.distroseries,
            packagename=review.package_name(),
            appname=review.app_name,
            version=review.version
        )

        self.assertEqual(1, len(reviews))
        self.assertEqual(review.package_name(), reviews[0].package_name)
        self.assertEqual(review.app_name, reviews[0].app_name)
        self.assertEqual(review.summary, reviews[0].summary)
        self.assertEqual(review.version, reviews[0].version)

    def test_get_reviews_page_2(self):
        # The rnr_client can get multiple pages of results.
        software_item = self.factory.makeSoftwareItem()
        repository = self.factory.makeRepository()
        for count in range(3):
            self.factory.makeReview(
                software_item=software_item, language='en',
                repository=repository)

        with patch_settings(API_REVIEWS_PAGINATION=2):
            reviews = RatingsAndReviewsAPI().get_reviews(
                language='en', origin=repository.origin,
                distroseries=repository.distroseries,
                packagename=software_item.package_name, page=2)

        self.assertEqual(1, len(reviews))

    def test_get_reviews_vary_only_on_accept(self):
        """Check that Vary header is set only to Accept."""
        response = self.client.get(
            '/reviews/api/1.0/reviews/filter/en/ubuntu/kinky/any/chess/')

        self.assertEqual(200, response.status_code)
        self.assertEqual(response['vary'], 'Accept')


class RnRGetReviewTestCase(RnRClientTestCase):

    def test_get_review(self):
        review = self.factory.makeReview()

        api = RatingsAndReviewsAPI()
        review_detail = api.get_review(review_id=int(review.id))

        self.assertEqual(review_detail.id, review.id)
        self.assertEqual(review_detail.origin, review.origin())
        self.assertEqual(review_detail.distroseries, review.distroseries())
        invalidate_paginated_reviews_for(review)


class RnRDeleteReviewTestCase(RnRClientTxTestCase):

    def test_delete_review(self):
        review = self.factory.makeReview()
        api = self.make_api_for_user(review.reviewer)
        response = api.delete_review(review_id=review.pk)

        self.assertEqual(review.id, response['id'])
        self.assertNotEqual(None, response['date_deleted'])


class RnRModifyReviewTestCase(RnRClientTxTestCase):

    def test_modify_review(self):
        review = self.factory.makeReview()
        api = self.make_api_for_user(review.reviewer)

        rating = 4
        summary = 'summary'
        review_text = 'review text'

        response = api.modify_review(review_id=int(review.id),
                                     rating=rating,
                                     summary=summary,
                                     review_text=review_text)

        self.assertEqual(review.id, response.id)
        self.assertEqual(rating, response.rating)
        self.assertEqual(summary, response.summary)
        self.assertEqual(review_text, response.review_text)


class RnRSubmitReviewTestCase(RnRClientTxTestCase):

    def _submit_review(self, review_request, user=None, mock_verify=None):
        """Submit a review request and return the result with a mock."""
        api = self.make_api_for_user(user, mock_verify=mock_verify)
        return api.submit_review(review=review_request)

    def _make_review_request(self, **kwargs):
        review_options = dict(
            package_name='pacman', summary='Too difficult',
            version='0.1.2', rating=3, arch_tag='i386',
            review_text='I keep getting mowed down by ghosts',
            language='en', origin='ubuntu', distroseries='lucky')
        review_options.update(kwargs)
        return ReviewRequest(**review_options)

    def test_submit_review(self):
        package_names = [
            'myapp',
            'com.example.myapp',
        ]
        for package_name in package_names:
            review = self._make_review_request(
                package_name=package_name, rating=4,
                review_text=u'This is great.')

            review_result = self._submit_review(review)

            self.assertEqual(package_name, review_result.package_name)
            self.assertEqual(4, review_result.rating)
            self.assertEqual(u'This is great.', review_result.review_text)
            self.assertFalse(review_result.hide)

    def test_submit_review_anonymous(self):
        api = RatingsAndReviewsAPI()
        self.assertRaises(ValidationException, api.submit_review)

    def test_submit_review_fetch(self):
        """Submit a review using a token/consumer that we don't have.

        The auth layer will need to fetch the information from SSO.
        """
        assert User.objects.count() == 0
        review = self._make_review_request()

        # XXX: This hack is avoid automatic user creation by
        # `self.make_api_for_user `.
        from django.contrib.auth.models import AnonymousUser
        review_result = self._submit_review(review, AnonymousUser())

        self.assertEqual('pacman', review_result.package_name)
        self.assertEqual(1, User.objects.count())

    @patch.multiple(settings, ALLOW_MULTIPLE_REVIEWS_FOR_TESTING=True)
    @patch('reviewsapp.forms.web_services.verify_packagename_in_repository')
    def test_submit_review_nonexisting_distroseries(self, mock_verify):
        mock_verify.return_value = True

        # First time for this package_name & distroseries - should call verify
        review = self._make_review_request(
            package_name='pacman', distroseries='lucky',
            review_text='First review')
        self._submit_review(review, mock_verify=mock_verify)
        self.assertEqual(mock_verify.call_count, 1)

        # Same package_name & distroseries - should not call verify
        review = self._make_review_request(
            package_name='pacman', distroseries='lucky',
            review_text='Second review')
        self._submit_review(review, mock_verify=mock_verify)

        self.assertEqual(mock_verify.call_count, 1)

        # Change the distroseries - should call verify
        review = self._make_review_request(
            package_name='pacman', distroseries='bogus',
            review_text='Third review')
        self._submit_review(review, mock_verify=mock_verify)
        self.assertEqual(mock_verify.call_count, 2)

    @patch.multiple(settings, ALLOW_MULTIPLE_REVIEWS_FOR_TESTING=True)
    @patch('reviewsapp.forms.web_services.verify_packagename_in_repository')
    def test_submit_review_different_archs(self, mock_verify):
        # First time for this package_name & arch - should call verify
        review = self._make_review_request(
            package_name='pacman', arch_tag='i386',
            review_text='First review')
        self._submit_review(review, mock_verify=mock_verify)
        self.assertEqual(mock_verify.call_count, 1)

        # Same package_name & arch - should not call verify
        review = self._make_review_request(
            package_name='pacman', arch_tag='i386',
            review_text='Second review')
        self._submit_review(review, mock_verify=mock_verify)
        self.assertEqual(mock_verify.call_count, 1)

        # Change the arch - should call verify
        review = self._make_review_request(
            package_name='pacman', arch_tag='amd64',
            review_text='Third review')
        self._submit_review(review, mock_verify=mock_verify)
        self.assertEqual(mock_verify.call_count, 2)

    def test_passive_mode(self):
        # The server is in passive mode by default
        review = self._make_review_request()

        review_result = self._submit_review(review)
        # In passive mode the review will not be hidden
        self.assertFalse(review_result.hide)

        # ensure that no ReviewModerationFlag was created
        self.assertEqual(ReviewModeration.objects.count(), 0)
        self.assertEqual(ReviewModerationFlag.objects.count(), 0)

    def test_active_mode(self):
        # Put the server in active mode
        RNRSettings.set_moderation_mode(RNRSettings.ACTIVE_MODERATION)

        review = self._make_review_request()

        review_result = self._submit_review(review)
        # In active mode the review will be hidden
        self.assertTrue(review_result.hide)

        # ensure that a ReviewModerationFlag was created
        self.assertEquals(ReviewModeration.objects.count(), 1)
        self.assertEquals(ReviewModerationFlag.objects.count(), 1)
        moderation = ReviewModeration.objects.all()[0]
        ReviewModerationFlag.objects.all()[0]
        # ensure that the review is in pending status
        self.assertEqual(moderation.status, ReviewModeration.PENDING_STATUS)


class RnRFlagReviewTestCase(RnRClientTestCase):

    def test_flag_review_returns_new_flag(self):
        user = self.factory.makeUser()
        api = self.make_api_for_user(user)
        review = self.factory.makeReview(reviewer=user)
        flag = api.flag_review(
            review_id=int(review.pk), reason='my_reason', text='my text')
        self.assertEqual('my_reason', flag['summary'])


class UsefulnessAPITestCase(RnRClientTxTestCase):
    def setUp(self):
        super(UsefulnessAPITestCase, self).setUp()
        self.user = self.factory.makeUser()

    def test_submit_useful(self):
        api = self.make_api_for_user(self.user)
        software_item = self.factory.makeSoftwareItem()
        review = self.factory.makeReview(software_item=software_item)
        response = api.submit_usefulness(
            review_id=int(review.id), useful="True")
        self.assertEqual('Created', response)
        count = Usefulness.objects.filter(review=review, useful=True).count()
        self.assertEqual(1, count)

    def test_submit_not_useful(self):
        api = self.make_api_for_user(self.user)
        software_item = self.factory.makeSoftwareItem()
        review = self.factory.makeReview(software_item=software_item)
        response = api.submit_usefulness(
            review_id=int(review.id), useful="False")
        self.assertEqual('Created', response)
        count = Usefulness.objects.filter(review=review, useful=True).count()
        self.assertEqual(0, count)

    def test_submit_useful_multiple_times(self):
        api = self.make_api_for_user(self.user)
        software_item = self.factory.makeSoftwareItem()
        review = self.factory.makeReview(software_item=software_item)
        response = api.submit_usefulness(
            review_id=int(review.id), useful="True")
        self.assertEqual('Created', response)
        count = Usefulness.objects.filter(review=review, useful=True).count()
        self.assertEqual(1, count)
        # submit same again and check that we don't get created
        response = api.submit_usefulness(
            review_id=int(review.id), useful="True")
        self.assertEqual('Not modified', response)
        count = Usefulness.objects.filter(review=review, useful=True).count()
        self.assertEqual(1, count)
        # submit again, this time with different usefulness
        response = api.submit_usefulness(
            review_id=int(review.id), useful="False")
        self.assertEqual('Updated', response)
        count = Usefulness.objects.filter(review=review).count()
        self.assertEqual(1, count)

    def test_get_usefulness(self):
        api = RatingsAndReviewsAPI()
        software_item = self.factory.makeSoftwareItem()
        review = self.factory.makeReview(software_item=software_item)
        user = self.factory.makeUser()
        usefulness = self.factory.makeUsefulness(review=review, user=user)

        response = api.get_usefulness(
            review_id=int(review.id), username=user.username)

        self.assertEqual(1, len(response))
        self.assertEqual(user.username, response[0].get('username'))
        self.assertEqual(review.id, response[0].get('review_id'))
        self.assertEqual(usefulness.useful, response[0].get('useful'))
