#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the web services utility """

from __future__ import absolute_import

__metaclass__ = type
__all__ = ['InvalidatePaginatedReviewsTestCase']

import urllib

from django.core.cache import cache
from django.core.urlresolvers import reverse

from reviewsapp.tests.helpers import patch_settings
from reviewsapp.tests.factory import TestCaseWithFactory
from reviewsapp.utilities import (
    cache_key_for_url,
    invalidate_paginated_reviews_for,
)


class InvalidatePaginatedReviewsTestCase(TestCaseWithFactory):
    def _url_for_package(
            self, package_name, origin='ubuntu',
            distro_series='lucid', language='en', version='1.0', page=None):
        kwargs = {
            'language': language,
            'origin': origin,
            'distroseries': distro_series,
            'version': version,
            'package_name': urllib.quote_plus(package_name),
        }
        if page is not None:
            kwargs['page'] = page

        return reverse('rnr-api-reviews-for-package', kwargs=kwargs)

    def _request_reviews_for_package(
            self, package_name, origin='ubuntu',
            distro_series='lucid', language='en', version='1.0', page=None):

        url = self._url_for_package(
            package_name, origin, distro_series, language, version, page)

        with patch_settings(API_REVIEWS_PAGINATION=2):
            response = self.client.get(url)

        return response

    def _populate_reviews_and_cache(
            self, package_name, origin='ubuntu',
            distro_series='lucid', language='en', version='1.0', pages=1):
        package = self.factory.makeSoftwareItem(package_name)
        repository = self.factory.makeRepository(
            origin=origin, distro_series=distro_series)

        # We ensure that API_REVIEWS_PAGINATION is set to 2 reviews per
        # page for our tests.
        reviews = []
        for review_num in range(pages * 2):
            review = self.factory.makeReview(
                software_item=package,
                language=language, repository=repository,
                version=version)
            reviews.append(review)

        # Now ensure the cache is populated by requesting each url.
        self._request_reviews_for_package(package_name)
        for page_num in range(1, pages + 1):
            self._request_reviews_for_package(package_name, page=page_num)

        return reviews

    def test_invalidates_all_pages(self):
        # The cache is invalidated for all pages.
        reviews = self._populate_reviews_and_cache('package_foo', pages=3)
        for page in (None, 1, 2, 3):
            url = self._url_for_package('package_foo', page=page)
            self.assertTrue(cache_key_for_url(url) in cache)

        invalidate_paginated_reviews_for(reviews[0])

        for page in (None, 1, 2, 3):
            url = self._url_for_package('package_foo', page=page)
            self.assertFalse(cache_key_for_url(url) in cache)

    def test_invalidates_with_space(self):
        # The cache is invalidated for all pages.
        reviews = self._populate_reviews_and_cache('package;App with space')
        invalidate_paginated_reviews_for(reviews[0])
        url = self._url_for_package('package;App with space')
        self.assertFalse(cache_key_for_url(url) in cache)

    def test_invalidates_single_review_request(self):
        review = self.factory.makeReview()
        url = reverse('rnr-api-review-for-review-id', args=[review.id])
        self.client.get(url)
        self.assertTrue(cache_key_for_url(url) in cache)

        invalidate_paginated_reviews_for(review)

        self.assertFalse(cache_key_for_url(url) in cache)
