#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'cache_key_for_url',
    'invalidate_paginated_reviews_for',
    'pnormaldist',
    'wilson_score',
    'get_db_engine',
    'is_using_postgres',
]

import math

from django.conf import settings
from django.core.cache import cache
from django.core.urlresolvers import reverse
from django.http import HttpRequest
from django.utils.cache import get_cache_key
from django.utils.http import urlquote_plus


def cache_key_for_url(url):
    """
    Return the key used by django's caching system for a (non-varied) url.
    """
    request = HttpRequest()
    request.path = url
    request.method = "GET"
    return get_cache_key(request)


def invalidate_paginated_reviews_for(review):
    # XXX achuni 2011-01-19: Here we're invalidating pages only for the exact
    # version, language, distroseries and origin of the review.  We'd also
    # need to invalidate partial matches, up to including all reviews
    # for the package (/reviews/any/any/any/any/packagename/).  This is
    # a lot of work.  Is it worth it?
    language = review.language
    origin = review.repository.origin
    distroseries = review.repository.distroseries
    package_name = review.softwareitem.package_name
    app_name = review.app_name
    version = review.version

    if app_name:
        package_app_name = u"{0};{1}".format(package_name, app_name)
    else:
        package_app_name = package_name

    # aaronp 12/08/2011: Moved import here to avoid circular import
    # failure. Caused by importing utilities in models.Review and
    # importing models.Review in utilities
    from reviewsapp.models import Review

    num_reviews_for_pkg = Review.objects.filter(
        repository__origin=origin, repository__distroseries=distroseries,
        softwareitem__package_name=package_name,
        app_name=app_name, hide=False, language=language,
        version=version).count()

    # Ensure we've got a list of all the possible page numbers,
    # including None (ie. first page is displayed if there is no page
    # number).
    pages = range(1, num_reviews_for_pkg + 1)
    pages.insert(0, None)
    for page in pages:
        url_kwargs = dict(
            language=language, origin=origin, distroseries=distroseries,
            version=version, package_name=urlquote_plus(package_app_name))
        if page:
            url_kwargs['page'] = page

        package_url = reverse('rnr-api-reviews-for-package', kwargs=url_kwargs)
        key = cache_key_for_url(package_url)
        if key in cache:
            cache.delete(key)

    # Delete the cache entry for the single review request also.
    url = reverse('rnr-api-review-for-review-id', args=[review.id])
    key = cache_key_for_url(url)
    if key in cache:
        cache.delete(key)


# The following functions pnormaldist and wilson_score are based on
# the ruby versions.
# See http://evanmiller.org/how-not-to-sort-by-average-rating.html
# As indicated in USC spec at: http://bit.ly/wyigxd
def pnormaldist(qn):
    '''
    Inverse normal distribution, based on the Ruby statistics2.pnormaldist
    '''
    b = [1.570796288, 0.03706987906, -0.8364353589e-3,
         -0.2250947176e-3, 0.6841218299e-5, 0.5824238515e-5,
         -0.104527497e-5, 0.8360937017e-7, -0.3231081277e-8,
         0.3657763036e-10, 0.6936233982e-12]

    if qn < 0 or qn > 1:
        raise ValueError("qn must be between 0.0 and 1.0")
    if qn == 0.5:
        return 0.0

    w1 = qn
    if qn > 0.5:
        w1 = 1.0 - w1
    w3 = -math.log(4.0 * w1 * (1.0 - w1))
    w1 = b[0]
    for i in range(1, 11):
        w1 = w1 + (b[i] * math.pow(w3, i))

    if qn > 0.5:
        return math.sqrt(w1 * w3)
    else:
        return -math.sqrt(w1 * w3)


def wilson_score(pos, n, power=0.2):
    if n == 0:
        return 0
    z = pnormaldist(1 - power / 2)
    phat = 1.0 * pos / n
    return (phat + z * z / (2 * n) - z * math.sqrt(
        (phat * (1 - phat) + z * z / (4 * n)) / n)) / (1 + z * z / n)


def get_db_engine():
    return settings.DATABASES['default']['ENGINE']


def is_using_postgres():
    return 'postgresql_psycopg2' in get_db_engine()
